/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

// This code is adapted from the book _Java in a Nutshell_ by David Flanagan.
// Written by David Flanagan.  Copyright (c) 1996 O'Reilly & Associates.

import java.io.*;
import java.net.*;
import results.*;
import java.util.*;

// Starts specific imports
import soif.*;
import parsing.*;
import query.*;
import parsing.filter.*;
import parsing.ranking.*;
import resource.*;
import wais.*;
import STARTSConfiguration.*;
import http.*;

/**
 * A server that listens on a port
 */
public class server extends Thread {
    public final static int DEFAULT_PORT = 6789;
    protected int port;
    protected ServerSocket listen_socket;
    ResourceDescription resource;

    /**
     * Exit with an error message, when an exception occurs.
     * @param e    The exception
     * @param msg  Error message
     */
    public static void fail(Exception e, String msg) {
        System.err.println(msg + ": " +  e);
        System.exit(1);
    }

    /**
     * Create a ServerSocket to listen for connections on;  start
     * the thread.
     * @param port port to listen for connections on
     */
    public server(int port) {
        if (port == 0) port = DEFAULT_PORT;
        this.port = port;
        try { listen_socket = new ServerSocket(port); }
        catch (IOException e) { fail(e, "Exception creating server socket"); }

		// Initialize the STARTS resource
		resource = new ResourceDescription();

        this.start();
    }

    /**
     * The body of the server thread.  Loop forever, listening for and
     * accepting connections from clients.  For each connection,
     * create a Connection object to handle communication through the
     * new Socket.
     */
    public void run() {
        try {
            while(true) {
                Socket client_socket = listen_socket.accept();
                Connection c = new Connection(client_socket, resource);
            }
        }
        catch (IOException e) {
            fail(e, "Exception while listening for connections");
        }
    }

    /**
     * Start the server up, listening on an optionally specified port
     * @param args Input arguments (the port)
     */
    public static void main(String[] args) {
        int port = 0;
        if (args.length == 1) {
            try { port = Integer.parseInt(args[0]);  }
            catch (NumberFormatException e) { port = 0; }
        }
        new server(port);
    }
}

/**
 * This class is the thread that handles all communication with a
 * client.
 */
class Connection extends Thread {
    protected Socket client;
    protected DataInputStream in;
    protected PrintStream out;
    ResourceDescription resource;

    /**
     * Initialize the streams and start the thread.
     * @param client_socket the Socket from which a connection is open
     * from a client.
     * @param resource The description of the resource that this server
     * is providing.
     */
    public Connection(Socket client_socket, ResourceDescription resource) {
        client = client_socket;
        try {
            in = new DataInputStream(client.getInputStream());
            out = new PrintStream(client.getOutputStream());
        }
        catch (IOException e) {
            try { client.close(); } catch (IOException e2) { ; }
            System.err.println("Exception while getting socket streams: " + e);
            return;
        }
        this.resource = resource;
        this.start();
    }

    /**
     * Provide server services.  Listen for a service specification
     * string and dispath to the handler for that service
     */
    public void run() {
        String serviceRequest;

            try {

                serviceRequest = in.readLine();

                // Request dispatcher. The CGI script invoked through HTTP turns URL's
                // into STARTS service requests.

                // query request goes to primary source and secondary sources.
                if (serviceRequest.equals("QUERY")) {
                    String primarySourceName = in.readLine();
		            SOIFObject soif = new SOIFObject(in);
		            String templateType = soif.GetTemplateType();

            		if (! templateType.equals("SQuery")) {
		                throw new SOIFException("Illegal Query SOIF");
            		}

                    out.print(HTTPHeader.MakeHeader(200, "text/plain"));

                    // send the query to the primary source
                    Query query = HandleQuery(primarySourceName, soif, out);

                    // now to secondary sources if any
                    if (query.otherSources != null) {
                        for (int i = 0; i < query.otherSources.length; i++) {
                            HandleQuery(query.otherSources[i], soif, out);
                        }
                    }
                }

                else if (serviceRequest.equals("SOURCEMETA")) {
                    HandleSourceMeta(in, out);
                }

                else if (serviceRequest.equals("SOURCECONTENT")) {
                    HandleSourceContent(in, out);
                }

                else if (serviceRequest.equals("RESOURCEMETA")) {
                    HandleResourceMeta(out);
                }

                // This should never happen
                else {
                    out.print(HTTPHeader.MakeHeader(500, "text/plain"));
                    out.println("Internal Error");
                }


            }

            // socket problems
            catch (IOException e1) {
                out.print(HTTPHeader.MakeHeader(500, "text/plain"));
                out.println(e1.getMessage());
            }

            // bad soif request
            catch (SOIFException e2) {
                out.print(HTTPHeader.MakeHeader(400, "text/plain"));
                out.println(e2.getMessage());
            }

            // Unsupported STARTS filter or ranking expression
            catch (UnsupportedQueryException e3) {
                out.print(HTTPHeader.MakeHeader(400, "text/plain"));
                out.println(e3.getMessage());
            }

            // error at the source engine
            catch (SourceException e4) {
                out.print(HTTPHeader.MakeHeader(500, "text/plain"));
                out.println(e4.getMessage());
            }

            // Parser uses general exception for parse errors in
            // filter or ranking expressions
            catch (Exception e6) {
                if (e6 instanceof ParseException) {
                    out.print(HTTPHeader.MakeHeader(400, "text/plain"));
                    out.println(e6.getMessage());
                }
                else {
                    out.print(HTTPHeader.MakeHeader(500, "text/plain"));
                    out.println("Internal parser error: " + e6.getMessage());
                }
            }
            finally {
                try {
                    in.close();
                    out.close();
                }
                catch (IOException e) {}
            }

        }

    /**
     * Handle the STARTS query service.
     * @param in   The input stream to read the query SOIF from.
     * @param out  The output stream to which to send the response
     * @param source   The source for the resource.
     * @return the query context of the query that was handled.
     * @exception SourceException if an error occurred with the source engine.
     * @exception SourceQueryException if the query was not able to be processed
     * by the source engine.
     * @exception Exception if a parsing error occurred.
     * @exception SOIFException if an error occurred converting the query
     * input to a SOIF object.
     * @exception UnsupportedQueryException if the query was not processable
     * as a STARTS query.
     * @exception IOException if an error occurred in socket I/O during
     * processing the query.
     */
    private Query HandleQuery(String sourceName, SOIFObject soif, PrintStream out)
        throws SourceException, SourceQueryException, Exception,
        SOIFException, UnsupportedQueryException, IOException {

        Query query;
        int numResults;
        SourceDescription source;

        // The CGI script turns the source related URL into a source name.
        if ((source = resource.GetSource(sourceName)) == null) {
            throw new UnsupportedQueryException(
                "Source not found at this resource: " + sourceName);
        }


        // For scanning and parsing of the filter expression
		parsing.filter.YYlex yylFilter;
		parsing.filter.YYparse yypFilter;

        // For scanning and parsing of the ranking expresssion
		parsing.ranking.YYlex yylRanking;
		parsing.ranking.YYparse yypRanking;


		query = new Query(soif, resource, source);

        // null out expressions that we don not support.
		if (!source.Ranking_p()) {
            query.rankingExpression = null;
		}
		if (!source.Filters_p()) {
            query.filterExpression = null;
		}

		// Must have either a ranking expression or a filter expression
		if (query.filterExpression== null &&
		    query.rankingExpression == null) {
		    throw new SOIFException(
		        "No query that this resource can process.");
		}

        // Parse, check, and translate the filter expression
        if (query.filterExpression != null &&
            query.filterExpression.length() > 0) {

			// Initialize the filter scanner and parser objects
			yylFilter = new parsing.filter.YYlex(
			    new StringBufferInputStream(
			        query.filterExpression), System.out);;
			yypFilter = new parsing.filter.YYparse(yylFilter);
			yypFilter.SetQuery(query);

			// Now run the parser on the filter
	        yypFilter.yyparse();
		    query.originalFilter = (ExpNode) yypFilter.yyval;

            // Do source and semantic sanity checking.
	        query.processedFilter =
	            ((STARTSnode) query.originalFilter).Check();


            // Now translate the filter expression
            query.sourceFilter =
                ((FilterNode) query.processedFilter).TranslateToFilter();
        }

        // Parse, check, and translate the ranking expression
        if (query.rankingExpression != null
            && query.rankingExpression.length() > 0) {

			// Initialize the ranking scanner and parser objects
			yylRanking = new parsing.ranking.YYlex(
			    new StringBufferInputStream(
			        query.rankingExpression), System.out);;
			yypRanking = new parsing.ranking.YYparse(yylRanking);
			yypRanking.SetQuery(query);

			// Now run the parser on the ranking

	        yypRanking.yyparse();
		    query.originalRanking = (ExpNode) yypRanking.yyval;

            // Now check for source support

	        query.processedRanking =
	            ((STARTSnode) query.originalRanking).Check();


            // Now translate the ranking expression
            query.sourceRanking =
                ((RankingNode)
                    query.processedRanking).TranslateToRanking();
            query.rankingTerms = new Vector();
            ((RankingNode) query.processedRanking).ExtractTerms(
                query.rankingTerms);
        }


        // Do the search at the source.
        numResults = query.source.DoSourceSearch(query);

        if (numResults > 0) {
            // Select only those results that meet score minimum
            numResults = query.results.MinScoreCheck();
        }

        // Sort the results
        if (numResults > 0 ) {
            query.results.SortResults(query.sortFields,
                query.sortDescending);
            // Trim results to size specified in query
            numResults = query.results.TrimResults();
        }


        // Assemble the common result set header (even if 0 results)
        SOIFObject resultSOIF = query.results.ResultSetToSOIF();
        StringBuffer resultsBuf =
            resultSOIF.SOIFToBuffer(true);

        out.println(resultsBuf.toString());

        // Assemble the document result SOIF's
        if (numResults > 0) {
            Vector docs = query.results.GetSortedDocuments();
            for (int i = 0; i < docs.size(); i++) {
                SOIFObject docSOIF =
                    ((Document) docs.elementAt(i)).DocumentToSOIF();
                StringBuffer docsBuf =
                    docSOIF.SOIFToBuffer(true);

                out.println(docsBuf.toString());
            }
        }
        return query;
    }

    /**
     * Handle the STARTS source metadata service
     * @param out  Output stream to which to send source metadata
     * @param source   Source for which metadata is to be provided
     * @exception UnsupportedQueryException if the request is for an unknown
     * source.
     * @exception IOException if an error occurs in socket I/O while
     * processing the request.
     */
    private void HandleSourceMeta(DataInputStream in, PrintStream out)
        throws UnsupportedQueryException, IOException {
        String sourceName = in.readLine();
        SourceDescription source;

        if ((source = resource.GetSource(sourceName)) == null) {
            throw new UnsupportedQueryException(
                "Source not found at this resource: " + sourceName);
        }

        SOIFObject sourceSOIF = source.SourceToSOIF();
        StringBuffer buf = sourceSOIF.SOIFToBuffer(true);
        out.print(HTTPHeader.MakeHeader(200, "text/plain"));
        out.println(buf.toString());
    }

    /**
     * Handle the STARTS source content service
     * @param out  Output strem to which to send source content data
     * @param source   Source for which content data is to be sent
     * @exception SourceException if an error occurs extracting content
     * data for the source
     * @exception SourceQueryException if the request is for an unknown
     * source.
     * @exception IOException if an error occurred in socket I/O during
     * processing the query.
     */
    private void HandleSourceContent(DataInputStream in, PrintStream out)
        throws SourceException, UnsupportedQueryException, IOException  {
        String sourceName = in.readLine();
        SourceDescription source;

        if ((source = resource.GetSource(sourceName)) == null) {
            throw new UnsupportedQueryException(
                "Source not found at this resource: " + sourceName);
        }

        out.print(HTTPHeader.MakeHeader(200, "text/plain"));
        SOIFObject sourceSOIF = source.ContentSummaryHeaderToSOIF();
        StringBuffer buf = sourceSOIF.SOIFToBuffer(false);
        out.println(buf.toString());
        source.ContentSummaryFieldsDump(out);
    }

    /**
     * Handle the STARTS resource metadata service
     * @param out  Output stream to which to send resource metadata
     */
    private void HandleResourceMeta(PrintStream out) {

        SOIFObject resourceSOIF = resource.ResourceToSOIF();
        StringBuffer buf = resourceSOIF.SOIFToBuffer(true);
        out.print(HTTPHeader.MakeHeader(200, "text/plain"));
        out.println(buf.toString());
    }
}
