/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package parsing;


import java.util.*;

public
class Object_Vector {

    protected static final int defaultsize = 16;

    protected int _length;
    protected Object _contents[];
    protected int _delta;

    public Object[] contents() { return _contents;}
    public void contents(Object v[]) { _contents = v; }

    public
    Object_Vector(int initsize, int delta)
    {
	super();
	_contents = new Object[initsize];
	_delta = delta;
    }

    public
    Object_Vector(int initsize)
    {
	this(initsize,0);
    }

    public
    Object_Vector()
    {
	this(defaultsize);
    }

    protected
    void
    extendto(int sz)
    {
	int oldsz = _contents.length;
	if(sz > oldsz) {
	    Object newbuf[];
	    int newsz;
	    if(oldsz == 0) oldsz = defaultsize;
	    if(_delta > 0) {
		newsz = oldsz + _delta;
	    } else {
		newsz = 2*oldsz;
	    }
	    newbuf = new Object[newsz];
	    if(oldsz > 0) System.arraycopy(_contents, 0, newbuf, 0, _length);
	    _contents = newbuf;
	}
    }

    public
    void
    setSize(int newsz)
    {
	if(newsz > _length) {
	    extendto(newsz);
// the following is suppressed on the assumption
// that the elementype is a scalar
//	} else {
//	    int i;
//	    for(i=newsz;i < _length;i++) {
//		_contents[i] = null;
//	    }
	}
	_length = newsz;
    }

    public
    int
    capacity()
    {
	return _contents.length;
    }

    public
    int
    size()
    {
	return _length;
    }

    public
    boolean
    isEmpty()
    {
	return _length == 0;
    }

    public
    boolean
    contains(Object e)
    {
	return indexOf(e, 0) >= 0;
    }

    public
    int
    indexOf(Object e)
    {
	return indexOf(e, 0);
    }

    public
    int
    indexOf(Object e, int index)
    {
	int i;
	for(i=index;i < _length;i++) {
	    if(e == _contents[i]) return i;
	}
	return -1;
    }

    public
    int
    lastIndexOf(Object e)
    {
	return lastIndexOf(e, _length);
    }

    public
    int
    lastIndexOf(Object e, int index)
    {
	int i;
	for(i=index-1;i >= 0;i--) {
	    if(e == _contents[i]) return i;
	}
	return -1;
    }

    public
    Object
    elementAt(int index)
    {
	if(index < 0 || index >= _length) {
	    throw new ArrayIndexOutOfBoundsException(""+index);
	}
	return _contents[index];
    }

    public
    Object
    firstElement()
    {
	return elementAt(0);
    }

    public
    Object
    lastElement()
    {
	return elementAt(_length-1);
    }

    public
    void
    setElementAt(Object e, int index)
    {
	if(index < 0 || index >= _length) {
	    throw new ArrayIndexOutOfBoundsException(index + "");
	}
	_contents[index] = e;
    }

    public
    void
    removeElementAt(int index)
    {
	if(index < 0 || index >= _length) {
	    throw new ArrayIndexOutOfBoundsException(index + "");
	}
	int i = _length - index - 1;
	if(i > 0) {
	    System.arraycopy(_contents, index+1, _contents, index, i);
	}
	_length--;
// again, assume elemtype is scalar
//	_contents[_length] = null;
    }

    public
    void
    insertElementAt(Object e, int index)
    {
	if(index < 0 || index >= _length + 1) {
	    throw new ArrayIndexOutOfBoundsException(index + "");
	}
	extendto(_length+1);
	System.arraycopy(_contents,index,_contents,index+1,_length - index);
	_contents[index] = e;
	_length++;
    }

    public
    void
    addElement(Object e)
    {
	extendto(_length+1);
	_contents[_length++] = e;
    }

    public
    boolean
    removeElement(Object e)
    {
	int i = indexOf(e);
	if(i >= 0) {
	    removeElementAt(i);
	    return true;
	}
	return false;
    }

    public
    void
    removeElementRange(int first, int count)
    {
	if(count < 0 || first < 0 || first+count > _length) {
	    throw new ArrayIndexOutOfBoundsException("range "+first+"/"+count);
	}
	int mv = (_length - count);
	if(count > 0 && mv > 0) {
	    System.arraycopy(_contents,first+count,_contents,0,mv);
	}
	_length -= count;
// again, assume elemtype is scalar
//	for(int i=0;i<count;i++) {_contents[_length+i] = null;}
    }

    public
    void
    removeAllElements()
    {
	removeElementRange(0,_length);
    }

    public
    void
    removeLastElement()
    {
	removeElementAt(_length-1);
    }

    public
    Object
    clone() throws CloneNotSupportedException
    {
	Object_Vector v = (Object_Vector)super.clone();
	Object newbuf[] = new Object[_length];
	System.arraycopy(_contents, 0, newbuf, 0, _length);
	v.contents(newbuf);
	return v;
    }

    public
    void
    copy(Object_Vector v)
    {
	fill(v.contents(),v.size());
    }

    public
    void
    fill(Object v[])
    {
	fill(v,v.length);
    }

    public
    void
    fill(Object v[], int count)
    {
	fill(v,0,count);
    }

    public
    void
    fill(Object v[], int offset, int count)
    {
	setSize(count);
	System.arraycopy(v, offset, _contents, 0, count);
    }

    public
    String
    toString()
    {
	int i;
	int len = size();
	StringBuffer b = new StringBuffer();

	b.append("[");
	for(i=0;i < len;i++) {
	    if(i > 0) b.append(", ");
	    b.append(_contents[i]);
	}
	b.append("]");
	return b.toString();
    }
}
