/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/
/**
 * Class added by Sandy Payette 6/23/97
 */

package query;

import java.util.*;


public class AttrSetDcore1 extends AttrSet {

// Characteristics of a Dublin Core (Dcore-1) attribute set object


    /**
     * Identifying name of the attribute set.
     */
    public String AttrSetName = "dcore-1";


    /**
     * Array of fields that are part of the Dublin Core (dcore-1)
     * attribute set.  The columns are:
     * Column 1 - field  (the field name)
     * Column 2 - required field indicator (true or false).
     *            "true" means a source MUST recognize the field
     *            "false" means a source may optionally recognize the field
     */

    public static String [] [] AttrSetFields = {
        {"TITLE", "true"},
        {"CREATOR", "true"},
        {"SUBJECT", "false"},
        {"DESCRIPTION", "false"},
        {"PUBLISHER", "false"},
        {"CONTRIBUTORS", "false"},
        {"DATE", "true"},
        {"TYPE", "false"},
        {"FORMAT", "false"},
        {"IDENTIFIER", "true"},
        {"SOURCE", "false"},
        {"LANGUAGE", "false"},
        {"RELATION", "false"},
        {"COVERAGE", "false"},
        {"RIGHTS", "false"}
    };

    private Hashtable attrSetFieldTable =
        new Hashtable(AttrSetFields.length);


    // Load the attribute set field table (attrSetFieldTable) which
    // contains all possible fields in the attribute set.

    protected void LoadFieldTable() {
        int i;

        for (i = 0; i < AttrSetFields.length; i++) {
            attrSetFieldTable.put(AttrSetFields[i][0], AttrSetFields[i][1]);
        }
    }


    /**
     * predicate to determine whether an attribute set supports
     * a particular field.
     * @param field    field to test
     * @return true or false depending on if the field is supported
     * by the attribute set.
     */

	public boolean attrSetFieldSupported_p(String field) {
	    if (!attrSetFieldTable.containsKey(field)) {
	        return(false);
	    }
        else {
            return(true);
        }
	}



    /**
     * Hashtable for identification of source-specific fields and for
     * translation of source-specific fields
     */

	private Hashtable sourceFields =
		new Hashtable();


    // Load source-specific fields hashtable which contains only the
    // attribute set fields supported by the source

	protected void LoadSourceFields(String [][] fieldsTranslation) {
	    int i;

		for (i = 0; i < fieldsTranslation.length; i++) {
		    if (fieldsTranslation[i][0].equals("dcore-1")) {

			    // load the source fields table with the key as the field name
			    // and the value the language token list
			    sourceFields.put(fieldsTranslation[i][1],
			                     fieldsTranslation[i][3]);
			}
		}

	}

	/**
     * Predicate to determine whether a specific field/attribute set
     * pair is supported by the source.
     * @param field    field to test
     * @return true of false depending on if the field/attribute set
     * pair is supported at the source.
     */

	public boolean FieldSupportBySource(String field) {
	    if (!sourceFields.containsKey(field)) {
	        return(false);
	    }
        else {
            return(true);
        }
	}

	/**
     * Hashtable for identification of source-specific fields and for
     * translation of source-specific fields
     */

	private Hashtable fieldsXlate =
	    new Hashtable();

    // Load Hashtable for identification of source-specific fields and for
    // translation of source-specific fields

	protected void LoadFieldXlate(String [][] fieldsTranslation) {
	    int i;

		for (i = 0; i < fieldsTranslation.length; i++) {
		    if (fieldsTranslation[i][0].equals("dcore-1")) {

			    // load the translation table with the key as the field name
			    // and the value the translated name
			    fieldsXlate.put(fieldsTranslation[i][1],
			                    fieldsTranslation[i][2]);
			}
		}

	}


    /**
     * Return a String that is a translation from STARTS field
     * to a WAIS field.
     * @param field    Field name of the source field to be translated.
     * @return the String translation of the input field.
     */

    public String TranslateField(String field) {
        return((String) fieldsXlate.get(field));
    }




    /**
     * Constructor
     */

    public AttrSetDcore1(String [][] fieldsTranslation) {


        LoadFieldTable();
        LoadSourceFields(fieldsTranslation);
        LoadFieldXlate(fieldsTranslation);

    }

}
