/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;

import resource.*;

/**
 * An expression tree node to handle STARTS boolean operators (both
 * filter and ranking expressions.  A BoolOp is an interior node
 * of an expression tree with two children.  Every BoolOp has one
 * property with attribute name "operator" and a value of type string.
 */

public class BoolOp extends ExpNode
    implements STARTSnode, FilterNode, RankingNode
{
	/**
	 * Build a new BoolOp.
	 * @param query    the STARTS query that is the context of the tree.
	 * @param operator the String operator
	 */
	public BoolOp(Query query, String operator) {
		super(query);
		SetPropertyValue("operator", operator);
	}


	/**
	 * return the operator
	 * @return The string version of the operator for this boolop
	 */
	public String GetOperator() {
		return((String) GetPropertyValue("operator"));
	}

    /**
     * Return a string that is the unparsed representation of this
     * BoolOp node
     * @return The unparsed string.
     */
	public String Unparse() {
        Vector children = GetChildren();

	    return(
	        "(" +
	        ((ExpNode) children.elementAt(0)).Unparse() +
	        " " +
	        GetOperator() +
	        " " +
	        ((ExpNode) children.elementAt(1)).Unparse()
	    );
	}

    /**
     * Return a new sub-tree that is the result of checking this node
     * for semantic inconsistencies and source support.
     * @exception UnsupportedQueryException If semantic checking of the
     * STARTS query fails.
     * @return The subtree modified by the checking operation.
     */
    public ExpNode Check() throws UnsupportedQueryException {
        Vector children = GetChildren();
        ExpNode child0;
        ExpNode child1;

        child0 = ((STARTSnode) children.elementAt(0)).Check();
        child1 = ((STARTSnode) children.elementAt(1)).Check();

       // children may dissapear in the checking process (e.g.,
       // stop word elimination
       if (child0 != null && child1 != null) {
            BoolOp newBO = new BoolOp(query, GetOperator());
            newBO.AddNewChild(child0);
            newBO.AddNewChild(child1);
            return(newBO);
        }
        else if (child0 != null) {
            return(child0);
        }
        else {
            return(child1);
        }
    }

    /**
     * Translate this BoolOp to the sub-tree that represents the
     * source-specific filter expression.
     * @exception SourceQueryException If there is an error in
     * translation to the source filter.
     * @return The abstract tree representation of the source filter exp.
     */
    public ExpNode TranslateToFilter() throws SourceQueryException {
        Vector children = GetChildren();

        ExpNode child1 = ((FilterNode) children.elementAt(0)).TranslateToFilter();
        ExpNode child2 = ((FilterNode) children.elementAt(1)).TranslateToFilter();

        return(query.source.BoolOpToFilter(this, child1, child2));
    }

    /**
     * Translate this BoolOp to the sub-tree that represents the
     * source-specific ranking expression.
     * @exception SourceQueryException If there is an error in
     * translation to the source ranking.
     * @return The abstract tree representation of the source ranking exp.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        Vector children = GetChildren();

        ExpNode child1 = ((RankingNode) children.elementAt(0)).TranslateToRanking();
        ExpNode child2 = ((RankingNode) children.elementAt(1)).TranslateToRanking();

        return(query.source.BoolOpToRanking(this, child1, child2));

    }

    /**
     * Extract the terms in the subtree that descendes from this
     * BoolOp.
     * @param termVector   A Vector into which the extracted terms
     * are placed.
     */
    public void ExtractTerms(Vector termVector) {
        Vector children = GetChildren();

        ((RankingNode) children.elementAt(0)).ExtractTerms(termVector);
        ((RankingNode) children.elementAt(1)).ExtractTerms(termVector);
    }
}

