/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;
import results.*;

/**
 * Superclass of all nodes that are components of the abstract syntax
 * trees used to hold filter and ranking expressions.
 */
public abstract class ExpNode
{
    /**
     * The children of this node
     */
	Vector children;

	/**
	 * The parent of this node.
	 */
	ExpNode parent;

	/**
	 * The number of this child among the children of its parent.
	 */
	protected int childNum;

    /**
     * Property list for this ExpNode
     */
	Hashtable properties;

    /**
     * The query context in which this Expnode exists.
     */
	protected Query query;

    /**
     * Create a new ExpNode.
     * @param  query   The query context of this Expnode
     */
	public ExpNode(Query query) {
		children = null;
		parent = null;
		properties = new Hashtable();
		this.query = query;
	}

	/**
	 * Add a new child to this node
	 * @param newChild the new child
	 */
	public void AddNewChild(ExpNode newChild) {
		if (children == null) {
			children = new Vector(1);
		}
		children.addElement(newChild);
		if (newChild != null) {
    		newChild.SetParent(this);
    		newChild.childNum = children.indexOf(newChild);
    	}
	}


	/**
	 * Return an Vector of the children of this ExpNode
	 * @return the Vector of chilren.
	 */
	public Vector GetChildren() {
		if (children.isEmpty() || children == null) {
			return null;
		}
		else {
			return children;
		}
	}

	/**
	 * Return the number of children that this node has
	 * @return the number of children that this node has.
	 */
	public int GetNumChildren() {
		return(children.size());
	}


	/**
	 * Return the parent of this node
	 * @return the ExpNode that is the parent of this node.
	 */
	public ExpNode GetParent() {
		return(parent);
	}

	/**
	 * Return the number of this child node relative to its parent
	 * @return the integer number that is this node's child number
	 */
	public int GetChildNum() {
	    return(childNum);
	}

	/**
	 * Return an enumeration which is the property names for this
	 * node.
	 * @return an Enumeration containing all the property names of
	 * this ExpNode.
	 */
	public Enumeration GetPropertyNames() {
		return(properties.keys());
	}

	/**
	 * Return the value of a specific property (or null if the
	 * property does not exist).  The caller should know the type
	 * of the value so it can cast it to the proper class.
	 * @return An Object that is the value of the requested property.
	 */
	public Object GetPropertyValue(String key) {
		return(properties.get(key));
	}

	/**
	 * Replace a child of this node
	 * @param newChild the child that is the replacement.
	 * @param childNum the number of the child to replace
	 */
	public void ReplaceChild(ExpNode newChild, int childNum) {
		children.setElementAt(newChild, childNum);
		newChild.SetParent(this);
	}

	/**
	 * set a specific property value
	 * @param name the name of the property to set
	 * @param value    the value of the property
	 */
	protected void SetPropertyValue(String name, Object value) {
		properties.put(name, value);
	}

    /**
     * Set the parent of this node
     * @param parent   The parent.
     */
	protected void SetParent(ExpNode parent) {
	    this.parent = parent;
	}

    /**
     * Return a string representation of this node
     * @return A String that is the unparsing of this node.
     */
    public abstract String Unparse();

    /**
     * Return the query context of this node
     * @return a Query that is the context of this node.
     */
    public Query GetQuery() {
        return query;
    }

 }

