/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import resource.*;

import java.util.*;


/**
 * An ExpNode that represents a Field node in either a ranking or
 * filter expression.  Field nodes have no children and two properties:
 * - field name: string
 * - attribute set: string
 */
public class Field extends ExpNode
    implements STARTSnode, FilterNode, RankingNode
{

	/**
	 * Build a new field using the default attribute set.
	 * @param query    the query context of this node
	 * @param  field   the field name for this field node.
	 */
	public Field(Query query, String field) {
		super(query);
		SetPropertyValue("field name", field);
		SetPropertyValue("attribute set",
		    query.defaultAttributeSet);
	}

    /**
     * Build a new field specifying the attribute set and field
     * name.
     * @param query    the query context for this node.
     * @param attributeSet the attribute set
     * @param  field   the field name
     */
	public Field(Query query, String attributeSet, String field) {
		super(query);
		SetPropertyValue("field name", field);
		SetPropertyValue("attribute set", attributeSet);
	}


	/**
	 * Return the field name of this node
	 * @return the string that is the field name (e.g. title)
	 */
	public String GetFieldName() {
		return((String) GetPropertyValue("field name"));
	}

	/**
	 * Return the attribute set of this Field
	 * @return the String that is the attribute set (e.g.,
	 * basic-1
	 */
	public String GetAttributeSet() {
		return((String) GetPropertyValue("attribute set"));
	}

/******************************************************************/

    /**
     * Predicate answering that this attribute set is supported by
     * the source in the query context.
     * @return true or false depending on whether the attribute set is supported
     *         by the source.
     * (Added by Sandy Payette 6/24/97)
     */

    public boolean AttrSetSupported_p() {
       return(query.source.AttrSetSupported(GetAttributeSet()));
    }


    /**
     *  Predicate answering that the field is part of the specified attribute set.
     *  Depending on what attrset/field combo we are evaluating, we want to evaluate
     *  the attribute set fields table in EITHER AttrSetBasic1 or AttrSetDcore1.
     *  Method attrSetFieldSupported uses the appropriate attribute set object in
     *  the context of the source and the query.
     * (Added by Sandy Payette 6/24/97)
     */

    public boolean FieldinAttrSet() {

        return((query.source.GetAttrSet(GetAttributeSet())).attrSetFieldSupported_p(
                GetFieldName()));

    }



    /**
     * Predicate answering that the field is supported BY THE SOURCE in
     * the query context.
     * @return true or false depending on wheter this Field is supported
     * by its Query context.
     *
     * (Modified by Sandy Payette 7/9/97 to use method FieldSupportBySource
     *  using the attribute set object in the context of the source and the query.
     */

	public boolean Supported_p() {
	    return((query.source.GetAttrSet(GetAttributeSet())).FieldSupportBySource(
	            GetFieldName()));

	    //return(query.source.FieldSupported_p(
	    //    GetAttributeSet(), GetFieldName()));
	}

/***************************************************************************/


    /**
     * Predicate answering whether the specfied language is supported
     * for this field.
     * @param language the language in ISO format
     * @return true or false depending on whether the input language
     * is supported for this field.
     */
	public boolean LanguageSupported_p(String language) {
	    return(query.source.LanguageSupportedForField_p(
	        GetAttributeSet(), GetFieldName(), language));
	}

    /**
     * Return the unparsed string represenation of this field node
     * @return the Sting representation of this Field node.
     */
	public String Unparse() {
	    String attributeSet = GetAttributeSet();
	    String field = GetFieldName();
	    if (attributeSet.equals(query.defaultAttributeSet)) {
	        return(field);
	    }
	    else {
    	    return(
	            "{" +
	            attributeSet +
    	        " " +
	            field + "}");
	    }
	}
/***************************************************************/

    /** Modified by Sandy Payette 6/24/97
     *  First check if the source supports the attribute set,
     *  then check if the field is legally part of this attribute set,
     *  then:
     *
     * Check this field node to see whether it is supported by the
     * source in the query context.  If not supported, then
     * return the "wildcard" field "any".  If supported then
     * return the field node
     */
    public ExpNode Check() throws UnsupportedQueryException {

        // no children to check

        String attributeSet = GetAttributeSet();
        String field = GetFieldName();


        if (!AttrSetSupported_p()) {
            throw new UnsupportedQueryException(
                "This source does not support the attribute set: " + attributeSet);
            }
        else
           if (!FieldinAttrSet()) {
               throw new UnsupportedQueryException(
               "Field " + field + " is not part of the attribute set" + " " + attributeSet);
               }

            else
                if (!Supported_p()) {

                    /** When Field is instantiated with "any" the general WAIS
                     *  dictionary and the general WAIS inverted index will be
                     *  used.  This exception is "attribute-set free" meaning
                     *  that the code that invokes the general dictionary/index
                     *  does not care what attribute set was specified in the query.
                     **/
                    return(new Field(query, "any"));
                    }
                else {
                    return(this);
                    }

    }

/****************************************************************/
    /**
     * Translate this field node in the context of a source
     * specific filter expression.
     * @return the sub-tree that is the source translated filter
     * equivalanet of this Field subtree.
     * @exception SourceQueryException if there are translation
     * errors.

     */
    public ExpNode TranslateToFilter() throws SourceQueryException {
        return(query.source.FieldToFilter(this));
    }


    /**
     * Translate this field node in the context of a source
     * specific ranking expression.
     * @return the subtree that is the source equivalent ranking
     * expression for this Field sub-tree.
     * @exception SourceQueryException if there are translations
     * errors.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        return(query.source.FieldToRanking(this));
    }

    /**
     * noop for extracting terms from an expnode subtree.  Must be
     * defined since the superclass defines it as abstract, but
     * nothing ever descends from Fields so do nothing.
     */
    public void ExtractTerms(Vector termVector) {
    }

}

