/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import resource.*;
import java.util.*;


/**
 * An ExpNode that represents a LString node in either a ranking or
 * filter expression.  LString nodes have no children and two
 * properties:
 * - string value: the value of the LString
 * - language: the language in RFC1877 format
 */
public class LString extends ExpNode
    implements STARTSnode, FilterNode, RankingNode
{


	/**
	 * Build an LString with no specified language
	 * @param query    the query context of this node
	 * @param  theString   string value of the LString
	 */
	public LString(Query query, String theString) {
		super(query);
		SetPropertyValue("string value", DeQuoteString(theString));
	}

	/**
	 * Build an LString with a specfied language
	 * @param query    query context of this node.
	 * @param  language    language in RFC1877 format
	 * @param theString    string value of the LString
	 */
	public LString( Query query, String language, String theString) {
		super(query);
		SetPropertyValue("string value", DeQuoteString(theString));
		SetPropertyValue("language", language);
	}

	/**
	 * Return the string value of this LString
	 * @return the string value of this LString
	 */
	public String GetString() {
		return((String) GetPropertyValue("string value"));
	}

    /**
     * Return the language of this LString
     * @return the language in ISO format.
     */
	public String GetLanguage() {
		return((String) GetPropertyValue("language"));
	}

    /**
     * trim the quotes (double) from a string
     * @param theString    string to dequote
     * @return the input string stripped of its enclosing quotes.
     */
	private String DeQuoteString(String theString) {
	    if (theString.startsWith("\"")) {
	        theString = theString.substring(1);
	    }
	    if (theString.endsWith("\"")) {
	        theString = theString.substring(0, theString.length() - 1);
	    }

	    return(theString);
	}

    /**
     * Return the string version of this LString
     * @return the string equivalent of this LString sub-tree
     */
	public String Unparse() {
        String language = GetLanguage();

        if (language != null) {
            return("[" + language + "\"" + GetString() +  "\"" + "]");
        }
        else {
            return("\"" + GetString() + "\"");
        }
    }


    /**
     * Check this LString node relative to the source.
     * @return The LString sub-tree modified by the Check
     * operation.  If the string in the Lstring is a stop word,
     * return null
     * @exception UnsupportedQueryException if the sub-tree
     * contains semantically unsupported elements
     */
    public ExpNode Check() throws UnsupportedQueryException {
        if (query.source.StopWord_p(GetString())) {
            throw new UnsupportedQueryException(
                "Invalid expression: a stopword has been used." );

            //return null;
        }
        else {
            return(this);
        }
    }


    /**
     * Translate this LString node in the context of a source
     * specific filter expression.
     * @return the sub-tree that is the source translated filter
     * equivalent of this LString subtree.
     */
    public ExpNode TranslateToFilter() throws SourceQueryException {
        return(query.source.LStringToFilter(this));
    }


    /**
     * Translate this LString node in the context of a source
     * specific ranking expressin.
     * @return the sub-tree that is the source translated ranking
     * equivalent of this LString subtree.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        return(query.source.LStringToRanking(this));
    }

    /**
     * noop for extracting terms from an expnode subtree.  Must be
     * defined since the superclass defines it as abstract, but
     * nothing ever descends from Fields so do nothing.
     */
    public void ExtractTerms(Vector termVector) {
    }

}

