/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;

import resource.*;

/**
 * An ExpNode that represents a Modifier List node in either a ranking
 * or filter expression.  Mofifier List nodes have no childrens
 * and one property:
 * - modifiers - the Vector of modifiers in the list
 */
public class ModifierList extends ExpNode
    implements STARTSnode, FilterNode, RankingNode
{

    /**
     * Construct a new modifier list
     * @param query    query context of the modifier list
     */
	public ModifierList(Query query) {

		super(query);
		Vector modifiers = new Vector();
		SetPropertyValue("modifiers", modifiers);
	}



    /**
     * Add a new modifier to this modifier list.  A modifier
     * is a string which optionally may include a prefix
     * attribute set.  Some examples are 'stem' or
     * 'basic-1 stem'
     * @param modifer  the new modifier to add
     */
	public void AddNewModifier(Modifier modifier) {
		Vector modifiers;

		modifiers = (Vector) GetPropertyValue("modifiers");
		modifiers.addElement(modifier);
		SetPropertyValue("modifiers", modifiers);
	}


    /**
     * Return the vector of modifiers in this modifier list
     * @return the vector of modifiers for this modifier list, Each
     * item in the vector is of type Modifier.
     */
	public Vector GetModifiers() {
		Vector modifiers;

		modifiers = (Vector) GetPropertyValue("modifiers");
        return(modifiers);
	}

	/**
	 * Replace a modifier in the modifier list
	 * @param modifiers the new modifier list
	 */
	public void SetModifiers(Vector modifiers) {
        SetPropertyValue("modifiers", modifiers);
    }

    /**
     * Predicate to determine whether the modifiers support specified
     * languages for this source
     * @param language the language to test.
     * @return true or false depending if the modifiers in this list
     * all support the specified language.
     */
    public boolean LanguageSupported_p(String language) {
	    Vector modifiers = GetModifiers();
	    int lim = modifiers.size();

	    for (int i = 0; i < lim; i++) {
            Modifier modifier = (Modifier) (modifiers.elementAt(i));
            if (!modifier.LanguageSupported_p(language)) {
                return(false);
            }
        }
        return(true);
    }

    /**
     * Return the string form of this modifier list
     * @return the String equivalent of this Modifier List sub-tree
     */
    public String Unparse() {
        Vector mods = GetModifiers();
        StringBuffer buf = new StringBuffer();

        for (int i = 0; i < mods.size(); i++) {
            Modifier m = (Modifier) mods.elementAt(i);
            String attrSet = m.GetAttributeSet();
            String mod = m.GetModifier();
            if (attrSet.equals(query.defaultAttributeSet)) {
                buf.append(mod + " ");
            }
            else {
                buf.append("{" + attrSet + " " + mod + "} ");
            }
        }
        return(buf.toString());
    }

    /**
     * scan through modifier list and return a new list containing only
     * those modifiers that are supported for the source context.
     * @return ExpNode The transformed subtree resulting from semantic
     * checks on this ModifierList sub-tree.
     * @exception UnsupportedQueryException if the checking process finds
     * errors in the ModifierList sub-tree.
     */
    public ExpNode Check()  throws UnsupportedQueryException {

        // no children to check
        Vector modifiers = GetModifiers();
        int lim = modifiers.size();
        ModifierList newModifiers = new ModifierList(query);

        for (int i = 0; i < lim; i++) {
            Modifier modifier = (Modifier) (modifiers.elementAt(i));
            if (modifier.Supported_p()) {
                newModifiers.AddNewModifier(modifier);
            }
        }

        return(newModifiers);
    }

    /**
     * Translate this modifier list node in the context of a source
     * specific filter expression.
     * @return the sub-tree that is the result of source translation of
     * this modifier list to a filter equivalent.
     */
    public ExpNode TranslateToFilter() throws SourceQueryException {
        return(query.source.ModifierListToFilter(this));
    }


    /**
     * Translate this modifier list in the context of a source
     * specific ranking expression.
     * @return the sub-tree that is the result of source translation of
     * this modifier list to a ranking equivalent.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        return(query.source.ModifierListToRanking(this));
    }

    /**
     * noop for extracting terms from an expnode subtree.  Must be
     * defined since the superclass defines it as abstract, but
     * nothing ever descends from Fields so do nothing.
     */
    public void ExtractTerms(Vector termVector) {
    }

}

