/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;
import resource.*;



/**
 * An ExpNode that represents a Proximity Operator in either a ranking
 * or filter expression.  ProxOp nodes have two children (both of
 * which are always terms) and two properties:
 * distance - an integer value which is the proximity distance
 * wordOrderMatters - a boolean value determining whether
 * word order matters in the proximity
 */
public class ProxOp extends ExpNode
    implements STARTSnode, FilterNode, RankingNode
{

    /**
     * Build a new ProxOp.
     * @param  query   the query context of the ProxOp
     * @param distance integer distance value of the proximity operator
     * @param wordOrderMatters boolean value determining whether word
     * order matters inthe proximity
     */
	public ProxOp(Query query, int distance, boolean wordOrderMatters) {
		super(query);
		SetPropertyValue("distance",
			new Integer(distance));
		SetPropertyValue("word order",
			new Boolean(wordOrderMatters));

	}

	/**
	 * return the proximity distance
	 * @return the integer value of the proximity distance.
	 */
	public int GetDistance() {
		return(
			((Integer)GetPropertyValue("distance")).intValue());
	}

	/**
	 * predicate determining whether word order matters in this
	 * proximity operator.
	 * @return true or false depending if word order matters for this
	 * proximity operator;
	 */
	public boolean WordOrderMatters_p() {
		return(
			((Boolean)GetPropertyValue("word order")).booleanValue());

	}

    /**
     * Return the String representation of this proximity sub-tree
     * @return the String equivalent of this ProxOp rooted sub-tree
     */
	public String Unparse() {
        Vector children = GetChildren();
        String wo;

        if (WordOrderMatters_p()) {
            wo = "T";
        }
        else {
            wo = "F";
        }

	    return(
	        "(" +
	        ((ExpNode) children.elementAt(0)) +
	        " " +
	        "prox[" +
	        String.valueOf(GetDistance()) +
	        "," +
	        wo +
	        "]" +
	        " " +
	        ((ExpNode) children.elementAt(1))
	    );
	}

    /**
     * Do sanity checking and source support checking for this
     * ProxOp.  For example, STARTS filter syntax does not prohibit
     * expressions like (author="foo") prox (title = "bar").  If
     * everything passes tests ok recurse down to the children.
     * @return The transformed sub-tree resulting from the check operation
     * @exception UnsupportedQueryException if an error is found in the
     * subtree.
     */
    public ExpNode Check() throws UnsupportedQueryException {
        Vector children = GetChildren();
        Term child0 = (Term) children.elementAt(0);
        Term child1 =  (Term) children.elementAt(1);

        if (!child0.GetField().GetFieldName().equalsIgnoreCase(
            child1.GetField().GetFieldName())) {
            throw new UnsupportedQueryException(
                "Fields of term must match for proximity operator");
        }

        child0 = (Term) ((STARTSnode) child0).Check();
        child1 = (Term) ((STARTSnode) child1).Check();

       // children may dissapear in the checking process (e.g., stop word elimination
       if (child0 != null && child1 != null) {
            ProxOp newPO= new ProxOp(query, GetDistance(),
                WordOrderMatters_p());
            newPO.AddNewChild(child0);
            newPO.AddNewChild(child1);
            return(newPO);
        }
        else if (child0 != null) {
            return(child0);
        }
        else {
            return(child1);
        }
    }

    /**
     * Return a new ExpNode representing the source-specific
     * translation of this ProxOp in the context of a filter expression.
     * @return The sub-tree that is the source filter equivalent of the
     * ProxOp sub-tree.
     * @exception SourceQueryException if an error occurs during the
     * translation.
     */
    public ExpNode TranslateToFilter() throws SourceQueryException {
        Vector children = GetChildren();

        ExpNode child1 = ((FilterNode) children.elementAt(0)).
            TranslateToFilter();
        ExpNode child2 = ((FilterNode) children.elementAt(1)).
            TranslateToFilter();

        return(query.source.ProxOpToFilter(this, child1, child2));
    }

    /**
     * Return a new ExpNode representing the source-specific
     * translation of this ProxOp in the context of a filter expression.
     * @return The sub-tree that is the source filter equivalent of the
     * ProxOp sub-tree.
     * @exception SourceQueryException if an error occurs during the
     * translation.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        Vector children = GetChildren();

        ExpNode child1 = ((RankingNode) children.elementAt(0)).TranslateToRanking();
        ExpNode child2 = ((RankingNode) children.elementAt(1)).TranslateToRanking();

        return(query.source.ProxOpToRanking(this, child1, child2));

    }

    /**
     * Recurse through this sub-tree to extract the terms within
     * @param termVector   Vector of terms extracted.
     */
    public void ExtractTerms(Vector termVector) {
        Vector children = GetChildren();

        ((RankingNode) children.elementAt(0)).ExtractTerms(termVector);
        ((RankingNode) children.elementAt(1)).ExtractTerms(termVector);
    }

}

