/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;
import java.io.*;

import soif.*;
import resource.*;
import results.*;
import STARTSConfiguration.*;

/**
 * A Query context for the STARTS server.  Everything done while
 * pressing a STARTS query (parsing, translation, etc.) is done in the
 * context of a Query object.
 */
public class Query {

    /**
     * The SOIFObject that is the input to the query.
     */
    public SOIFObject queryInput;

    /**
     * Version of STARTS of the query.
     */
    public String version;

    /**
     * The string filter expression in the query.
     */
    public String filterExpression;

    /**
     * The string ranking expression in the query.
     */
    public String rankingExpression;

    /**
     * Whether stop words should be dropped in the query processing
     */
    public boolean dropStopWords;

    /**
     * The default attribute set for the query.
     */
    public String defaultAttributeSet;

    /**
     * The Default language for the query.
     */
    public String defaultLanguage;

    /**
     * The fields that should be returned for each document in the
     * query results set.
     */
    public Field answerFields[];

    /**
     * Fields that should be used for sorting the results
     */
    public Field sortFields[];

    /**
     * Boolean array that is isomorphic to sortFields determining
     * whether the respective field should be sorted in ascending
     * or descending order.
     */
    public boolean sortDescending[];

    /**
     * Minimum document score of documents that should be included in
     * the results set.
     */
    public float minimumDocScore;

    /**
     * Maximum number of documents that should be returned as part of
     * the result set.
     */
    public int maximumNumDocuments;

    /**
     * Other sources to which the query should be submitted
     */
    public String otherSources[];

    /**
     * SourceDescription context for the query.
     */
    public SourceDescription source;

    /**
     * Resource context for the query.
     */

    /**
     * The SOIF that is the result object in response to the query.
     */
    public SOIFObject queryOutput;

    /**
     * Abstract tree version of the submitted filter expression.
     */
    public ExpNode originalFilter;    // tree of original filter exp.

    /**
     * Abstract tree version of the processed (source and sanity
     * checked) version of the filter expression.
     */
    public ExpNode processedFilter;

    /**
     * Abstract tree version of the filter expression submitted to the
     * source engine.
     */
    public ExpNode sourceFilter;

    /**
     * Abstract tree version of the submitted ranking expression.
     */
    public ExpNode originalRanking;

    /**
     * Abstract tree version of the processed (source and sanity
     * checked) version of the ranking expression.
     */
    public ExpNode processedRanking;

    /**
     * Abstract tree version of the filter expression submitted to the
     * source engine.
     */
    public ExpNode sourceRanking;

    /**
     * Vector of Terms in the ranking expression (used to construct
     * term statistics for the result set.
     */
    public Vector rankingTerms;

    /**
     * The ResultSet object for the query.
     */
    public ResultSet results;

    /**
     * Construct a new query object
     * @param  qInput  the SOIFObject that is the query input
     * @param resource  Resource context for this query.
     * @param source    Source target for this query.
     */
    public Query(SOIFObject qInput, ResourceDescription resource,
        SourceDescription source) throws SOIFException {
        this.queryInput = qInput;

        if (qInput.GetAttributeValue("Version") != null) {
            version = qInput.GetAttributeValue("Version");
        }
        else {
            version = ServerConfiguration.STARTSVersion;
        }
        filterExpression = qInput.GetAttributeValue("FilterExpression");
        rankingExpression = qInput.GetAttributeValue("RankingExpression");
        if (qInput.GetAttributeValue("DropStopWords") != null) {
            if (qInput.GetAttributeValue("DropStopWords").equalsIgnoreCase("T")) {
                dropStopWords = true;
            }
            else if (qInput.GetAttributeValue("DropStopWords").equalsIgnoreCase("F")) {
                dropStopWords = false;
            }
            else {
                throw new SOIFException("Illegal value for DropStopWords: " +
                    qInput.GetAttributeValue("DropStopWords"));
            }
        }
        else {
            dropStopWords = ServerConfiguration.dropStopWords;
        }

        if (qInput.GetAttributeValue("DefaultAttributeSet") != null) {
            defaultAttributeSet = qInput.GetAttributeValue("DefaultAttributeSet");
        }
        else {
            defaultAttributeSet = ServerConfiguration.defaultAttributeSet;
        }

        defaultLanguage = qInput.GetAttributeValue("DefaultLanguage");

        this.source = source;

        if (qInput.GetAttributeValue("Sources") != null) {
            StringTokenizer st =
                new StringTokenizer(qInput.GetAttributeValue("Sources"));

            otherSources = new String[st.countTokens()];
            int i = 0;
            while (st.hasMoreTokens()) {
                otherSources[i] = st.nextToken();
                if (resource.GetSource(otherSources[i]) == null) {
                    throw new SOIFException(
                        "Unknown source in query: " + otherSources[i]);
                }
                i++;
            }
        }

        SetAnswerFields(qInput);

        SetSortFields(qInput);

        if (qInput.GetAttributeValue("MinDocumentScore") != null) {
            try {
                minimumDocScore =
                    Float.valueOf(qInput.GetAttributeValue(
                        "MinDocumentScore")).floatValue();
            }
            catch (NumberFormatException e) {
                throw new SOIFException(
                    "Illegal value for MinDocumentScore");
            }
        }
        else {
            minimumDocScore = ServerConfiguration.minimumDocScore;
        }


        if (qInput.GetAttributeValue("MaxNumberDocuments") != null) {
            try {
                maximumNumDocuments =
                    Integer.valueOf(qInput.GetAttributeValue(
                        "MaxNumberDocuments")).intValue();
            }
            catch (NumberFormatException e) {
                throw new SOIFException(
                    "Illegal value for MaximumNumberDocuments");
            }
        }
        else {
            maximumNumDocuments =
                ServerConfiguration.maximumNumDocuments;
        }
    }

    /**
     * Set the answer fields for the query object.  These are either
     * the STARTS default answer fields if the query input specifies
     * none, or the result of parsing the string in the query input.
     * @exception SOIFException if an error occrs parsing the answer
     * fields string submitted in the query.
     */
    private void SetAnswerFields(SOIFObject qInput)  throws SOIFException {
        Vector fVector = new Vector();
        String input;

        // Nothing specified in the query input, use the defaults.
        if ((input = qInput.GetAttributeValue("AnswerFields")) == null) {
            int size = ServerConfiguration.defaultAnswerFields.length;
            answerFields = new Field[size];
            for (int i = 0; i < size; i++) {
                answerFields[i] =
                    new Field(this, ServerConfiguration.defaultAnswerFields[i]);
            }
        }

        // Otherwise parse the input
        else {
            StringBufferInputStream buf = new StringBufferInputStream(input);
            int parenCount = 0;
            String attrSet = null, fieldString = null;
            StringBuffer sb = new StringBuffer();
            char c = ' ';

            // a simple parser to handle a field list
            while (buf.available() > 0) {

                c = (char) buf.read();

                // '[' groups attr sets and fields - no multiple nesting allowed
                if (c == '[') {
                    parenCount++;
                    if (parenCount > 1) {
                        throw new SOIFException(
                            "Illegal nesting in Answer Fields list");
                    }

                    // establish a new nesting
                    else {
                        attrSet = null;
                        fieldString = null;
                    }
                }

                // ']' closes a attrset field grouping
                else if (c == ']') {
                    parenCount--;
                    if (parenCount <0) {
                        throw new SOIFException(
                            "Illegal nesting in Answer Fields list");
                    }

                    // close the group - we MUST have both an attrSet and field
                    else {

                        if (sb.length() > 0) {
                            if (fieldString != null) {
                                throw new SOIFException(
                                    "too many tokens in brackets in answer fields specificatioN");
                            }
                            fieldString = sb.toString();
                            sb = new StringBuffer();
                        }


                        // legal, add the new field to the vector
                        if (attrSet != null && fieldString != null) {
                            fVector.addElement(new Field(this, attrSet, fieldString));
                            attrSet = null;
                            fieldString = null;
                        }

                        // grouping without both elements
                        else {
                            throw new SOIFException(
                                "No attribute specifier in answer fields list");
                        }
                    }
                }

                // space type character
                else if (Character.isSpace(c)) {

                    // we have something in the acculumator buffer
                    if (sb.length() > 0) {

                        // in a nesting
                        if (parenCount > 0) {

                            // first token in nesting is attribute set
                            if (attrSet == null) {
                                attrSet = sb.toString();
                                sb = new StringBuffer();
                            }

                            // next token is field
                            else if (fieldString == null) {
                                fieldString = sb.toString();
                                sb = new StringBuffer();
                            }

                            // woops more than one token
                            else {
                                throw new SOIFException(
                                    "More than one token in brackets in answer fields list");
                            }
                        }

                        // not in a nesting, so this space terminates a field token
                        else {
                            fVector.addElement(new Field(this, sb.toString()));
                            sb = new StringBuffer();
                        }
                    }
                    else {
                        // just swallow the space
                    }
                }

                // // a non-space character.  Append it to the acculumator.
                else {
                    sb.append(c);
                }
            }

            // we've swallowed the entire string, clean up
            if (parenCount > 0) {
                throw new SOIFException(
                    "Illegal open grouping in answer fields specification");
            }

            // clean up a remaining token.
            if (sb != null  &&  c != ']') {
                fVector.addElement(new Field(this, sb.toString()));
                sb = new StringBuffer();
            }

            answerFields = new Field[fVector.size()];
            fVector.copyInto(answerFields);
        }


    }

    /**
     * Set the sort fields for the query object.  These are either
     * the STARTS default sort fields if the query input specifies
     * none, or the result of parsing the string in the query input.
     * @exception SOIFException if an error occurs parsing the SortFields
     * string in the query input.
     */
    private void SetSortFields(SOIFObject qInput)  throws SOIFException {
        Vector fVector = new Vector();
        Vector adVector = new Vector();
        String input;

        // Nothing specified - use the defaults.
        if ((input = qInput.GetAttributeValue("SortByFields")) == null) {
            int size = ServerConfiguration.defaultSortFields.length;
            sortFields = new Field[size];
            sortDescending = new boolean[size];
            for (int i = 0; i < size; i++) {
                sortFields[i] =
                    new Field(this, ServerConfiguration.defaultSortFields[i]);
                sortDescending[i] = ServerConfiguration.sortDescending;
            }

        }

        // otherwise parse the input.
        else {
            StringBufferInputStream buf = new StringBufferInputStream(input);
            int parenCount = 0, bracketCount = 0;
            String attrSet = null, fieldString = null, ad = null;;
            Field f = null;
            StringBuffer sb = new StringBuffer();;

            // a simple parser to handle a field list
            while (buf.available() > 0) {

                char c = (char) buf.read();

                // '[' groups attr sets and fields - no multiple nesting allowed
                if (c == '[') {

                    // can only appear with in a field/sort order pair
                    if (parenCount == 0) {
                        throw new SOIFException(
                            "Illegal nesting in Sort fields List");
                    }


                    bracketCount++;
                    if (bracketCount > 1) {
                        throw new SOIFException(
                            "Illegal nesting in Sort Fields list");
                    }

                    // establish a new nesting
                    else {
                        attrSet = null;
                        fieldString = null;
                    }
                }

                // ']' closes a attrset field grouping
                else if (c == ']') {
                    bracketCount--;
                    if (bracketCount <0) {
                        throw new SOIFException(
                            "Illegal nesting in Sort Fields list");
                    }

                    // close the group - we MUST have both an attrSet and field
                    else {

                        if (sb.length() > 0) {
                            if (fieldString != null) {
                                throw new SOIFException(
                                    "too many tokens in brackets in answer fields specificatioN");
                            }
                            fieldString = sb.toString();
                            sb = new StringBuffer();
                        }

                        // legal, add the new field to the vector
                        if (attrSet != null && fieldString != null) {
                            f = new Field(this, attrSet, fieldString);
                            attrSet = null;
                            fieldString = null;
                        }

                        // grouping without both elements
                        else {
                            throw new SOIFException(
                                "No attribute specifier in Sort fields list");
                        }
                    }
                }

                // '(' opens a field/sort order pair
                else if (c == '(') {

                    parenCount++;

                    // cant multiple nest
                    if (parenCount > 1) {
                        throw new SOIFException(
                            "Illegal nesting in Sort Fields list");
                    }

                    // establish a new field/sort order pair
                    else {
                        f = null;
                        attrSet = null;
                        fieldString = null;
                        ad = null;
                    }
                }

                // ')' closes a field/sort order pair
                else if (c == ')') {
                    parenCount--;

                    if (parenCount < 0) {
                        throw new SOIFException(
                            "Illagel nesting in Sort Fields list");
                    }

                    // close the field/sort order pair - we MUST have all
                    // components
                    else {
                        if (sb.length() > 0) {
                            if (ad != null) {
                                throw new SOIFException(
                                    "too many tokens in parens in sort fields specificatioN");
                            }
                            ad = sb.toString();
                            sb = new StringBuffer();
                        }

                        if (f != null && ad != null) {
                            fVector.addElement(f);
                            adVector.addElement(ad);
                            f = null;
                            ad = null;
                        }

                        // elements missing
                        else {
                            throw new SOIFException(
                                "Illegal format in sort fields list");
                        }
                    }
                }

                // space type character
                else if (Character.isSpace(c)) {

                    // we have something in the acculumator buffer
                    if (sb.length() > 0) {

                        // in a attr-set/field nesting
                        if (bracketCount > 0) {

                            // first token in nesting is attribute set
                            if (attrSet == null) {
                                attrSet = sb.toString();
                                sb = new StringBuffer();
                            }

                            // next token is field
                            else if (fieldString == null) {
                                fieldString = sb.toString();
                                sb = new StringBuffer();
                            }

                            // woops more than one token
                            else {
                                throw new SOIFException(
                                    "More than one token in brackets in answer fields list");
                            }
                        }

                        // in a parened field/order specifier pair
                        else if (parenCount > 0) {
                            // could be a field without an attribute set
                            if (f == null) {
                                f = new Field(this, sb.toString());
                                sb = new StringBuffer();
                            }

                            // or it could be a sort order spec
                            else if (ad == null) {
                                ad = sb.toString();
                                sb = new StringBuffer();

                                // check for legal sort specifier
                                if (! (ad.equalsIgnoreCase("a") || ad.equalsIgnoreCase("d"))) {
                                    throw new SOIFException(
                                        "Sort specifier must be a or d");
                                }
                            }

                            // woops to many tokens
                            else {
                                throw new SOIFException(
                                    "too many tokens within () in answer fields specification");
                            }
                        }


                        // this will never happen due to check for non-spaces out
                        // side of brackets
                        else {
                        }
                    }

                    // space after space
                    else {
                        // just swallow the space
                    }
                }

                // a non-space character.
                else {

                    // can't appear on-nested
                    if (parenCount == 0) {
                        throw new SOIFException(
                            "Illegal format in sort order specification");
                    }

                    // just append to accumulator
                    else   {
                        sb.append(c);
                    }
                }
            }

            if (parenCount > 0) {
                throw new SOIFException(
                    "Unclose group in sort fields specification");
            }
            sortFields = new Field[fVector.size()];
            fVector.copyInto(sortFields);
            sortDescending = new boolean[adVector.size()];
            for (int i = 0; i < adVector.size(); i++) {
                sortDescending[i] = ((String) adVector.elementAt(i)).equalsIgnoreCase("d");
            }
        }

    }

}
