/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;
import resource.*;

/**
 * An Expnode that represents a Ranking List node in a STARTS ranking
 * expression.  RankingList nodes have a variable number of children
 * and no properties.
 */
public class RankingList extends ExpNode
    implements STARTSnode, RankingNode
{

    /**
     * Contruct a new ranking list node.
     */
	public RankingList(Query query) {
		super(query);
	}

    /**
     * Return the string representation of the sub-tree descending
     * from this ranking list node
     * @return the String equivalent of this RankingList sub-tree
     */
	public String Unparse() {
        Vector children = GetChildren();
        StringBuffer buf = new StringBuffer("list(");

        for (int i = children.size() - 1; i >= 0; i--) {
            buf.append(((ExpNode) children.elementAt(i)).Unparse() + " ");
        }

        buf.append(")");
        return(buf.toString());
    }

    /**
     * Return a new sub-tree that is the result of checking (for both
     * semantic sanity and source support) the sub-tree that descends
     * from this RankingList node.
     * @return the subtree tht is the result of the check operation.
     * @exception UnsupportedQueryException if the RankingList subtree
     * contains semantic errors
     */
    public ExpNode Check( )
        throws UnsupportedQueryException {

        // No checking so descend to chilren
        RankingList newRL = new RankingList(query);
        Vector children = GetChildren();
        int lim = children.size();
        int numChildren = 0;

        for (int i = 0; i < lim; i++) {
            ExpNode newChild = ((STARTSnode) children.elementAt(i)).Check();
            if (newChild != null) {
                newRL.AddNewChild(newChild);
                numChildren++;
            }
        }

        if (numChildren > 0) {
            return(newRL);
        }

        else {
            return(null);
        }
    }

    /**
     * Return the sub-tree that represents the source-specific
     * translation of this ranking list sub-tree
     * @return a sub-tree that is the source translated ranking
     * equivalent of this RankingList sub-tree
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        Vector children = GetChildren();
        Vector newChildren = new Vector(children.size());

        // translate the children first
        for (int i = 0 ; i < children.size(); i++) {
            newChildren.addElement(
                ((RankingNode)children.elementAt(i)).TranslateToRanking());
        }

        return(query.source.RankingListToRanking(
            this, newChildren));
    }

    /**
     * Recurse from this sub-tree to extract the terms in the sub-tree
     * @param rankingTerms The vector of terms.
     */
    public void ExtractTerms(Vector rankingTerms) {
        Vector children = GetChildren();

        for (int i = 0; i < children.size(); i++) {
            ((RankingNode) children.elementAt(i)).ExtractTerms(rankingTerms);
        }
    }
}

