/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;

import regexp.*;
import resource.*;

/**
 * An ExpNode that represents a Term in either a ranking or filter
 * expression.  Term nodes have no properties and three children,
 * which are, in order, a Field node, a ModifierList node (which
 * may be null), and an LString node
 */
public class Term extends ExpNode
    implements STARTSnode, FilterNode, RankingNode
{
    static int FIELDCHILDNUM = 0;
    static int MODLISTCHILDNUM = 1;
    static int LSTRINGCHILDNUM = 2;


    /**
     * Construct a new Term node with only an LString as input (field
     * is set to the default null, and modifier list is null).
     * @param query    The query context of this term node.
     * @param  lString The LString node that is a child of this term.
     */
	public Term(Query query, LString lString) {
		super(query);

		AddNewChild(new Field(query, "any"));
		AddNewChild(null);
		AddNewChild(lString);
	}

    /**
     * Construct a new Term node with the default field (any) and
     * a modifier list and LString.
     * @param query    The query context of this term node.
     * @param modifierList ModifierList to add as a child.
     * @param  lString LString node to add as a child.
     */
	public Term(Query query, ModifierList modifierList, LString lString) {
		super(query);

		AddNewChild(new Field(query, "any"));
		AddNewChild((ExpNode) modifierList);
		AddNewChild((ExpNode) lString);
	}

    /**
     * Construct a new Term node with a field, modifier list, and
     * LString
     * @param  query   The query context of this term node.
     * @param  field   The field node to add as a child.
     * @param modifierList ModifierList to add as a child.
     * @param  lString LString node to add as a child.
     */
	public Term(Query query, Field field, ModifierList modifierList, LString lString) {
		super(query);

		AddNewChild((ExpNode) field);
		AddNewChild((ExpNode) modifierList);
		AddNewChild((ExpNode) lString);
	}

    /**
     * Construct a new Term node with a field and LString specified
     * @param query    The query context of this term node.
     * @param field    The field node to add as a child
     * @param lString  LString node to add as a child.
     */
	public Term(Query query, Field field, LString lString) {
		super(query);

		AddNewChild(field);
		AddNewChild(null);
		AddNewChild(lString);
	}

    /**
     * Return the LString node that is a child of this term
     * @return the LString object that is the child of this term.
     */
	public LString GetlString() {
		Vector children = GetChildren();
		return((LString) children.elementAt(LSTRINGCHILDNUM));
	}

    /**
     * Return the ModifierList node that is a child of this term
     * @return the ModifierList object that is the child of this term.
     */
	public ModifierList GetModifierList() {
		Vector children = GetChildren();
		return((ModifierList) children.elementAt(MODLISTCHILDNUM));
	}

    /**
     * Return the Field node that is a child of this term
     * @return the Field object that is the child of this term
     */
	public Field GetField() {
		Vector children = GetChildren();
		return((Field) children.elementAt(FIELDCHILDNUM));
	}

    /**
     * Return the string represenation of the sub-tree that descends
     * from this term node.
     * @return the String equivalent of this Term sub-tree.
     */
	public String Unparse() {
	    Vector children = GetChildren();

	    String modString;
	    if (children.elementAt(MODLISTCHILDNUM) == null) {
	        modString = "";
	    }
	    else {
	        modString =
	            ((ExpNode) children.elementAt(MODLISTCHILDNUM)).Unparse()
	            + " ";
	    }

	    return(
	        "(" +
	        ((ExpNode) children.elementAt(FIELDCHILDNUM)).Unparse() +
	        " " +
            modString +
	        ((ExpNode) children.elementAt(LSTRINGCHILDNUM)).Unparse() +
	        ")"
	    );
	}


    /**
     * Translate the sub-tree that descends from this term node
     * in the context of a source-specific filter expression.
     * @return the source filter sub-tree that is the equivalent of this
     * Term subtree.
     * @exception SourceQueryException if an error occurs during the
     * translation.
     */
    public ExpNode TranslateToFilter() throws SourceQueryException {
        ExpNode field = GetField().TranslateToFilter();
        ExpNode modList = GetModifierList();
        if (modList != null) {
            modList = ((FilterNode) modList).TranslateToFilter();
        }
        ExpNode lString = GetlString().TranslateToFilter();
        return(query.source.TermToFilter(this, field, modList, lString));
    }

    /**
     * Translate the sub-tree that descends from this term node
     * in the context of a source-specific ranking expression.
     * @return the source ranking sub-tree that is the equivalent of this
     * Term subtree.
     * @exception SourceQueryException if an error occurs during the
     * translation.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        ExpNode field = GetField().TranslateToRanking();
        ExpNode modList = GetModifierList();
        if (modList != null) {
            modList = ((RankingNode) modList).TranslateToRanking();
        }
        ExpNode lString = GetlString().TranslateToRanking();
        return(query.source.TermToRanking(this, field, modList, lString));
    }

    /**
     * Check the sub-tree that descends from this term node for
     * semantic inconsistencies and see whether it is supported
     * by the source in the query context.
     * @return the transforme Term subtree that results from the
     * checking operations.
     * @exception UnsupportedQueryException if an error is encountered
     * in the Term sub-tree during the checking operation.
     */
    public ExpNode Check() throws UnsupportedQueryException {
        Field newField;
        ModifierList newMods = null;
        LString newLString;

        Vector children = GetChildren();

        Field field = (Field) children.elementAt(FIELDCHILDNUM);
        ModifierList modifiers = (ModifierList) children.elementAt(MODLISTCHILDNUM);
        LString lString = (LString) children.elementAt(LSTRINGCHILDNUM);

        // Some sanity checking for funny semantic incosistencies

        // make sure the lString for a date field is in ISO form
        if (field.GetFieldName().equalsIgnoreCase("date-last-modified")) {
            Regexp reg = Regexp.compile("\\d\\d\\d\\d\\-\\d\\d\\-\\d\\d");
            if (reg.match(lString.GetString(),0) == null) {
                throw new UnsupportedQueryException(
                    "Argument for date field must be in YYYY-MM-DD format");
            }
        }

        // recurse to check the children first.
        newField = (Field) field.Check();
        if (modifiers != null) {
            newMods = (ModifierList) modifiers.Check();
        }

        // lString checking removes stop words.  This may result in a null lString which results
        // in a null term
        newLString = (LString) lString.Check();
        if (newLString == null) {
            return(null);
        }
        else {
            return(new Term(query, newField, newMods, newLString));
        }
    }

    /**
     * Add this term to the vector of terms extracted from
     * a scan through a sub-tree
     */
    public void ExtractTerms(Vector termVector) {
        termVector.addElement(this);
        // don't recurse since no terms can descend from this
    }
}


