/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package query;

import java.util.*;
import resource.*;

/**
 * An ExpNode that represents a weighted ranking operator in a STARTS
 * ranking expressin.  Weighted Ranking nodes have one child and one
 * property:
 * ranking value - a float value that is the weight.
 */
public class WeightedRanking extends ExpNode
    implements STARTSnode, RankingNode
{

    /**
     * Construct a new WeightedRanking node
     * @param query    The query context of this node
     * @param rankingValue The weight of the node
     */
	public WeightedRanking(Query query, float rankingValue) {
		super(query);
		SetPropertyValue("ranking value",
			new Float(rankingValue));
	}

    /**
     * Return the ranking value of this node
     * @return the ranking value of this node.
     */
	public float GetRankingValue() {
		return(((Float)GetPropertyValue("ranking value")).floatValue());
	}

    /**
     * Return the string representation of the sub-tree descending
     * from the WeightedRanking node
     * @return String the string equivalent of this WeightedRanking
     * subtree.
     */
    public String Unparse() {
        Vector children = GetChildren();

        return(
            ")" +
            ((ExpNode) children.elementAt(0)).Unparse() +
            " " +
            String.valueOf(GetRankingValue()) +
            ")"
        );
    }

    /**
     * Return a new sub-tree that is the result of checking (for both
     * semantic sanity and source support) the sub-tree that descends
     * from this WeightedRanking node.
     * @return the transformed Weighted Ranking sub-tree as a result
     * of the check process.
     * @exception UnsupporteQueryException if an error is found in the
     * subtree during the checking process.
     */
    public ExpNode Check()
        throws UnsupportedQueryException {
        Vector children = GetChildren();
        ExpNode newChild;
        WeightedRanking newWR = new WeightedRanking(query, GetRankingValue());
        newChild = ((STARTSnode) children.elementAt(0)).Check();

        // the child may have been reduced to null in checking
        if (newChild != null) {
            newWR.AddNewChild(newChild);
            return newWR;
        }
        else {
            return null;
        }
    }

    /**
     * Return the sub-tree that represents the source-specific
     * translation of this weighted ranking sub-tree
     * @return the source equivalent ranking sub-tree of this Weighted
     * ranking sub-tree
     * @exception SourceQueryException if an error occurs in the
     * translation process.
     */
    public ExpNode TranslateToRanking() throws SourceQueryException {
        Vector children = GetChildren();
        // translate the child first.
        ExpNode xChild =
            ((RankingNode)children.elementAt(0)).TranslateToRanking();
        return(query.source.WeightedRankingToRanking(
            this, xChild));
    }

    /**
     * Recurse from this sub-tree to extract the terms in the sub-tree
     * @param rankingTerms The vector of terms.
     */
    public void ExtractTerms(Vector termVector) {
        Vector children = GetChildren();

        ((RankingNode) children.elementAt(0)).ExtractTerms(termVector);
    }
}

