/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package resource;

import java.io.*;
import java.util.*;

import regexp.*;
import wais.*;

/**
 * A class representation of a document in the CSTR collection.  This
 * class allows us to extract information from the document file (e.g,
 * title, author, etc).
 */
public class CSTRDocument extends WAISResultDocument {

    /**
     * translation table to go from STARTS field names to the RFC1807
     * tags in the document source.  Unfortunately this does not match
     * the translation from STARTS field names to WAIS field names (due
     * to the two character restriction for WAIS field names.
     */

/*************************************************************************************/
/** Original hashtable changed.  (Sandy Payette 7/24/97)
    This is problematic because it assumes the implementation
    of only one attribute set, namely basic-1.  Upon enhancing the system to incorporate
    additional attribute sets such as Dublin Core (dcore-1), this translation table
    became quirky.  The first problem here is that the attribute set is not part
    of the key.  Even if we created a concatenated key of attribute-set and field,
    this is still awkward, since we no longer have the one-to-one mapping between
    a field and a document tag. Instead we see a one-to-many relationship, for example,
    "basic-1 linkage" equates to "ID" and "dcore-1 identifier" equates to "ID".
    So, to clean this up, the translation is now between a WAIS field and a document tag.
    Attribute-set/field combinations have already been translated to WAIS fields by
    the time we get to the point of creating a result document.  WAIS fields are more
    appropriately mapped to the actual document tags. One WAIS field will map to one
    document tag.  This also makes sense given that the source document classes
    (CSTRDocument and LINUXDocument) as sub-classes of WAISResultDocument.

    static Hashtable transTable = new Hashtable();
    static {
        transTable.put("title", "TITLE");
        transTable.put("author", "AUTHOR");
        transTable.put("date-last-modified", "ENTRY");
        transTable.put("linkage", "ID");
        transTable.put("body-of-text", "BODY");
        transTable.put("identifier", "ID");
**/

    static Hashtable transTable = new Hashtable();
    static {
        transTable.put("ti", "TITLE");
        transTable.put("au", "AUTHOR");
        transTable.put("dm", "ENTRY");
        transTable.put("id", "ID");
        transTable.put("bd", "BODY");

    }

/*************************************************************************************/

    /**
     * Given a STARTS field name, return the value of that field in the
     * document.
     * @param path the full path of the document file.
     * @param fieldName    the STARTS field name for which information is
     * to be extracted.
     * @param maxLen   the maximum length of the value returned.
     * @return The value of the specified field in the CSTR document.
     * @exception SourceException if an error occurred which getting the
     * field.
     */
    public String GetFieldValueFromDB(String path, String fieldName,
        int maxLen) throws SourceException {
        String line, retVal;
        Regexp reg = Regexp.compile("^([A-Z\\-]+)::");
        Result result;
        String xField = (String) transTable.get(fieldName.toLowerCase());
        if (xField == null) {
            return null;
        }
        StringBuffer buf = new StringBuffer();
        FileInputStream inStream;
        DataInputStream dStream;

        // field not present in the translation table so value must be null.
        if (xField == null) {
            return null;
        }

        try {
            /* !!!!! PLATFORM SPECIFIC CODE !!!!! */

            /* !!!!! NT VERSION !!!!! */

            /**
            int sp = path.lastIndexOf("/");
            path = "E:\\starts_stuff\\cstrdocs\\" + path.substring(sp+1);
            **/

            /* !!!!! NO SOLARIS VERSION !!!!! */

            /* !!!!! END PLATFORM SPECIFIC CODE !!!!! */

            inStream = new FileInputStream(new File(path));
            dStream = new DataInputStream(inStream);

            boolean found = false;

            // loop until end of file (but we will break out of it within
            // when we find the field we are looking for)
            while (dStream.available() > 0) {
                line = dStream.readLine(); // read a line

                // If we haven't found the tag yet...
                if (!found) {

                    // If the line begins with a RFC1807 tag...
                    if ((result = reg.match(line, 0)) != null) {
                        String tag = result.getMatch(1);
                        int pos = result.getMatchEnd();

                        // If the tag is the one we are looking for, append
                        // the rest of the line to the buffer
                        if (xField.equalsIgnoreCase(tag)) {
                            found = true;
                            buf.append(line.substring(pos));
                        }

                        // if we hit the body (and we are not looking for the body)
                        // then don't waste time scanning through all the rest of the
                        // lines.
                        else if (tag.equalsIgnoreCase("BODY")) {
                            break;
                        }

                        // no else since we simply ignore lines without the
                        // tag that we are looking for
                    }
                }

                // otherwise we have found the tag (but haven't hit another
                // tag yet.
                else {

                    // continuation line (does not begin with a tag) - append
                    // it.
                    if ((result = reg.match(line, 0)) == null) {
                        buf.append(" ");
                        buf.append(line);
                    }

                    // another tag line
                    else {
                        String tag = result.getMatch(1);
                        int pos = result.getMatchEnd();

                        // Another instance of the tag, append
                        // the rest of the line to the buffer
                        if (xField.equalsIgnoreCase(tag)) {
                            buf.append(line.substring(pos));
                        }

                        // new tag, we are done
                        else {
                            break;
                        }
                    }
                }
                if (buf.length() > maxLen) {
                    break;
                }
            }

            retVal = buf.toString().trim();

            // chop to maxLen if necessary.
            if (retVal.length() <= maxLen) {
                inStream.close();
                return retVal;
            }
            else {
                inStream.close();
                return(retVal.substring(0, maxLen));
            }
        }
        catch (IOException e) {
            throw new SourceException(e.getMessage());
        }
    }
}
