/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package resource;

import java.util.*;
import java.io.*;
import java.net.*;

import soif.*;
import STARTSConfiguration.*;
import wais.*;
import query.*;
import results.*;

/**
 * Description of the characteristics specific to the CSTR source.
 * The bulk of this class is a set of static varliable definitions
 * that are used to set the values in the SourceDescription super
 * class.  These settings are used in various parts of the implementatin
 * to do things like query translations.  The implementation is
 * organized so that the service provider should create a new
 * <SourceName>SourceDescription class for each source provided by
 * the resource.  Note that this is a type of WAISSourceDescription and
 * many of the settings are done in that superclass.
 */
public final class CSTRSourceDescription extends WAISSourceDescription
{

    /**
     * identifying name of the source.
     */
 	private static String sourceID = "cstr";

	/**
	 * A String value giving the URL where the source should
	 * be queried
	 */
	public static String linkage =
		"http://" + ServerConfiguration.hostName + "/STARTS/QUERY/cstr";

	/**
	 * A String value giving the URL of the source content summary.
	 */
	public static String contentSummaryLinkage =
		"http://" + ServerConfiguration.hostName + "/STARTS/CONTENTSUMMARY/cstr";

	/**
	 * A string indicating the URL from which to get the query
	 * results for a sample document collection.
	 */
	public static String sampleDataBaseResultsURL =
		"http://" + ServerConfiguration.hostName + "/STARTS/SAMPLE/cstr";

    /**
     * URL at which source metadata is available.
     */
   	private String metadataURL = "http://" + ServerConfiguration.hostName + "/STARTS/SOURCEMETA/cstr";

    /**
     * The syntax of the metadata available at the URL above.
     */
   	private String metadataSyntax = "Stanford-1";

/*****************************************************************/

    /**  (Sandy Payette 7/8/97)
     *
     * Translation tables to go from STARTS fields to WAIS fields.
     *
     * This array should list only the fields THAT ARE SUPPORTED by THIS SOURCE.
     * The full list of fields for any particular attribute set, irrespective of
     * a particular source, can be found in attribute set objects (see class AttrSet
     * and its sub-classes).
     *
     * Formatted as a two-D "array of arrays"
     *
     * Array defined as follows:
     * column 0 being the STARTS attribute set,
     * column 1 being the STARTS field, and
     * column 2 the WAIS equivalent.
     * column 3 a set of tokens indicating languages supported for this field
     * When this array is used in the context of a specific attribute
     * set object, the attribute set column is evaluated to load up
     * attribute-set-specific hash tables.  (See class AttrSet and its sub-classes.)
     *
     * Note that this information is source, not engine, specific.
     */


	protected static String[][]fieldsTranslation = {
		{"basic-1", "author", "au", ""},
		{"basic-1", "title", "ti", ""},
		{"basic-1", "date-last-modified", "dm", ""},
		{"basic-1", "any", "any", ""},
		{"basic-1", "linkage", "id", ""},
		{"basic-1", "body-of-text", "bd", ""},
		{"dcore-1", "TITLE", "ti", ""},
		{"dcore-1", "CREATOR", "au", ""},
	    {"dcore-1", "DATE", "dm", ""},
	    {"dcore-1", "IDENTIFIER", "id", ""}
	};

    /**
     * Added by Sandy Payette 6/24/97
     * Hashtable to indicate what attribute sets are supported by this source,
     * and to instantiate that attribute set class the source is loaded.
     * The method "AttrSetSupported" (see super class SourceDescription)
     * will allow a lookup for valid attribute sets.
     */


     static Hashtable attrSetsTable = new Hashtable();

     // Load the sources hashtable

     static {
        attrSetsTable.put("basic-1", new AttrSetBasic1(fieldsTranslation));
        attrSetsTable.put("dcore-1", new AttrSetDcore1(fieldsTranslation));
     }

    //  Note: The static method to get an item out of the hashtable
    //  is found in the super class SourceDescription.

/******************************************************************************/

	/**
	 * A String array where each element indicates,
	 * in RFC1877 format, a language present at the source.
	 */
	public static String[] sourceLanguage = {
		"en-US"
	};

	/**
	 * A String giving the name of the source.
	 */
	public static String sourceName = "Sample CSTR documents";

	/**
	 * A String value giving a free text abstract of this source.
	 */
	public static String sourceAbstract =
		"This is the reference implementation of STARTS using a freeWAIS-sf source";

	/**
	 * A String value giving a free text description of the contraints
	 * or legal prerequisites for accessing the source.
	 */
	public static String accessConstraints =
		"Unrestricted";

	/**
	 * A String value indicating the administrator of the source
	 */
	public static String administrator =
		"Carl Lagoze - Cornell University";

    /**
     * settings for meta attributes in other attribute sets.  This
     * should be a three dimensional array where each element in
     * a row is as follows:
     * 0 - name of the attribute set
     * 1 - name of the attribute
     * 2 - value of the attribute set
     * For example, a row might be initialized to:
     * {"foo-1", "emporer", "Caeser"}
     */
	public static String otherMetaAttributes[][][] = {
	};

/*********************************************************************/


    /**
     * Hashtable holding all supported fields.  key is
     * <attribute-set> <field>
     * value is set of language tokens or empty string if no
     * special languages.
     *
     * Modified by Sandy Payette 7/9/97
     * Now uses the fieldsTranslation array.  Previous version of code had
     * another, redundant array specifying all fields for the source.
     * This hashtable supports Source methods to list all fields for all attribute sets.
     */

	public static Hashtable fieldsTable =
		new Hashtable(fieldsTranslation.length);


	// load the above hash table.
	static {
		int i;
		String fieldkey;


		for (i = 0; i < fieldsTranslation.length; i++) {

		    fieldkey = fieldsTranslation[i][0] + " " + fieldsTranslation[i][1];
		    fieldsTable.put(fieldkey, fieldsTranslation[i][3]);


		}
    }

    public static Enumeration listfields() {
            return(fieldsTable.keys());
    }


    /**
     * Constructor which loads the generic super class with the
     * source specific information.
     */
    public CSTRSourceDescription() {
        super(sourceID);

        // Generic Source stuff
    	((SourceDescription)this).sourceID = sourceID;
    	((SourceDescription)this).metadataURL = metadataURL;
    	((SourceDescription)this).metadataSyntax = metadataSyntax;
        ((SourceDescription)this).fieldsTable = fieldsTable;
        ((SourceDescription)this).attrSetsTable = attrSetsTable;
        ((SourceDescription)this).sampleDataBaseResultsURL = sampleDataBaseResultsURL;
        ((SourceDescription)this).sourceLanguage = sourceLanguage;
        ((SourceDescription)this).sourceName = sourceName;
        ((SourceDescription)this).linkage = linkage;
        ((SourceDescription)this).contentSummaryLinkage = contentSummaryLinkage;
        ((SourceDescription)this).sourceAbstract = sourceAbstract;
        ((SourceDescription)this).accessConstraints = accessConstraints;
        ((SourceDescription)this).administrator = administrator;
        ((SourceDescription)this).otherMetaAttributes = otherMetaAttributes;

        LoadStopWordsFile();


    }


/***********************************************************************/
// sdp - moved this to AttrSet sub classes

    /**
     * Return a String that is a translation from STARTS field
     * to WAIS field.
     * @param attrSet  Attribute set of the WAIS field to be translated.
     * @param field    Field name of the WAIS field to be translated.
     * @return the String translation of the input attribute set/field
     * pair.
     */
    //public  String TranslateField(String attrSet, String field) {
    //    return((String) fieldsXlate.get(attrSet + " " + field));
    //}

/**********************************************************************/
    /**
     * Return an ISO-1807 representation of when the source was modified
     * @return ISO 1807 date of when the source was modified.
     */
    public String GetDateModified() {
        return("1996-10-26");
    }

    /**
     * Return an ISO-1807 representation of when metadata for this
     * source expires.
     * @return ISO-1807 date of when this metadata expires.
     */
    public String GetDateExpires() {
        return("1996-12-1");
    }

}




