/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package resource;

import java.io.*;
import java.util.*;

import regexp.*;
import wais.*;

/**
 * A class representation of a document in the LINUX  collection.  This
 * class allows us to extract information from the document file (e.g,
 * title, author, etc).
 */
public class LINUXDocument extends WAISResultDocument {

    /**
     * translation table to go from STARTS field names to the usenet
     * tags in the document source.  Unfortunately this does not match
     * the translation from STARTS field names to WAIS field names (due
     * to the two character restriction for WAIS field names.
     */

/******************************************************************************/
/** Original hashtable changed.  (Sandy Payette 7/24/97)
    This is problematic because it assumes the implementation
    of only one attribute set, namely basic-1.  Upon enhancing the system to incorporate
    additional attribute sets such as Dublin Core (dcore-1), this translation table
    became quirky.  The first problem here is that the attribute set is not part
    of the key.  Even if we created a concatenated key of attribute-set and field,
    this is still awkward, since we no longer have the one-to-one mapping between
    a field and a document tag. Instead we see a one-to-many relationship, for example,
    "basic-1 linkage" equates to "ID" and "dcore-1 identifier" equates to "ID".
    So, to clean this up, the translation is now between a WAIS field and a document tag.
    Attribute-set/field combinations have already been translated to WAIS fields by
    the time we get to the point of creating a result document.  WAIS fields are more
    appropriately mapped to the actual document tags. One WAIS field will map to one
    document tag.  This also makes sense given that the source document classes
    (CSTRDocument and LINUXDocument) as sub-classes of WAISResultDocument.

    static Hashtable transTable = new Hashtable();
    static {
        transTable.put("title", "Subject");
        transTable.put("author", "From");
        transTable.put("date-last-modified", "Date");
        transTable.put("linkage", "Message-ID");
**/

    static Hashtable transTable = new Hashtable();
    static {
        transTable.put("sb", "Subject");
        transTable.put("fr", "From");
        transTable.put("dt", "Date");
        transTable.put("id", "Message-ID");

    }

/**********************************************************************************/

    /**
     * Given a STARTS field name, return the value of that field in the
     * document.
     * @param path the full path of the document file.
     * @param fieldName    the STARTS field name for which information is
     * to be extracted.
     * @param maxLen   the maximum length of the value returned.
     * @return The value of the specified field in the CSTR document.
     * @exception SourceException if an error occurred which getting the
     * field.
     */
    public String GetFieldValueFromDB(String path, String fieldName,
        int maxLen) throws SourceException {
        String line, retVal;
        Regexp reg = Regexp.compile("^([A-Za-z\\-]+):");
        Result result;
        String xField = (String) transTable.get(fieldName.toLowerCase());
        if (xField == null) {
            return null;
        }
        StringBuffer buf = new StringBuffer();
        RandomAccessFile inStream;

        // field not present in the translation table so value must be null.
        if (xField == null) {
            return null;
        }

        try {
            /* !!!!! PLATFORM SPECIFIC CODE !!!!! */

            /* !!!!! NT VERSION !!!!! */

            /**
            int sp = path.lastIndexOf("/");
            path = "E:\\starts_stuff\\linuxdocs\\" + path.substring(sp+1);
            **/

            /* !!!!! NO SOLARIS VERSION !!!!! */

            /* !!!!! END PLATFORM SPECIFIC CODE !!!!! */

            inStream = new RandomAccessFile(new File(path), "r");

            inStream.seek(0);
            long fileLength = inStream.length();
            boolean found = false;

            // loop until end of file (but we will break out of it within
            // when we find the field we are looking for)
            while (inStream.getFilePointer() < fileLength && !found) {
                line = inStream.readLine(); // read a line


                // If the line begins with a header tag...
                if ((result = reg.match(line, 0)) != null) {
                    String tag = result.getMatch(1);
                    int pos = result.getMatchEnd();

                    // body-of-text is untagged, for others check if this is the
                    // tag.
                    if (!fieldName.equalsIgnoreCase("body-of-text") &&
                        xField.equalsIgnoreCase(tag)) {
                        found = true;
                        buf.append(line.substring(pos));
                    }

                    // no else since we simply ignore lines without the
                    // tag that we are looking for
                }

                // untagged line
                else {

                    // for usenet articles all untagged lines are the body.
                    if (fieldName.equalsIgnoreCase("body-of-text")) {
                        buf.append(line + " ");
                    }
                    // we are past all tagged lines and will never find the field
                    // we are looking for
                    else {
                        break;
                    }
                }

                if (buf.length() > maxLen) {
                    break;
                }
            }

            retVal = buf.toString().trim();

            // chop to maxLen if necessary.
            if (retVal.length() <= maxLen) {
                inStream.close();
                return retVal;
            }
            else {
                inStream.close();
                return(retVal.substring(0, maxLen));
            }
        }
        catch (IOException e) {
            throw new SourceException(e.getMessage());
        }
    }
}
