/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package resource;

import soif.*;
import resource.*;
import java.util.*;
import STARTSConfiguration.*;
import query.*;

import java.io.*;

/**
    Abstract superclass for all source descriptions to which STARTS
    queries can be directed.
*/
public abstract class SourceDescription {

    // characteristics of a source object


    /**
     * identifying name of the source.
     */
    protected String sourceID;

    /**
     * URL at which source metadata is available.
     */
    protected String metadataURL;

    /**
     * The syntax of the metadata available at the URL above.
     */
    protected String metadataSyntax;

/*************************************************************************/

    /** Added by Sandy Payette 6/24/97
     *  Hashtable holding all supported attribute sets.
     *  Key is attribute set.
     *  Value is attribute set class contructor method
     */
    protected Hashtable attrSetsTable;




    /** Added by Sandy Payette 6/24/97
     *  Method to determine whether an attribute set is supported by
     *  the source.
     *  @param attributeSet  attribute set to test.
     *  @return true of false depending on whether the field/attribute set
     *  pair is supported at the source.
     */


    public boolean AttrSetSupported(String attributeSet) {
	    if (!attrSetsTable.containsKey(attributeSet)) {
	        return(false);
	    }
        else {
            return(true);
        }
	}

	/** Added by Sandy Payette 7/1/97
	 *  Method to return the attribute set object.
	 *  This us used when we want to access the invoke the
	 *  variables and methods of a particular sub-class of the super
	 *  class AttrSet.
	 */

	public AttrSet GetAttrSet(String attributeSet) {
	    return((AttrSet) attrSetsTable.get(attributeSet));
	}


/*************************************************************************/

    /**
     * Hashtable holding all supported fields.  key is
     * <attribute-set> <field>
     * value is set of language tokens or empty string if no
     * special languages.
     */
    protected Hashtable fieldsTable;

    /**
     * Hashtable included all supported modifiers.  key is
     * <attribute-set> <modifier>
     * value is set of language tokens or empty string if no special
     * languages
     */
    protected Hashtable modifiersTable;

    /**
     * Hashtable of all supported operators.  Key is operator,
     * value is null
     */
    protected Hashtable operatorsTable;

    /**
     * A boolean indicating if proximity is supported.
     */
    protected boolean proximitySupported;

    /**
     * Boolean values indicating whether the source supports ranking.
     */
    protected boolean supportsRanking;

    /**
     * Boolean values indicating whether the source supports filters.
     */
    protected boolean supportsFilters;

	/**
	 * A two element double array indicating the lower and
	 * upper bounds for relevance scores returned for each
	 * search hit.
	 */
    protected float[] scoreRange;

	/**
	 * A string array giving the name of the ranking algorithm.
	 */
    protected String rankingAlgorithm;

    /**
     * A string array giving the identifiers of the tokenizers
     * used by the source.  Each element of the array is formatted
     * as follows:
     * <language-list> tokenizer-id
     * where <language-list> is an optional list of languages
     * for which the respective tokenizer is used and
     * tokenizer-id is the (single token) name of the tokenizer
     * An example array element is:
     * "en-US sp-MX Acme-1"  meaning the the tokenizer Acme-1
     * is used for both US english and Mexican spanish.
     */
    protected String[] tokenizerIDList;

	/**
	 * A string indicating the URL from which to get the query
	 * results for a sample document collection.
	 */
    protected String sampleDataBaseResultsURL;

	/**
	 * A boolean value indicating whether stop words can be
	 * turned off at this source
	 */
    protected boolean turnOffStopWords;

	/**
	 * A String indicating the default Meta attribute set
	 */
    protected String defaultMetaAttributeSet;

	// settings for mbasic-1 attributes

	/**
	 * A String array where each element indicates,
	 * in RFC1877 format, a language present at the source.
	 */
    protected String[] sourceLanguage;

	/**
	 * A String giving the name of the source.
	 */
    protected String sourceName;

	/**
	 * A String value giving the URL where the source should
	 * be queried
	 */
    protected String linkage;

	/**
	 * A String value giving the URL of the source content summary.
	 */
    protected String contentSummaryLinkage;

	/**
	 * A String value giving a free text abstract of this source.
	 */
    protected String sourceAbstract;

	/**
	 * A String value giving a free text description of the contraints
	 * or legal prerequisites for accessing the source.
	 */
    protected String accessConstraints;

	/**
	 * A String value indicating the acministrator of the source
	 */
    protected String administrator;

    /**
     * settings for meta attributes in other attribute sets.  This
     * should be a three dimensional array where each element in
     * a row is as follows:
     * 0 - name of the attribute set
     * 1 - name of the attribute
     * 2 - value of the attribute set
     * For example, a row might be initialized to:
     * {"foo-1", "emporer", "Caeser"}
     */
    protected String otherMetaAttributes[][][];

/***********************************************************************/
// sdp - moved to Attrset class

    /**
     * Return a String that is a translation from STARTS field
     * to Source specific field.
     * @param attrSet  Attribute set of the source field to be translated.
     * @param field    Field name of the source field to be translated.
     * @return the string translation of the input attribute set/field
     * pair to the source specific field.
     */

    //public  abstract String TranslateField(String attrSet, String field);
/*************************************************************************/
    /**
     * Return a String that is a translation from STARTS modifier
     * to source modifier.
     * @param attrAndMod   Combined attribute set and modifier (space
     * separated) to be translated.
     * @return The source-specific string translation of the input
     * STARTS attribute set/modifier pair.
     */
    public abstract String TranslateModifier(String attr, String Mod);

    /**
     * Return a String that is a translation from STARTS operator to
     * source operator.
     * @param op   The STARTS operator to be translated
     * @return the source-specific string translation of the STARTS
     * operator.
     */
    public abstract String TranslateOperator(String op);

    /**
     * Return a string array that is the set of stop words at this
     * source.
     * @return the array of stop words for the source.
     */
    public abstract String[] GetStopWords() ;

    /**
     * Return true or false specifying whether the input string is a
     * stop word in the source.
     * @param string   word to test whether it is a stop word.
     * @return true or false depending on whether the input source
     * word is a stop word for this source.
     */
    public abstract boolean StopWord_p(String string);

    /**
     * Return an ISO-1807 representation of when the source was modified
     * @return the ISO representation of when this source was last
     * modified.
     */
    public abstract String GetDateModified();

    /**
     * Return an ISO-1807 representation of when metadata for this
     * source expires.
     * @return the ISO formatted date of when the returned metadata for
     * this source expires.
     */
    public abstract String GetDateExpires();

    /**
     * Return a SOIF object that contains the Content Summary header
     * information for this source. (Note: the content summary also
     * includes field/term information that is retrieved via a method
     * call to ContentSummaryFieldsDump)
     * @exception SourceException if there is an error at the source
     * deriving this header information.
     */
    public abstract SOIFObject ContentSummaryHeaderToSOIF() throws SourceException;

    /**
     * Send to an output stream the content summary description
     * for this source.
     * @param out the output stream to which to send the content
     * summary description.
     * @exception SourceException if an error occurs deriving the
     * content summary information from the source
     */
    public abstract void ContentSummaryFieldsDump(PrintStream out)
        throws SourceException;


    // Translators for STARTS filter and ranking ExpNodes to native expressions

    /**
     * Return the translated version of a STARTS Term node to the
     * source equivalent.
     * @param term term node to translated.
     * @param fieldXlate   tranlsated version of the field that descends
     * from this term.
     * @param modListXlate translated version of the modifier list that
     * descends from this term.
     * @param lStringXlate translated version of the LString that
     * descends from this term.
     * @return the source equivalent filter subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public  abstract ExpNode TermToFilter(
        Term term, ExpNode fieldXlate, ExpNode modListXlate,
        ExpNode lStringXlset)
        throws SourceQueryException;

    /**
     * Return the translated version of a STARTS term node within a
     * ranking expresssion.
     * @param term the STARTS term node.
     * @param  fieldXlate  translated version of the field node that
     * descends from this term.
     * @param modListXlate  translated version fo the modifier list node
     * that descends from this term.
     * @param lStringXlate translated version of the lString node that
     * descends from this term.
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public  abstract ExpNode TermToRanking(Term term, ExpNode fieldXlate,
        ExpNode modListXlate,
        ExpNode lStringXlset)
        throws SourceQueryException;


    /**
     * Translate from a STARTS BoolOp to the source equivalent.
     * @param bo   STARTS BoolOop to translate
     * @param leftChild    translated version of the left child of
     * this boolop
     * @param rightChild   trnslated version of the right child of this
     * boolop
     * @return the source equivalent filter subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode BoolOpToFilter(
        BoolOp bo, ExpNode leftChild, ExpNode rightChild)
        throws SourceQueryException;

    /**
     * Return the translated version of a STARTS BoolOp node within
     * a ranking expression.
     * @param field    the STARTS BoolOp node
     * @param leftChild    translated left child of this boolop
     * @param rightChild   translated right child of this boolop
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode BoolOpToRanking(
        BoolOp bo, ExpNode leftChild, ExpNode rightChild)
        throws SourceQueryException;

    /**
     * Translate from a STARTS field tree node to a source one.
     * @param field    field node to translate
     * @return the source equivalent filter subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode FieldToFilter(Field field)
        throws SourceQueryException;

    /**
     * Return the translated version of a STARTS field node within
     * a ranking expression.
     * @param field    the STARTS field node
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode FieldToRanking(Field field)
        throws SourceQueryException;

    /**
     * Translate from a STARTS LString tree node to a source one.
     * @param string   LString to translate
     * @return the source equivalent filter subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode LStringToFilter(LString string)
        throws SourceQueryException;

    /**
     * Return the translated version of a STARTS LString node within
     * a ranking expression.
     * @param string    the STARTS LString node
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode LStringToRanking(LString string)
        throws SourceQueryException;

    /**
     * Translate from a STARTS modifier list tree node to a source one.  In
     * @param ml   modifier list to translate
     * @return the source equivalent filter subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode ModifierListToFilter(ModifierList ml)
        throws SourceQueryException;

    /**
     * Return the translated version of a STARTS modifier list node within
     * a ranking expression.
     * @param ml    the STARTS modifier list node
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode ModifierListToRanking(ModifierList ml)
        throws SourceQueryException;

    /**
     * Translate from a STARTS ProxOp to the source equivalent.
     * @param bo   STARTS ProxOp to translate
     * @param leftChild    translated version of the left child of
     * this ProxOp
     * @param rightChild   trnslated version of the right child of this
     * ProxOp
     * @return the source equivalent filter subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode ProxOpToFilter(ProxOp po,
        ExpNode leftTerm, ExpNode rightTerm)
        throws SourceQueryException;

    /**
     * Return the translated version of a STARTS ProxOp node within
     * a ranking expression.
     * @param field    the STARTS BoolOp node
     * @param leftChild    translated left child of this Prox
     * @param rightChild   translated right child of this ProxOp
     */
    public abstract ExpNode ProxOpToRanking(
        ProxOp po, ExpNode leftChild, ExpNode rightChild)
        throws SourceQueryException;

    /**
     * Return the translated version of a starts Ranking list node
     * within a ranking expression.
     * @param wr   STARTS  ranking list node to be translated.
     * @param child. Vector of translated childrenof this ranking
     * list node.
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode RankingListToRanking(RankingList rl, Vector children)
        throws SourceQueryException;

    /**
     * Return the translated version of a starts WeightedRanking node
     * within a ranking expression.
     * @param wr   STARTS weighted ranking node to be translated.
     * @param child    translated version of the single childed node
     * of this weighted ranking node.
     * @return the source equivalent ranking subtree.
     * @exception SourceQueryException if an error occurs in the
     * translation.
     */
    public abstract ExpNode WeightedRankingToRanking(WeightedRanking wr,
        ExpNode child)
        throws SourceQueryException;

    /**
     * Submit a search to the source search engine.  Return the number
     * of "hits" that the search engine returned.
     * @param query    the query on which to base the search.
     * @exception if there is an error performing the search at the
     * source.
     */
    public abstract int DoSourceSearch(Query query) throws SourceException;

    /**
     * Predicate to determine whether this source supports ranking
     * @return true or false depending on whether this source
     * supports ranking expressions
     */
    public boolean Ranking_p() {
        return supportsRanking;
    }

    /**
     * Predicate to determine whether this source supports filter
     * expressions.
     * @return true or false depending on if this source supports
     * filter expressions.
     */
    public boolean Filters_p() {
        return supportsFilters;
    }


/**************************************************************************/
// sdp - commented out.  Implemented this method in attribute set classes.

/**
	public  boolean FieldSupported_p(String attributeSet, String field) {
	    if (!fieldsTable.containsKey(attributeSet + " " + field)) {
	        return(false);
	    }
        else {
            return(true);
        }
	}
**/
/**************************************************************************/

    /**
     * predicate to determine whether a specific modifier/attribute
     * set pair is supported.
     * @param attibuteSet attribute set to test.
     * @param modifier modifier to test.
     * @return true or false depending on whether the selected
     * attribute set/modifier pair is supported for this source.
     */
	public  boolean ModifierSupported_p(String attributeSet, String modifier) {
	    if (!modifiersTable.containsKey(attributeSet + " " + modifier)) {
	        return(false);
	    }
	    else {
            return(true);
	    }
	}

    /**
     * predicate to determine whether a specific language is
     * supported for a field
     * @param attributeSet attribute set of the field to test
     * @param field    field to test
     * @param language language to test.
     * @return true or false depending on whether the input
     * language is supported for the input attribute set/field
     * pair.
     */
	public  boolean LanguageSupportedForField_p(
	    String attributeSet, String field, String language) {

	    if (!fieldsTable.containsKey(attributeSet + " " + field)) {
	        return(false);
	    }
	    else {
    	    String value = (String)(fieldsTable.get(attributeSet + " " + field));
	        StringTokenizer tokens = new StringTokenizer(value);
	        String languageList;
	        while (tokens.hasMoreTokens()) {
	            languageList = tokens.nextToken();
	            if (languageList.equalsIgnoreCase(language)) {
	                return(true);
	            }
	        }
	    return(false);
	    }
	}

    /**
     * predicate to determine whether a specific language is
     * supported for a modifier
     * @param attributeSet attribute set of the modifier
     * @param modifier modifier to test
     * @param language language to test
     * @return true or false depending on whether the input language
     * is supported for the input attribute set/modifier pair.
     */
	public  boolean LanguageSupportedForModifier_p
	    (String attributeSet, String modifier, String language) {

	    if (!modifiersTable.containsKey(attributeSet + " " + modifier)) {
	        return(false);
	    }
	    else {
    	    String value = (String)(modifiersTable.get(attributeSet + " " + modifier));
	        StringTokenizer tokens = new StringTokenizer(value);
	        String languageList;
	        while (tokens.hasMoreTokens()) {
	            languageList = tokens.nextToken();
	            if (languageList.equalsIgnoreCase(language)) {
	                return(true);
	            }
	        }
	    return(false);
	    }
	}

	/**
	 * predicate to determine whether a specific operator is
	 * supported
	 * @param operator operator to test.
	 * @return true or false depending on whether the input operator
	 * is supported for this source.
	 */
	public  boolean OperatorSupported_p(String operator) {
	    return(operatorsTable.contains(operator));
	}

    /**
     * return the ID of this source.
     * @return the ID of this source.
     */
	public String GetSourceID() {
	    return sourceID;
	}

    /**
     * Return the metadata URL of this source.
     * @return the URL to query for metadata about this source.
     */
	public String GetMetadataURL() {
	    return metadataURL;
	}

    /**
     * Return the metadatasynatx of this source.
     * @return the syntax in which metadata for this source is
     * represented (e.g., basic-1).
     */
    public String GetMetadataSyntax() {
        return metadataSyntax;
    }

    /**
     * Return a SOIF object that contains the meta attributes of this
     * source.
     * @return SOIF object containing meta attributes of this source.
     */
    public SOIFObject SourceToSOIF() {
        SOIFObject outSOIF = new SOIFObject("SMetaAttributes");
        StringBuffer buf;

        outSOIF.SetAttributeValue("Version",
            ServerConfiguration.STARTSVersion);
        outSOIF.SetAttributeValue("SourceID",
            sourceID);

/************************************************************************************/
// Added new SOIF attribute to indicate what attribute sets the source supports
// (Sandy Payette 7/28/97)

        // attribute sets supported
        buf = new StringBuffer();
        for (Enumeration e = attrSetsTable.keys();
            e.hasMoreElements();) {

            String key = (String) e.nextElement();
            buf.append(key + " ");

        }
        outSOIF.SetAttributeValue("AttributeSetsSupported", buf.toString());
/***********************************************************************************/

        // fields supported
        buf = new StringBuffer();
        for (Enumeration e = fieldsTable.keys();
            e.hasMoreElements();) {

            String key = (String) e.nextElement();
            String lang = (String) fieldsTable.get(key);
            if (lang.length() == 0) {
                buf.append("[" + key + "]");
            }
            else {
                buf.append("(" + "[" + key + "]" + lang + ")");
            }
        }
        outSOIF.SetAttributeValue("FieldsSupported", buf.toString());

        // Modifiers supported
        buf = new StringBuffer();
        for (Enumeration e = modifiersTable.keys();
            e.hasMoreElements();) {

            String key = (String) e.nextElement();
            String lang = (String) modifiersTable.get(key);
            if (lang.length() == 0) {
                buf.append("{" + key + "}");
            }
            else {
                buf.append("(" + "{" + key + "}" + lang + ")");
            }
        }
        outSOIF.SetAttributeValue("ModifiersSupported", buf.toString());

        // Query parts supported
        buf = new StringBuffer();
        if (supportsRanking) {
            buf.append("R");
        }
        if (supportsFilters) {
            buf.append("F");
        }
        outSOIF.SetAttributeValue("QueryPartsSupported", buf.toString());

        if (scoreRange == null) {
            outSOIF.SetAttributeValue(
                "ScoreRange", "-infinity +infinity");
        }
        else {
            outSOIF.SetAttributeValue(
                "ScoreRange",
                String.valueOf(scoreRange[0]) +
                " " +
                String.valueOf(scoreRange[1]));
        }

        outSOIF.SetAttributeValue("RankingAlgorithm",
            rankingAlgorithm);

        // TokenizerIDList
        buf = new StringBuffer();
        for (int i = 0;
            i < tokenizerIDList.length;
            i++) {

            buf.append("(" + tokenizerIDList[i] + ")");
        }

        outSOIF.SetAttributeValue("SampleDatabaseResults",
            sampleDataBaseResultsURL);

        // Stop word list
        buf = new StringBuffer();
        String stopWords[] = GetStopWords();
        if (stopWords == null) {
            outSOIF.SetAttributeValue("StopWordList", "");
        }
        else {
            for (int i = 0; i < stopWords.length; i++) {
                buf.append(stopWords[i] + " ");
            }
            outSOIF.SetAttributeValue("StopWordList", buf.toString());
        }

        if (turnOffStopWords) {
            outSOIF.SetAttributeValue("TurnOffStopWords", "true");
        }
        else {
            outSOIF.SetAttributeValue("TurnOffStopWords", "false");
        }

        outSOIF.SetAttributeValue("DefaultMetaAttributeSet",
            defaultMetaAttributeSet);

        // source languages
        buf = new StringBuffer();
        for (int i = 0; i < sourceLanguage.length; i++) {
            buf.append(sourceLanguage[i] + " ");
        }
        outSOIF.SetAttributeValue("source-language", buf.toString());

        outSOIF.SetAttributeValue("source-name", sourceName);
        outSOIF.SetAttributeValue("linkage", linkage);
        outSOIF.SetAttributeValue("content-summary-linkage",
            contentSummaryLinkage);
        outSOIF.SetAttributeValue("date-changed", GetDateModified());
        outSOIF.SetAttributeValue("date-expires", GetDateExpires());
        outSOIF.SetAttributeValue("abstract", sourceAbstract);
        outSOIF.SetAttributeValue("access-constraints", accessConstraints);
        outSOIF.SetAttributeValue("contact", administrator);

        return outSOIF;

    }


}
