/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package results;

import java.lang.*;
import java.util.*;
import soif.*;
import STARTSConfiguration.*;
import query.*;

/**
 * The documents in the results set from a STARTS query
 */
public abstract class Document {

    /**
     * query that produced this document.
     */
    protected Query query;

    /**
     * unnormalized score of this document.
     */
    protected float rawScore;

    /**
     * sources in which the document appears
     */
    protected String sources[];

    /**
     * size of the document in bytes.
     */
    protected long documentSize;

    /**
     * number of tokens in the document.
     */
    protected long documentCount;

    /**
     * array of term statistics (for each ranking term) in the
     * document.
     */
    protected TermStatistics termStats[];

    /**
     * Dummy constructor so we can instantiate sub-classes with
     * newInstance
     */
    public Document() {}

    /**
     * construct a new document and add it to the result set.
     * @param query    the query that for which this document is a
     * "hit".
     */
    public Document(Query query) {
        query.results.AddNewDocument(this);
        this.query = query;
    }

    /**
     * construct a new document and set its attributes.
     * @param rawScore raw score of this document.
     * @param documentSize size of the document in bytes.
     * @param numTokens    number of tokens in the document.
     * @param termStats    statistis on ranking terms in this document.
     */
    public Document(
        float rawScore, String sources[], long documentSize, long numTokens,
        TermStatistics termStats[]) {

        this.rawScore = rawScore;
        this.documentSize = documentSize;
        this.sources = new String[sources.length];
        System.arraycopy(sources, 0,
            this.sources, 0, sources.length);
        this.termStats = new TermStatistics[termStats.length];
        System.arraycopy(termStats, 0, this.termStats, 0, termStats.length);
    }

    /**
     * Return the float raw score of a document
     * @return the float raw score value.
     */
    public float GetScore() {
        return rawScore;
    }

    /**
     * Produce a SOIF object containing the information for the
     * document.
     * @return the SOIFObject representation of the document data (
     * document is a member of a query result set)
     * @exception SourceException if an error occurs deriving the
     * document data from the source.
     */
    public SOIFObject DocumentToSOIF() throws resource.SourceException{
        SOIFObject outSOIF = new SOIFObject("SQRDocument");

        outSOIF.SetAttributeValue("Version",
           ServerConfiguration.STARTSVersion);
        outSOIF.SetAttributeValue("RawScore",
           String.valueOf(rawScore));

        // Sources
        StringBuffer sourceBuf = new StringBuffer();
        for (int i = 0; i < sources.length; i++) {
            sourceBuf.append(sources[i] + " ");
        }
        outSOIF.SetAttributeValue("Sources", sourceBuf.toString());

/************************************************************************************/
/*  Sandy Payette 7/24/97
    Made changes to have the attribute-set name part of the outSOIF.
    Previously, just the field name was returned with the SOIF value which was
    not accomodating of multiple attribute sets.  Now, the attribute set name will
    be explicitly reported in the SQRDocument, for example:
    [basic-1 author]{12}: Lagoze, Carl
*/

        // Field values
        for (int i = 0; i < query.answerFields.length; i++) {
            String fieldName = (
                                "[" +
                                (query.answerFields[i].GetAttributeSet()) +
                                " " +
                                (query.answerFields[i].GetFieldName()) +
                                "]"
                                );

            String fieldValue = GetFieldValue(query.answerFields[i], Integer.MAX_VALUE);

            if (fieldValue != null) {
                outSOIF.SetAttributeValue(fieldName, fieldValue);
                //outSOIF.SetAttributeValue(query.answerFields[i].GetFieldName(), fieldValue);
            }
        }
/************************************************************************************/

        // Term Statistics
        if (termStats != null) {
            StringBuffer statBuf = new StringBuffer();
            for (int i = 0; i < termStats.length; i++) {
                statBuf.append(termStats[i].GenerateStatsString() + " ");
            }
            outSOIF.SetAttributeValue("TermStats", statBuf.toString());
        }

        outSOIF.SetAttributeValue("DocSize", String.valueOf(documentSize));
        outSOIF.SetAttributeValue("DocCount", String.valueOf(documentCount));

        return outSOIF;
    }

    /**
     * Return the string value of a given field (e.g., author, title)
     * in the document
     * @param  f   the field to return a value for.
     * @param  maxLen  the maximum length of the String to be returned
     * (may be infinity meaning that there is no limit on length
     * of string to be returned).
     * @return the value of the selected field in the source document.
     */
    public abstract String GetFieldValue(Field f, int maxLen)
        throws resource.SourceException;

}
