/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package results;

import java.lang.*;
import java.util.*;

import query.*;
import soif.*;
import STARTSConfiguration.*;

/**
 * The results set produced from a STARTS query
 */
public class ResultSet {

    /**
     * The query context of the result set.
     */
    Query query;

    /**
     * Vector of Documents that result from processingthe query.
     */
    Vector documents;

    /**
     * Sorted version of above. (Sorted according to the spec. in the
     * query).
     */
    Vector sortedDocuments;

    /**
     * Create a new results set
     * @param query    query context of the result set
     * @param numDocuments     number of documents in the results set
     */
    public ResultSet(Query query, int numDocuments) {

        this.query = query;
        query.results = this;
        if (numDocuments > 0) {
            documents = new Vector(numDocuments);
        }
    }

    /**
     * Add a new document to the result set
     */
    public void AddNewDocument(Document document) {
        documents.addElement(document);
    }

    /**
     * Return the vector of documents in the result set
     * @return the Vector of documents in the result set.
     */
    public Vector GetDocuments() {
        return(documents);
    }

    /**
     * Return the vector of sorted documents in the result set
     * @return a sorted vector of the documents in the result set.
     */
    public Vector GetSortedDocuments() {
        return(sortedDocuments);
    }

    /**
     * Return a SOIFObject that is the SOIF representation of this
     * result set.
     * @return a SOIF object representing the data in the reslt set.
     */
    public SOIFObject ResultSetToSOIF() {
        SOIFObject soif = new SOIFObject("SQResults");
        String ranking, filter;

        ranking = query.processedRanking != null ? query.processedRanking.Unparse() : "";
        filter = query.processedFilter != null ? query.processedFilter.Unparse() : "";

        soif.SetAttributeValue("Version", ServerConfiguration.STARTSVersion);
        soif.SetAttributeValue("Sources", query.source.GetSourceID());
        soif.SetAttributeValue("ActualFilterExpression", filter);
        soif.SetAttributeValue("ActualRankingExpression", ranking);
        if (documents == null) {
            soif.SetAttributeValue("NumDocSOIFs", "0");
        }
        else {
            soif.SetAttributeValue("NumDocSOIFs", String.valueOf(documents.size()));
        }

        return soif;
    }

    /**
     * Remove from the document list those documents whose score is
     * less than the minimum score for the query.
     * @return the number of documents remaining after the culling
     * process.
     */
    public int MinScoreCheck() {
        int numDocs = 0;

        for (int i = 0; i < documents.size(); i++) {
            Document doc = (Document)documents.elementAt(i);
            if (doc.GetScore() < query.minimumDocScore) {
                documents.removeElement(doc);
                i--;
            }
            else {
                numDocs++;
            }
        }

        return numDocs;
    }

    /**
     * Trim the result set down to the size specified for the query
     * @return the number of documents remaining after the trimming
     * process.
     */
    public int TrimResults() {
        if (documents.size() > query.maximumNumDocuments) {
            documents.setSize(query.maximumNumDocuments);
        }
        return documents.size();
    }

    /**
     * perform a multi-key sort based on the sort specification (SortFields and
     * SortDescending) in the query context for this resultset.
     * @exception SourceException if an error occurs at the source
     * deriving the information necessary for the sort.
     */
    public void SortResults(Field sortFields[], boolean sortDescending[])
        throws resource.SourceException {
        Vector docs = new Vector(documents.size());
        Hashtable sortTree;
        sortedDocuments = new Vector(docs.size());

        // avoid the trouble if you can.
        if (documents.size() == 1) {
            sortedDocuments.addElement(documents.elementAt(0));
            return;
        }

        // Initialize a unsorted vector, each element of which is the position of
        // a document in the ResultSet.documents vector.  This is the inital seed
        // for the sort
        for (int i = 0; i < documents.size(); i++) {
            docs.addElement(new Integer(i));
        }

        // Create a tree (using hashtables) where each level of the tree maps
        // to the next sort field and the nodes at that level (the entries in the
        // hash table) are the values in the result set for that field. For example,
        // assume that the sort is (author, title).  In this case the top level
        // hash table will have keys that match the authors in the result set.
        // The values of these top level keys will be hashtables.  Each of these
        // hashtables will have keys that match the titles for the respective
        // author.  Finally at the leaves (bottom level hash tables), the value is
        // an Integer vector where the value of each Integer object is the entry
        // in the ResultSet.Documents array that has the values of the sort fields
        // defined by its parents.
        sortTree = HashResults(docs, 0);

        // Now use the tree to fill sortedDocuments with the sort order of the
        // Documents vector.  sortedDOcuments is a vector of Integers, where the
        // value of the Integer object maps to a position in the Document vector of
        // the next sorted document.
        SortTables(sortTree, sortedDocuments, 0);
    }

    /**
     * Recursively create a sort tree.  Each level of the tree represents
     * a sort field in the sortFields vector.  Each node at a level is a
     * value for that respective field.
     * @param  docs    Vector of documents to be used to generate nodes at this
     * level of the sort tree.
     * @param sortKeyNum   The position in the sortFields array that we are
     * currently processing.
     * @return a nested hash table that represents the result set arranged
     * according to the sort field specification in the query.
     * @exception if an error occurs at the source derivig the information
     * required for the hashing process
     */
    private Hashtable HashResults  (
        Vector docs, int sortKeyNum) throws resource.SourceException{

        Hashtable newTable = new Hashtable();
        String fieldVal;

        // loop through the document vector.  For each element, map to the
        // document in the documents array that maps to that element.  Get the
        // value of the respective Field for that document and create an entry
        // in the hash table.
        for (int i = 0; i < docs.size(); i++) {
            Integer docNum = (Integer) docs.elementAt(i);
            Document doc = (Document) documents.elementAt(docNum.intValue());

            // score is a special "field"
            String fieldString = query.sortFields[sortKeyNum].GetFieldName();
            if (fieldString.equalsIgnoreCase("score")) {
                fieldVal = String.valueOf(doc.GetScore());
            }
            // otherwise we have to fetch the field from the document
            else {
                fieldVal = doc.GetFieldValue(query.sortFields[sortKeyNum], 20);
            }
            if (!newTable.containsKey(fieldVal)) {
                newTable.put(fieldVal, new Vector(1));
            }
            Vector v = (Vector) newTable.get(fieldVal);
            v.addElement(docNum);
        }

        // recurse down if there are more sort fields to process
        if (++sortKeyNum < query.sortFields.length) {
            for (Enumeration e = newTable.keys(); e.hasMoreElements(); ) {
                String key = (String) e.nextElement();
                Vector v = (Vector) newTable.get(key);
                newTable.put(key, HashResults(v, sortKeyNum));
            }
        }

        return(newTable);
    }

    /**
     * Recursively process the sort tree produced by HashResults.  At each level
     * of the tree sort the set of nodes at that level.  Then (in order)
     * recurse down to the next level of the tree.  When the leaf level is hit,
     * append the vector of document numbers to the end of the sorted Vector.
     * @param sortTable    the sort tree produced from HashResults.
     * @param  sorted  Vector of sorted documents that is built.
     * @param sortKeyNum   level in the sortFields vector currently being
     * processed.
     */
    private void SortTables(Hashtable sortTable, Vector sorted, int sortKeyNum) {
        String keyArray[] = new String[sortTable.size()];
        int i = 0, j;

        // Fill an array with the keys at this tree level.
        for (Enumeration e = sortTable.keys(); e.hasMoreElements(); ) {
            String key = (String) e.nextElement();
            keyArray[i++] = key;
        }

        // bubble sort the keys. in ascending order
        for (i = keyArray.length - 1; i > 0 ; i--) {
            for (j = 0; j < i; j++) {
                if (Compare(
                    keyArray[j], keyArray[j+1], query.sortFields[sortKeyNum]) > 0) {
                    String temp = keyArray[j];
                    keyArray[j] = keyArray[j+1];
                    keyArray[j+1] = temp;
                }
            }
        }

        // If we are not at the leaf level of the tree recurse, The order of the
        // recursion is sorted (either ascending or descending) depending on the
        // boolean value in sortDescending for this level.
        if (sortKeyNum < query.sortFields.length - 1) {
            if (query.sortDescending[sortKeyNum]) {
                sortKeyNum++;   // down to the next key
                for ( i = keyArray.length - 1; i >= 0; i--) {
                    String key = keyArray[i];
                    Hashtable subTable = (Hashtable) sortTable.get(key);
                    SortTables(subTable, sorted, sortKeyNum);
                }
            }
            else {
                sortKeyNum++;   // down to the next key
                for ( i = 0; i < keyArray.length; i++) {
                    String key = keyArray[i];
                    Hashtable subTable = (Hashtable) sortTable.get(key);
                    SortTables(subTable, sorted, sortKeyNum);
                }
            }
        }

        // Otherwise we are at the leaf level.  Step through the leaf keys
        // (either ascending or descending) and append to the sorted vector the documents
        // corresponding to the integer values in the leaf vector.
        else {
            if (query.sortDescending[sortKeyNum]) {
                for ( i = keyArray.length - 1; i >= 0; i--) {
                    String key = keyArray[i];
                    Vector docNums = (Vector) sortTable.get(key);
                    for ( j = 0; j < docNums.size(); j++) {
                        int dn = ((Integer) docNums.elementAt(j)).intValue();
                        sorted.addElement(documents.elementAt(dn));
                    }
                }
            }
            else {
                for ( i = 0; i < keyArray.length; i++) {
                    String key = keyArray[i];
                    Vector docNums = (Vector) sortTable.get(key);
                    for ( j = 0; j < docNums.size(); j++) {
                        int dn = ((Integer) docNums.elementAt(j)).intValue();
                        sorted.addElement(documents.elementAt(dn));

                    }
                }
            }
        }

    }

    /**
     * Compare two values and return -1, 0, 1 for less than, equal to, or
     * greater than.
     * @param val1 first value to compare
     * @param  val2 second value to compare
     * @param sortField    the field type of the values.
     * @return -1 if val1 is less than val 2, 0 if val1 is equal to val2,
     * or 1 if val1 is greater than val2
     */
    private int Compare(String val1, String val2, Field sortField) {
        String fieldName = sortField.GetFieldName();

        // score is special numerical comparison.
        if (fieldName.equalsIgnoreCase("score")) {
            float fv1 = Float.valueOf(val1).floatValue();
            float fv2 = Float.valueOf(val2).floatValue();
            if (fv1 < fv2) {
                return -1;
            }
            else if (fv1 > fv2) {
                return 1;
            }
            else {
                return 0;
            }
        }

        // otherwise string comparison
        else {
            return(val1.compareTo(val2));
        }
    }

}
