/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package soif;

import java.io.*;
import java.net.*;
import java.util.*;

import regexp.*;

/**
 * A set of data encapsulated in a SOIF
 */
public class SOIFObject
{

    /**
     * The template type of the SOIF
     */
	private String templateType;

    /**
     * The URL in the SOIF
     */
	private URL URL;

    /**
     * Property list for the SOIF
     */
	private Hashtable attributeValues;

    /**
     * Allows us to pull things out of the property list in the order they
     * were inserted.
     */
	private Vector keyList;

    /**
     * Common delimitter that separates SOIF attributes and values
     */
	static final String soifDelimitter = ":	";


    /**
        Build a new SOIF object
    */
	public SOIFObject() {
		attributeValues = new Hashtable();
		keyList = new Vector();
	}

  /*
   * Build a new SOIF object with a set template type
   * @param templateType the template type of the SOIF
   */
	public SOIFObject(String templateType) {
	    this();
	    this.templateType = templateType;
	}

  /**
   * Build a new SOIF object from an input stream
   * @param soifInput input stream from which the SOIF should be built.
   */
	public SOIFObject(InputStream soifInput) throws SOIFException {
		this();

		String soifText;
		byte b[];

		try {
		    DataInputStream dStream = new DataInputStream(soifInput);
		    soifText = dStream.readLine();
		}
		catch (IOException e) {
			throw new SOIFException("error reading SOIF input");
		}

ParseSOIFBuffer(soifText);
	}

  /**
   * Return a string array of the attribute names in this SOIF
   * @return An array of String that are the attribute (property)
   * names for the SOIF instance
   */
	public String[] GetAttributeNames() {
		Enumeration elements = attributeValues.elements();
		String attributeNames[] = new String[attributeValues.size()];
		int i = 0;

		while (elements.hasMoreElements()) {
			attributeNames[i++] = (String) elements.nextElement();
		}
		return(attributeNames);
	}

  /**
   * Return a count of the number of attributes in this SOIF
   */
	public int GetNumAttributes() {
		return(attributeValues.size());
	}

  /**
   * Return the value of a specified attribute in this SOIF
   * @param attribute name of the attribute to get the value of.
   */
	public String GetAttributeValue(String attribute) {
		if (!attributeValues.containsKey(attribute)) {
			return null;
		}
		return((String) attributeValues.get(attribute));
	}

  /**
   * Set an attribute/value pair for this SOIF
   * @param attribute name of the attribute to set.
   * @param value value to set the attribute to.
   */
	public void SetAttributeValue(String attribute, String value) {
		attributeValues.put(attribute, value);
		keyList.addElement(attribute);

	}

  /**
   * Return the template type of this SOIF object.
   */
	public String GetTemplateType() {
		return(templateType);
	}

  /**
   * Return the URL of this SOIF object
   */
	public URL GetURL() {
		return(URL);
	}

  /**
   * Return a StringBuffer which is a marshalled version of this SOIF
   * @param close A boolean value determining whether the soif should be closed
   * (I.E. whether the closing brace should be appended to the soif).
   */
	public StringBuffer SOIFToBuffer(boolean close) {
		StringBuffer output = new StringBuffer();
		Enumeration keys = attributeValues.keys();
		String attribute, value;

		output.append("@" + GetTemplateType() + "{");
		for (int i = 0; i < keyList.size(); i++) {
   			value = (String) attributeValues.get(keyList.elementAt(i));
            output.append(UnparseSOIFElement((String) keyList.elementAt(i), value));
		}
        if (close) {
    		output.append(SOIFClose());
    	}

		return(output);
	}

    /**
     * Turn a attribute/value pair into a strintg.
     * @param attribute The attribute
     * @param value The attribute's value.
     */
    public static String UnparseSOIFElement(String attribute, String value) {
        StringBuffer output = new StringBuffer();

        output.append(attribute);
        output.append("{" + String.valueOf(value.length()) + "}");
		output.append(soifDelimitter + value);
		output.append("\n");

		return output.toString();
	}

	/**
	 * Return the SOIF close symbol (the brace)
	 *
	 */
	public String SOIFClose() {
	    return("}");
	}

  /**
   * Parse a marshalled SOIF object extracting all its attribute/value
   * pairs
   * @param soifString the string to parse.
   */
	private void ParseSOIFBuffer(String soifString) throws SOIFException {
		String soifBody, couldBeURL;
		Regexp reg;
		Result result;
		int pos;

		soifString.trim();

        // separate the template type and the body (inside {})
		if (!soifString.startsWith("@")) {
			throw new SOIFException("Illegal SOIF file format");
		}
        if ((pos = soifString.indexOf("{")) < 0) {
			throw new SOIFException("Illegal SOIF file format");
		}
		templateType = soifString.substring(1, pos).trim();
		soifBody = soifString.substring(pos + 1).trim();
		if (!soifBody.endsWith("}")) {
			throw new SOIFException("Illegal SOIF file format");
		}
        soifBody = soifBody.substring(0, soifBody.length() - 1);

        // first token in the body may be a url
        StringTokenizer st = new StringTokenizer(soifBody);
        try {
            couldBeURL = st.nextToken();
            URL = new URL(couldBeURL);
			soifBody = soifBody.substring(couldBeURL.length()).trim();
		}
		catch (MalformedURLException e) {
			// we don't care if there is no URL, It is optional and if there is none, then the
			// body is just attribute/value pairs.
		}
		catch (NoSuchElementException e1) {
			throw new SOIFException("Illegal SOIF file format");
		}


		ParseSOIFBody(soifBody);
	}

  /**
   * Parse the body of the SOIF extracting the attribute/value pairs
   * @param body the body to parse
   */
	private void ParseSOIFBody(String body) throws SOIFException {
		int curPos = 0, bodyLength = body.length();
		attributeValues = new Hashtable();
		String attribute, attributeName, value;
		int valueLength;

		// pattern for an attribute (alpha-num string followed by count - e.g. foo{10})
		Regexp reg = Regexp.compile("\\s*(\\S+)\\{(\\d+)\\}:\\t");
		Result result;

		while (curPos < body.length()) {
			if ((result = reg.match(body, curPos)) == null) {
				throw new SOIFException("Illegal SOIF file format - can't find attribute name");
			}

			curPos += result.getMatch(0).length();
			attributeName = result.getMatch(1);
			valueLength = (new Integer(result.getMatch(2))).intValue();
			try {
				value = body.substring(curPos, curPos + valueLength);
			}
			catch (StringIndexOutOfBoundsException e) {
				throw new SOIFException("Illegal SOIF file format - value too short");
			}

			SetAttributeValue(attributeName, value);
			curPos += valueLength;
		}

	}


}

