/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.io.*;
import STARTSConfiguration.*;

/**
 * A WAIS dictionary file, containing all the terms and information
 * on those terms in the source.  Refer to Section 4.3.6 of the
 * freeWAIS documentation for a description of the structure of
 * dictionary files.
 */
public class WAISDictionary {

    // constants for manipulating dictionary files

    /**
     * The number of entries in file block
     */
    static int entriesInBlock = 1000;

    /**
     * Bytes in an entry in the dictinary file
     */
    static int entrySize = 29;

    /**
     * Number of bytes for the term in the entry
     */
    static int bytesForTerm = 21;

    /**
     * Number of bytes for the pointer to the inverted index file
     */
    static int bytesForPointer = 4;

    /**
     * Number of bytes for the number of documents in which an
     * entry appears.
     */
    static int bytesForDocsNum = 4;

    /**
     * Offset from beginning of file for the entry giving the number
     * of directories in the dictionary
     */
    static int numDirectoriesOffset = 2;

    /**
     * Number of bytes which contain the number of directories
     */
    static int bytesForNumDirectories = 2;

    /**
     * Offset from the beginning of the dictionary file of the first
     * directory.
     */
    static int firstDirectoryOffset = 4;

    /**
     * The file object for the dictionary file.
     */
    RandomAccessFile dctFile;

    /**
     * Contruct a WAIS dictionary object from a database name and
     * (possibly null) database field name.
     * @param  indexLocation   The pathname of the directory in
     * which index files for the repsecitve source are located.
     * @param  database    The source name of the database
     * @param field    The field name of the dictionary (may be
     * null).
     */
    public WAISDictionary(String indexLocation, String database,
        String field) throws resource.SourceException {

        String fileName = database;

        // special case any which means just use the general dictionary
        if (field.equalsIgnoreCase("any")) {
            field = null;
        }

        if (field != null) {
            fileName = fileName + "_field_" + field;
        }
        fileName = fileName + ".dct";

        try {
            dctFile = new RandomAccessFile
                (new File(indexLocation, fileName), "r");
        }
        catch (IOException e) {
            throw new resource.SourceException(e.getMessage());
        }
    }

    public void Close() {
        try {
            dctFile.close();
        }
        catch (IOException e) {}
    }

    /**
     * Return the next dictionary term at the current file position.
     * Note that the client of this method MUST NOT make any other
     * method calls on a current WAISDictionary object in between
     * calls to this method (others may change the file pointer)
     */
    public WAISDictionaryTerm GetNextTerm() throws resource.SourceException {

        int numDirectories;
        byte buf[];

        try {

            if (dctFile.getFilePointer()  < dctFile.length()) {
                // special case for the first call, skip past the directory
                // blocks
                if (dctFile.getFilePointer() == 0) {
                    dctFile.seek(numDirectoriesOffset);
                    buf = new byte[bytesForNumDirectories];
                    dctFile.read(buf, 0, bytesForNumDirectories);
                    numDirectories = (int) ReadBytes(buf);

                    // now we are at the start of the first directory.  Skip
                    // forward over all of them to the first term.
                    dctFile.skipBytes(numDirectories * entrySize);
                }

                return(TermAtPos(dctFile.getFilePointer()));
            }
            else {
                return(null);
            }
        }
        catch (IOException e) {
            throw new resource.SourceException(e.getMessage());
        }
    }

    /**
     * Return a WAISDictionary term object representing an entry
     * in the WAISDictionary.
     * @param termString   The term to locate in the dictionary.
     */
    public WAISDictionaryTerm FindTerm(String termString)
        throws resource.SourceException {

        WAISDictionaryTerm term = null;
        int numDirectories;
        byte buf[];
        long groupStart;

        try {
            // first search through the header directories in the dictionary
            // file to find the appropriate block group.
            dctFile.seek(numDirectoriesOffset);
            buf = new byte[bytesForNumDirectories];
            dctFile.read(buf, 0, bytesForNumDirectories);
            numDirectories = (int) ReadBytes(buf);

            // Initiate a binary search through the directories
            term =
                PrivateFindTerm(false,
                    termString, firstDirectoryOffset,
                    firstDirectoryOffset + (numDirectories * entrySize));

            // Now we have the start of the block group containing the term
            // Intiate a binary search through the block group.
            groupStart = term.GetInvertedFileLocation();
            term =
                PrivateFindTerm(true,
                    termString, groupStart, groupStart  + (entriesInBlock * entrySize));
        }
        catch (IOException e) {
            throw new resource.SourceException(e.getMessage());
        }


        return term;
    }

    /**
     * Binary search through a set of blocks to locate a term.
     * @param exact    If true then we must find the exact term
     * or return null, If false, it is ok if when we hit the end
     * of recursion, we return the next lexigraphically smaller
     * value (this is for directory searching).
     * @param termString   The term to find.
     * @param  startOff    The offset inthe dictionary file
     * from which to start the search.
     * @param endOff   The offset in the deictionary file at which
     * to end the search.
     */
    private WAISDictionaryTerm PrivateFindTerm(
        boolean exact,
        String termString, long startOff, long endOff)
        throws resource.SourceException, IOException {
        String curTerm;
        byte buf[] = new byte[20];    // maximum 20 bytes in a term
        int i = 0, compare;
        long midPoint = startOff;

        // seek to the midpoint between the offsets
        long numBlocks = (endOff - startOff) / entrySize;

        if (numBlocks > 1) {
            midPoint = (((int) numBlocks / 2) * entrySize) + startOff;
        }
        else {
            midPoint = startOff;
        }
        dctFile.seek(midPoint);

        // loop through block bytes to collect the 0 terminated term
        do {
            buf[i] = dctFile.readByte();
        } while (buf[i++] != 0);

        curTerm = new String(buf, 0, 0, i - 1);

        // watch out if we are a point in the block that is not full (null curTerm)
        if (curTerm.length() > 0) {
            compare = curTerm.compareTo(termString);
        }
        else {
            compare = 1;
        }

        // We've found it!!
        if (compare == 0) {
            return(TermAtPos(midPoint));
        }

        // haven't found it yet.
        else {
            // reached the end of recursion
            if ((endOff - startOff) == entrySize) {

                // If we needed an exact then give up
                if (exact) {
                    return null;
                }

                // however if only an 'next smaller' match required, return it.
                else {
                    // must back up to next smaller one.
                    if (compare > 0) {
                        midPoint -= entrySize;
                    }
                    // gather information
                    return(TermAtPos(midPoint));
                }
            }
            else {

                if (compare < 0) {
                    return(PrivateFindTerm(
                        exact, termString, midPoint + entrySize, endOff));
                }
                else if (compare > 0) {
                    return(PrivateFindTerm(exact, termString, startOff, midPoint));
                }
            }
        }
    return(null);
    }

    /**
     * Return the WAISDictionaryTerm object that represents the term
     * entry at an offset in the dictionary file.
     * @param  offset  offset of the entry to return.
     */
    private WAISDictionaryTerm TermAtPos(long offset)
        throws IOException {
        long pointer;
        int numDocuments;
        byte buf[], charBuf[] = new byte[1];
        StringBuffer term = new StringBuffer(20);

        dctFile.seek(offset);

        // gather information
        do {
            charBuf[0] = dctFile.readByte();
            if (charBuf[0] != 0) {
                term.append(new String(charBuf, 0));
            }
        } while (charBuf[0] != 0);

        dctFile.seek(offset + bytesForTerm);

        buf = new byte[bytesForPointer];
        dctFile.read(buf, 0, bytesForPointer);
        pointer = ReadBytes(buf);

        buf = new byte[bytesForDocsNum];
        dctFile.read(buf, 0, bytesForDocsNum);
        numDocuments = (int) ReadBytes(buf);

        return(new WAISDictionaryTerm(term.toString(), pointer, numDocuments));
    }


    /**
     * Convert the byte buffer to a long value
     * @param buf  buffer of bytes to convert.
     */
    private long ReadBytes(byte buf[]) {
        long answer = 0;
        for (int i = 0; i < buf.length; i++) {

            // This looks like a java bug to me!!!  Even though buf is
            // a byte array, if the high bit is on it fills a high byte
            // with on bits, so something like 0x81 turns into 0xff81.
            // This happens actually at the time of the read.  So, I
            // work around this by zeroing the high byte. (Boy did I
            // waste time on this!!
            answer = (answer << 8) + (buf[i] & 0xff);
        }
        return answer;
    }

}




