/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.io.*;
import java.util.*;

import STARTSConfiguration.*;
import resource.*;

/**
 * A class to hold information on the documents in the WAIS source.  This
 * information is spread in strange ways across the document file and file name
 * file.  This class loads the information into hash tables so we can quickly
 * do things in the inverted index file.  If these in memory tables weren't created
 * each extraction of information from the inverted index file would require a linear
 * search through the document file.  WARNING!! For a very
 * large database with a very large number of documents this method of
 * creating an in memory table of this information wouldn't work.
 */
public class WAISDocuments {

    // Constants for manipulating doc file

    /**
     * Offset in the document file of the first document
     */
    static int firstDocOffset = 2;

    /**
     * Length of a document entry.
     */
    static int docEntryLength = 25;

    /**
     * Offset in an entry of the filename for an entry.
     */
    static int filenameOffsetLength = 3;

    /**
     * Offset in an entry of the document length value
     */
    static int lengthOffset = 14;

    /**
     * Bytes in entry for document length
     */
    static int lengthLen = 4;

    /**
     * The WAIS document table.  See section 4.3.4 of the freeWAIS
     * documentation for explanation of contents.
     */
    RandomAccessFile docFile;

    /**
     * The WAIS file name list.  See section 4.3.2 of the freeWAIS
     * documentation for explanation of contents.
     */
    RandomAccessFile fnFile;

    /**
     * A hashtable that maps from a fully qualified file name in the
     * source to the entry number for the respective file in the
     * WAIS database.  This entry is needed for recogniing the document
     * in the inverted index file.
     */
    Hashtable documentTable = new Hashtable();

    /**
     * A hashtable that maps from a fully qualified file name to the
     * number of terms in the document.
     */
    Hashtable documentSizeTable = new Hashtable();

    /**
     * Construct a WAISDocument instance.
     * @param indexLocation    The source specific location of the
     * indexes.
     * @param dataBase The indentifier of the database.
     */
    public WAISDocuments(String indexLocation, String dataBase)
        throws resource.SourceException {

        String docFileName = dataBase + ".doc";
        String fnFileName = dataBase + ".fn";

        try {
            docFile = new RandomAccessFile(
                new File(indexLocation, docFileName), "r");
            fnFile = new RandomAccessFile(
                new File(indexLocation, fnFileName), "r");
        }
        catch (IOException e) {
            throw new resource.SourceException("Can't open document files " + e.getMessage());
        }
        System.out.println(
            "Starting load of documents for source " + dataBase);
        FillDocumentTable();
        System.out.println(
            "Finished load of documents for source " + dataBase);
        try {
            docFile.close();
            fnFile.close();
        }
        catch (IOException e) {}
    }

    /**
     * given a filename (fully qualified) in the database, return the
     * entryNumber of that filename in the document table.
     * @param fileName The file name for which data is to be returned.
     */
    public int GetDocumentEntryNumber(String fileName) throws resource.SourceException {
        if (!documentTable.containsKey(fileName)) {
            throw new resource.SourceException("Can't find a filename in the database");
        }
        return(((Integer)documentTable.get(fileName)).intValue());
    }

    /**
     * given a filename return the number of terms.
     * @param fileName The file name for which data is to be returned.
     */
    public long GetDocumentNumberTerms(String fileName) throws resource.SourceException {
        if (!documentSizeTable.containsKey(fileName)) {
            throw new resource.SourceException("Can't find a filename in the database");
        }
        return(((Long)documentSizeTable.get(fileName)).longValue());
    }


    /**
     * return the number of documents in the tables.
     */
    public int GetNumberDocuments() {
        return documentTable.size();
    }

    /**
     * Fill in the hash table with information gleaned form the document
     * file and filename file.
     */
    private void FillDocumentTable() throws resource.SourceException {
        int entryNumber = 1;    // it appears that they start with 1.
        long fileLength;
        byte buf[] = new byte[filenameOffsetLength];
        byte lenbuf[] = new byte[lengthLen];
        byte charBuf[] = new byte[1];
        StringBuffer fileName;
        long filePos = firstDocOffset;
        long fileNumTerms, filenameOffset;

        try {
            fileLength = docFile.length();
        }
        catch (IOException e) {
            throw new resource.SourceException("Error reading document files");
        }

        // loop through the document file reading each entry.  For
        // each entry find the corresponding filename in the filename
        // list file.  Then deposit in the documentTable hashtable an
        // entry with key filename and value entrynumber.  This entry
        // number is the one used in the inverted  file to show the
        // document in which a term appears.
        while (filePos < fileLength) {
            try {
                docFile.seek(filePos);

                // read in the offset in the filename list (first three bytes)
                for (int i = 0; i < filenameOffsetLength; i++) {
                    buf[i] = docFile.readByte();
                }
                filenameOffset = ReadBytes(buf);

                // documentation states that a zero filled block should
                // be skipped
                if (filenameOffset == 0) {
                    filePos += docEntryLength;
                    continue;
                }
                // otherwise get the filename from the filename file
                fnFile.seek(filenameOffset);
                fileName = new StringBuffer(64);
                do {
                    fnFile.read(charBuf, 0, 1);
                    if (charBuf[0] != 0) {
                        fileName.append(new String(charBuf, 0));
                    }
                } while (charBuf[0] != 0);

                // now deposit in the hash table
                documentTable.put(fileName.toString(),
                    new Integer(entryNumber++));


                // now get the number of tokens in the document and
                // deposit in hash table
                docFile.seek(filePos + lengthOffset);
                for (int i = 0; i < lengthLen; i++) {
                    lenbuf[i] = docFile.readByte();
                }
                fileNumTerms = ReadBytes(lenbuf);
                documentSizeTable.put(fileName.toString(),
                    new Long(fileNumTerms));

                docFile.skipBytes(docEntryLength - filenameOffsetLength);
                filePos += docEntryLength;
            }
            catch (IOException e) {
                throw new resource.SourceException("Error reading document files");
            }
        }
    }


    /**
     * Convert a byte array to number
     * @param buf  The byte array to convert.
     */
    private long ReadBytes(byte buf[]) {
        long answer = 0;
        for (int i = 0; i < buf.length; i++) {

            // This looks like a java bug to me!!!  Even though buf is
            // a byte array, if the high bit is on it fills a high byte
            // with on bits, so something like 0x81 turns into 0xff81.
            // This happens actually at the time of the read.  So, I
            // work around this by zeroing the high byte. (Boy did I
            // waste time on this!!
            answer = (answer << 8) + (buf[i] & 0xff);
        }
        return answer;
    }

}
