/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.io.*;
import STARTSConfiguration.*;
import resource.*;

/**
 * Class to map to a WAISInvertedIndex file. The sole constructor allows us to
 * open up an inverted index file given a database name (source) and a
 * field.  From this file we can extract data about terms and data about terms
 * with repect to documents.
 */
public class WAISInvertedIndex {

    // constants for manipulating inverted index files

    /**
     * Offset in an index entry to the number of postings
     */
    static int numPostingsOffset = 1;

    /**
     * Size in bytes of the number of postings entry
     */
    static int numPostingsSize = 4;

    /**
     * Size in bytes of the docid entry
     */
    static int docidSize = 4;

    /**
     * Size in bytes of the file character position entry for the term
     */
    static int charPositionSize = 4;

    /**
     * Size in bytes of the term weight entry
     */
    static int weightSize = 4;

    /**
     * Size in bytes of the first occurrance or the term entry
     */
    static int firstOccSize = 3;

    /**
     * Size in bytes of the entry indicating the number of occurrances
     * of the term in a document
     */
    static int totalOccsSize = 4;

    /**
     * The file containing the inverted index
     */
    RandomAccessFile invFile;

    /**
     * Create (open up) a WAISInverted index that corresponds to the
     * respective source/field
     * @param indexLocation    Source specific location of index files.
     * @param database ID of the database
     * @param  field   Field identifier for the database (may be null).
     */
    public WAISInvertedIndex(String indexLocation, String dataBase,
        String field) throws resource.SourceException {

        String fileName = dataBase;

        // special case any which means just use the general inverted index
        if (field.equalsIgnoreCase("any")) {
            field = null;
        }

        if (field != null) {
            fileName = fileName + "_field_" + field;
        }
        fileName = fileName + ".inv";

        try {
            invFile = new RandomAccessFile
            (new File(indexLocation, fileName), "r");
        }
        catch (IOException e) {
            throw new resource.SourceException(e.getMessage());
        }
    }

    /**
     * Close the inverted index file.  Garbage collection (a finalize
     * method) could do this but at an unreliable time.
     */
    public void Close() {
        try {
            invFile.close();
        }
        catch (IOException e) {}
    }

    /**
     * Get the term info (not document specific) for a given term at a given
     * offset in the inverted index.  The pieces of information we can get
     * at this level are:
     * - the total number of occurrences of the term in the source
     * (multiple occurences in documents count)
     * - number of postings (documents) for the term in the inverted
     * index file - this should match the information in the dictionary
     * file.
     * @param term The term to extract information for.
     * @param offset   The offset in the inverted index file at which the
     * term resides.
     */
    public WAISTermInfo GetWAISTermInfo(String term, long offset)
        throws resource.SourceException {

        byte buf[];
        long documentFrequency = 0;
        long totalOccs = 0;

        try {
            // backpeddle past the term (\n terminated) to the beginning of the
            // total occurrences bytes
            invFile.seek(offset - term.length() - 1 - totalOccsSize);
            buf = new byte[totalOccsSize];
            invFile.read(buf, 0, totalOccsSize);
            totalOccs = ReadBytes(buf);

            // skip forward to the number of postings bytes
            invFile.seek(offset + numPostingsOffset);
            buf = new byte[numPostingsSize];
            invFile.read(buf, 0, numPostingsSize);
            documentFrequency = ReadBytes(buf);
            return(new WAISTermInfo((int) documentFrequency, (int) totalOccs));
        }
        catch (IOException e) {
            throw new resource.SourceException(e.getMessage());
        }
    }


    /**
     * Given an offset into the inverted index and an entry number of
     * a document return data the terms within that document.  The info.
     * we can get here is:
     * - a float value which gives the weight of the term with respect
     * to the document.
     * - the number of the occurrences of the term in the document.
     * @param offset  The offset into the inverted index file of the
     * term.
     * @param docEntryNum The entry number of the document for which
     * we wish to extract term specific information.
     */
    public WAISTermDocInfo GetWAISTermDocInfo(long offset, int docEntryNum)
        throws resource.SourceException {

        byte buf[];
        int numPostings;
        int postingNum = 0;
        int occListSize;
        int docid;
        float weight;
        DataInputStream dStream;

        try {
            invFile.seek(offset + numPostingsOffset);
            buf = new byte[numPostingsSize];
            invFile.read(buf, 0, numPostingsSize);

            // numPostings gives us the number of documents that the
            // term appears in
            numPostings = (int) ReadBytes(buf);
            invFile.skipBytes(4);   // skip over to the first posting

            // Ok, now we are at the first posting for this term.  Loop
            // through the postings to find the one that corresponds to the
            // docEntryNum
            while (postingNum < numPostings) {

                // extract the docid, size, wieght, and number of postings
                // for this term/document pair.  You may wonder why I am doing
                // this without even knowing if this is the right document yet.
                // Even if this is the wrong document, the only way we can
                // move to the next postingis by reading everything, since
                // the size of the posting is variable. (YUCH!!)
                buf = new byte[docidSize];
                invFile.read(buf, 0, docidSize);
                docid = (int) ReadBytes(buf);
                buf = new byte[charPositionSize];
                invFile.read(buf, 0, charPositionSize);
                occListSize = (int) ReadBytes(buf);
                buf = new byte[weightSize];
                invFile.read(buf, 0, weightSize);
                dStream =
                    new DataInputStream(new ByteArrayInputStream(buf));
                weight = dStream.readFloat();

                // now skip over the first occurrence (which is of fixed size)
                // and set numOccs (which will hold the number of occurrences
                // of the term in the respective document) to 1.
                invFile.skipBytes(firstOccSize);
                int occListRead = 3;    // number of bytes read from occList
                int numOccs = 1;

                // now consume all the other term occurrence pointers.  The
                // rest are of vairable size and must be painfully read
                // byte by byte.  This process returns the number of additional
                // entries of the term in the respective document.
                numOccs += ConsumeCompressedIntegers(
                    occListSize - occListRead);

                // OK, weve got everything. If this is the document we were
                // looking for then return the term info. for this document.
                if (docEntryNum == docid) {
                    return(new WAISTermDocInfo(weight, numOccs));
                }
                else {
                    postingNum++;
                }
            }

            // we'll only reach here if we didn't find the docEntryNum in the
            // list of postings for the term.  This means that the term does
            // not occur in the respective document.
            return null;
        }
        catch (IOException e) {
            throw new resource.SourceException("Inconsistency in WAIS inverted index");
        }
    }

    /**
     * Very ugly and probably slow code to consume a list of compressed integers
     * in an inverted index file and count the number integers in the list.
     * We only know the number of bytes that remain to be read.  Each entry
     * can be a variable number of bytes.  Return the number of these funny
     * entries we read (each of which corresponds to an entry of the term
     * in the respective document).
     * @param  bytesRemaining  Number of bytes to read
     */
    private int ConsumeCompressedIntegers(int bytesRemaining)
        throws IOException {

        byte buf;
        int numOccs = 0;

        while (bytesRemaining > 0) {
            buf = invFile.readByte();

            // continue bit is high bit 8.  If it is not on, then we have found
            // the end of an occurrence.
            if (buf <= 127) {
                numOccs++;
            }
            bytesRemaining--;
        }

        return(numOccs);
    }

    /**
     * Convert a buffer of bytes to a long value.
     * @param buf  The buffer of bytes.
     */
    private long ReadBytes(byte buf[]) {
        long answer = 0;
        for (int i = 0; i < buf.length; i++) {

            // This looks like a java bug to me!!!  Even though buf is
            // a byte array, if the high bit is on it fills a high byte
            // with on bits, so something like 0x81 turns into 0xff81.
            // This happens actually at the time of the read.  So, I
            // work around this by zeroing the high byte. (Boy did I
            // waste time on this!!
            answer = (answer << 8) + (buf[i] & 0xff);
        }
        return answer;
    }


}
