/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.util.*;

import query.*;
import results.*;
import resource.*;

/**
 * An ExpNode that represents a proximity operatory in a WAIS query
 * expression.  Each WAISProxOp has two children, which are the
 * operands of the proximity operator and two properties:
 * - ordered - a boolean determining whether word order
 * counts in the proximity.
 * - unit - the "number of words" value for the proximity.
 */
public class WAISProxOp extends ExpNode  implements SourceNode{

    /**
     * Construct a new WAISProxOp.
     * @param query    The query context of the proximity operator
     * @param ordered  A boolean determining whether word order
     * matters
     * @param  unit    The "number of words between" value for the
     * proximity.
     */
    public WAISProxOp(Query query, boolean ordered, int unit) {
        super(query);
        SetPropertyValue("ordered", new Boolean(ordered));
        SetPropertyValue("unit", new Integer(unit));
    }

    /**
     * Predicate to determine whether word order matters for the
     * proximity.
     */
    public boolean Ordered_p() {
        Boolean ordered = (Boolean) GetPropertyValue("ordered");
        return(ordered.booleanValue());
    }

    /**
     * Return the proximity distance value.
     */
    public int ProxDistance() {
        Integer unit = (Integer) GetPropertyValue("unit");
        return(unit.intValue());
    }

    /**
     * Return the string the represents the unparsed representation
     * of the sub-tree rooted at the WAISProxOp
     */
    public String Unparse() {
        String child0String, child1String, proxOp;
        Vector children;

        children = GetChildren();
        child0String = ((ExpNode) children.elementAt(0)).Unparse();
        child1String = ((ExpNode) children.elementAt(1)).Unparse();

        if (Ordered_p()) {
            proxOp = "w/" + String.valueOf(ProxDistance());
        }
        else {
            proxOp = "pre/" + String.valueOf(ProxDistance());
        }

        return(child0String + " " + proxOp + " " + child1String);
    }

    /**
     * This is a slot to do some primitive ranking on this sub-tree.
     * Unimplemented now except for the recursion
     * @param doc  The document to rank using this sub-tree
     */
    public float RankBy(Document doc) {
        Vector children = GetChildren();
        float lRank = ((SourceNode) children.elementAt(0)).RankBy(doc);
        float rRank = ((SourceNode) children.elementAt(1)).RankBy(doc);
        return ((float) 0.0);
    }


}
