/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.util.*;

import query.*;
import results.*;
import resource.*;

/**
 * An ExpNode that represents a relational Operator in a WAIS query
 * expression.  Each WAISBoolOp has one child, which is always a
 * WAISWordLiteral, and two properties:
 * operator - a string that is the relational operator (e.g. =, >)
 * field - the database field for the operator (e.g. au, ti)
 */
public class WAISRelOp extends ExpNode implements SourceNode{

    /**
     * Construct a new WAISRelOp.
     * @param query    Query context of the relational operator
     * @param operator String that is the operator.
     * @param field    Database field for the operator.
     */
    public WAISRelOp(Query query, String operator, String field) {
        super(query);
        SetPropertyValue("operator", operator);
        SetPropertyValue("field", field);
    }

    /**
     * Return the string that is the operator for this node
     */
    public String GetOperator() {
	    return((String) GetPropertyValue("operator"));
    }

    /**
     * Return the string that is the database field for this
     * operator.
     */
    public String GetField() {
        return((String) GetPropertyValue("field"));
    }


    /**
     * Return the string that represents the unparsed representation
     * of the sub-tree rooted at this WAISRelOp.
     */
    public String Unparse() {
        String childString;
        Vector children;
        String fieldU, operatorU;

        fieldU = GetField();
        operatorU = GetOperator();

        children = GetChildren();
        childString = ((ExpNode) children.elementAt(0)).Unparse();

        // Must handle numeric operators different from string (no () around
        // numeric
        if (!(operatorU.equals("==") || operatorU.equals(">") ||
            operatorU.equals("<"))) {
            // Nesting parents neccesary for non-wordliteral
            if (! children.elementAt(0).getClass().getName().equals(
                "resource.WAISWordLiteral")) {
                    childString = "(" + childString + ")";
            }
            else {  // more than one token we must form a phrase search
                StringTokenizer st = new StringTokenizer(childString);
                if (st.countTokens() > 1) {
                    childString = "(\"" + childString + "\")";
                }
            }
        }

        // SPecial case for any - just send back the childString
        if (fieldU.equalsIgnoreCase("any")) {
            return childString;
        }
        else {
            return(GetField() + GetOperator() + childString);
        }

    }

    /**
     * This is a slot to do some primitive ranking on this sub-tree.
     * Unimplemented now except for the recursion
     * @param doc  The document to rank using this sub-tree
     */
    public float RankBy(Document doc) {
        Vector children = GetChildren();
        float lRank = ((SourceNode) children.elementAt(0)).RankBy(doc);
        return ((float) 0.0);
    }

}
