/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.util.*;

import results.*;
import query.*;
import resource.*;

/**
 * A document returned by the WAIS search engine.  The main purpose of
 * this class is  to take the result string returned by the WAIS search
 * engine and turn it into a generic result.
 */
public class WAISResultDocument extends Document {

    /**
     * Full pathname of the file in which lies this document, so we
     * can extract information from it.
     */
    String pathname;

    /**
     * No argument constructor which allows us to generically sub-class
     * from this class
     */
    public WAISResultDocument() {}

    /**
     * Construct a new WAISResult document from a "hit" string returned
     * from the WAIS search engine.
     * @param query    The query context of this result.
     * @param engineString A "hit" returned from WAIS with the following
     * format:
     * - raw score (integer)
     * - number of lines in the document (integer)
     * - document length in bytes (integer)
     * - document id - a single quoted token with three sub-tokens
     * - start position in respective file of this document
     * - end position in respective file of this document
     * - the pathname of the file in which this document appears
     * - headline of this document
     */
    public WAISResultDocument(Query query, float rawScore,
        long documentSize, String pathname)
        throws  resource.SourceException {
        super(query);

        try {
            this.rawScore = rawScore;
            this.documentSize = documentSize;
            sources = new String[1];
            sources[0] = query.source.GetSourceID();
            if (pathname.endsWith("\'")) {
                pathname = pathname.substring(0, pathname.length() - 1);
            }
            this.pathname = pathname;
            int documentEntryNum;

            // now comes the harder stuff which we have to extract from
            // the ancillary WAIS files.

            WAISDocuments dt =
                ((WAISSourceDescription) query.source).GetDocumentTable();

            documentCount = dt.GetDocumentNumberTerms(pathname);

            // the entry number for the document allows us to index
            // into other tables.
            documentEntryNum = dt.GetDocumentEntryNumber(pathname);

            // Now gather statistics for each of the terms in the
            // ranking expression
            if (query.rankingTerms != null) {

                termStats =
                    new TermStatistics[query.rankingTerms.size()];
                for (int i = 0; i < query.rankingTerms.size(); i++) {
                    Term t = (Term) query.rankingTerms.elementAt(i);
                    String field = t.GetField().GetFieldName();
                    String attrSet = t.GetField().GetAttributeSet();

   /*****************************************************************************/
   // sdp - changed to reflect the move of the TranslateField method
   //       to attribute set classes (7/2/97)

                    String xlateField =
                        query.source.GetAttrSet(attrSet).TranslateField(field);
                        //query.source.TranslateField(attrSet, field);

   /*****************************************************************************/

                    String termVal = t.GetlString().GetString();

                    // open the dictionary for this field
                    WAISDictionary dict =
                        new WAISDictionary(
                            ((WAISSourceDescription) query.source).GetIndexLocation(),
                            query.source.GetSourceID(), xlateField);

                    // find the token in the dictionary
                    WAISDictionaryTerm token =
                        dict.FindTerm(termVal);
                    dict.Close();

                    // term not in dictionary
                    if (token == null) {
                        termStats[i] =
                            new TermStatistics(
                                field, termVal, 0, (float) 0.0, 0);
                    }

                    // Term in dictionary, get its data.
                    else {
                        float weight = (float) 0.0;
                        int termFrequency = 0;

                        // can get the number of documents that the term
                        // appears in from the dictionary file
                        int docFrequency = token.GetNumberDocuments();

                        // have to go to the inverted file for the rest.
                        long invertedFileLocation =
                            token.GetInvertedFileLocation();
                        WAISInvertedIndex invIndex =
                            new WAISInvertedIndex(
                                ((WAISSourceDescription)query.source).GetIndexLocation(),
                                query.source.GetSourceID(), xlateField);

                        // term info in the inverted index will give us
                        // the total number of occurrences of this term
                        // in the source.
                        WAISTermInfo termInfo =
                            invIndex.GetWAISTermInfo(termVal,
                            invertedFileLocation);
                        int totalOccs = termInfo.GettotalOccs();

                        if (totalOccs > 0) {
                            // finally get information for the term with respect to
                            // the document.
                            WAISTermDocInfo termDocInfo =
                                invIndex.GetWAISTermDocInfo(
                                invertedFileLocation, documentEntryNum);

                            // term didn't occur in the document
                            if (termDocInfo == null) {
                                weight = 0.0f;
                                termFrequency = 0;
                            }
                            // found the term, get document specific information.
                            else {
                                weight = termDocInfo.GetWeight();
                                termFrequency = termDocInfo.GetTermFrequency();
                            }
                        }

                        // load up the term stats entry with thhe information.
                        termStats[i] = new TermStatistics(field, termVal,
                            termFrequency, weight, docFrequency);
                        invIndex.Close();
                    }

                }
            }


        }
        catch (NoSuchElementException e) {
            throw new resource.SourceException(
            "Invalid response from WAIS search engine");
        }

    }

    /**
     * Return a string that is the document value for a specific field
     * e.g. author.
     * @param f    The field to return a value for.
     * @param maxLen   The maximum length of the string value returned.
     */
    public String GetFieldValue(Field f, int maxLen)
        throws resource.SourceException{
        Class docClass;
        Object docClassObj;

        // using the sourceID of the query create the name of the class
        // Then instantiate the source specific document class for this
        // document hit.   The implementation convention is that source
        // specific document classes are called <source>Document (e.g.,
        // CSTRDocument.
        String sourceID = query.source.GetSourceID();
        String className = sourceID.toUpperCase() + "Document";
        try {
            docClass = Class.forName("resource." + className);
            docClassObj = docClass.newInstance();
        }
        catch (InstantiationException e1) {
            throw new resource.SourceException(e1.getMessage());
        }
        catch (ClassNotFoundException e1) {
            throw new resource.SourceException(e1.getMessage());
        }
        catch (IllegalAccessException e1) {
            throw new resource.SourceException(e1.getMessage());
        }
/***********************************************************************************/
/* Sandy Payette (7/24/97)
       I'm passing the WAIS translated field value to the source document class
       (e.g., CSTRDocument, LINUXDocument).  The idea is to go from a WAIS field
       to a source document tag not from an attribute set field to a document tag.
       By the time we get to creating the result document, the attribute-set fields
       have already been translated to WAIS fields.  The Document class sub-classes
       WAISResultDocument, so we really should be dealing with WAIS fields at this point.
       Also, when we consider that there is a one-to-many relationship between
       attribute-set fields and a document tags, it is preferable to implement
       the one-to-one relationship of a WAIS field to a document tag.
       (See transTable in CSTRDocument and LINUXDocument.)
*/


 String translatedField =
    query.source.GetAttrSet(f.GetAttributeSet()).TranslateField(f.GetFieldName());

        // now through the magic of sub-classing this will actually go
        // to the source specific class method (e.g.
        // <source>Document.GetFieldValueFromDB

 return (((WAISResultDocument) docClassObj).GetFieldValueFromDB
                                            (pathname, translatedField, maxLen));

        // original code was:
        //return (((WAISResultDocument) docClassObj).GetFieldValueFromDB
        //    (pathname, f.GetFieldName(), maxLen));

    }

/***********************************************************************************/

    // declare this method so we can do generic call to
    // ((WAISResultDocument) object).GetFieldValueFromDB  The actual
    // call will be routed to the subclass of (WAISResultDocument) that
    // is correct for the document.
    public String GetFieldValueFromDB(String path, String fieldName,
        int maxLen)
        throws resource.SourceException {
        // never used, only sub-classes implement this
        return null;
    }
}
