/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.util.*;

import query.*;
import results.*;

/**
 * The class for combining the results returned by filter queries
 * and ranking queries to the WAIS search engine.  WAIS does not
 * have a facility for internally handling.  So our way around it is
 * to submit two queries - one based on the filter expression and
 * one on the ranking expression.  The final result set includes
 * all documents in the filter expression with their socre adjusted
 * as follows - if the document appears in the result set from the
 * ranking expression, use the ranking score as the score; if the
 * documennt does not appear in the result set from the ranking
 * expressing, set its score to 0.
 */
public class WAISResultSet {

    /**
     * Query context of this results set
     */
    private Query query;

    /**
     * Hashtable to hold filter results.  The key is the pathname
     * of the result document
     */
    private Hashtable filterResults;

    /**
     * Hashtable to hold ranking results.  The key is the pathname
     * of the result document
     */
    private Hashtable rankingResults;

    /**
     * Instantiate a new WAISResultSet and create the hashtables.
     */
    public WAISResultSet(Query query) {
        this.query = query;
        filterResults = new Hashtable();
        rankingResults = new Hashtable();
    }

    /**
     * Add a new document to the filter results set
     * @param result   The result string from the WAIS search engine.
     */
    public void AddFilterResult(String result)
        throws resource.SourceException {
        WAISResult wr = new WAISResult(query, result);
        filterResults.put(wr.pathname, wr);
    }

    /**
     * Add a new document to the ranking results set
     * @param result   The result string from the WAIS search engine.
     */
    public void AddRankingResult(String result)
        throws resource.SourceException{
        WAISResult wr = new WAISResult(query, result);
        rankingResults.put(wr.pathname, wr);
    }

    /**
     * Merge the filter and ranking result sets and create a generic
     * results set including the merged documents.  Return the final
     * number of results
     */
    public int MergeFilterAndRanking() throws resource.SourceException {
        Hashtable mergedResults;

        // if there are no filter results, just return the restuls
        // of the ranking expression.
        if (filterResults.size() == 0) {
            mergedResults = rankingResults;
        }

        // if there are no ranking results, just return the results
        // of the filter expression.
        else if (rankingResults.size() == 0) {
            mergedResults = filterResults;
        }

        // otherwise do the merge.
        else {
            mergedResults = filterResults;

            // enumerate through the filter results, if document
            // is in ranking set use the score from the ranking set.
            // othersize set the score to 0.0.
            for (Enumeration e = filterResults.keys();
                e.hasMoreElements();) {
                String key = (String) e.nextElement();
                WAISResult fr = (WAISResult) filterResults.get(key);
                if (rankingResults.containsKey(key)) {
                    WAISResult rr = (WAISResult) rankingResults.get(key);
                    fr.rawScore = rr.rawScore;
                }
                else {
                    fr.rawScore = 0.0f;
                }
            }
        }

        // Create a new generic result set with these document.
        query.results = new ResultSet(query, mergedResults.size());

        for (Enumeration e = mergedResults.keys(); e.hasMoreElements();) {
            String key = (String) e.nextElement();
            WAISResult mr = (WAISResult) mergedResults.get(key);

            WAISResultDocument wrd = new WAISResultDocument(
                query, mr.rawScore, mr.documentSize,
                mr.pathname);
        }

        return (mergedResults.size());
    }
}

/**
 * private class to encapsulate the informatiohn returned in a WAIS
 * result.
 */
class WAISResult {

    /**
     * Rawscore of the document
     */
    float rawScore;

    /**
     * number of lines in the document.
     */
    long numLines;

    /**
     * Size of the document in bytes.
     */
    long documentSize;

    /**
     * Source that the document appears in.
     */
    String sources[];

    /**
     * Unique pathname of the document (maybe should be a URL or
     * URN in the future?
     */
    String pathname;

    /**
     * Instantiate a new WAISResult
     * @param query    Query context of the result
     * @param engineString String returned from WAIS containing the
     * document inforamtion
     */
    WAISResult(Query query, String engineString)
        throws resource.SourceException {
        StringTokenizer st = new StringTokenizer(engineString);

        try {
            // extract readily available document information first.
            rawScore = (new Float(st.nextToken())).floatValue();
            numLines = (new Long(st.nextToken())).longValue();
            documentSize = ((new Long(st.nextToken())).longValue());
            sources = new String[1];
            sources[0] = query.source.GetSourceID();
            // skip next two tokens
            st.nextToken();
            st.nextToken();
            pathname = st.nextToken();
            if (pathname.endsWith("\'")) {
                pathname = pathname.substring(0, pathname.length() - 1);
            }
        }
        catch (NoSuchElementException e) {
            throw new resource.SourceException(
            "Invalid response from WAIS search engine");
        }
    }
}

