/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.util.*;
import java.net.*;

import resource.*;
import query.*;
import soif.*;
import STARTSConfiguration.*;
import regexp.*;

import java.io.*;

/**
 * This interface encapsulates the stuff that is special about WAIS
 * type sources
 */
public abstract class WAISSourceDescription extends SourceDescription {

    /**
     * Host on which the WAIS search engine is running
     */
    protected static String WAISHost = ServerConfiguration.hostName;

    /**
     * Port on which the WAIS search engine is listenning
     */
    protected static String WAISPort = "5000";

    /**
     * The set of standard stop words for the freeWAIS search engine.
     * Unfortunately these are hardcoded in the file ir/stoplist.c
     * and must be, therefore, hardcoded here.
     */
    protected static String stopWords[] = {
        "a",
        "about",
        "above",
        "according",
        "across",
        "actually",
        "adj",
        "after",
        "afterwards",
        "again",
        "against",
        "all",
        "almost",
        "alone",
        "along",
        "already",
        "also",
        "although",
        "always",
        "among",
        "amongst",
        "an",
        "and",
        "another",
        "any",
        "anyhow",
        "anyone",
        "anything",
        "anywhere",
        "are",
        "aren't",
        "around",
        "as",
        "at",
        "b",
        "be",
        "became",
        "because",
        "become",
        "becomes",
        "becoming",
        "been",
        "before",
        "beforehand",
        "begin",
        "beginning",
        "behind",
        "being",
        "below",
        "beside",
        "besides",
        "between",
        "beyond",
        "billion",
        "both",
        "but",
        "by",
        "c",
        "can",
        "can't",
        "cannot",
        "caption",
        "co",
        "co.",
        "could",
        "couldn't",
        "d",
        "did",
        "didn't",
        "do",
        "does",
        "doesn't",
        "don't",
        "down",
        "during",
        "e",
        "each",
        "eg",
        "eight",
        "eighty",
        "either",
        "else",
        "elsewhere",
        "end",
        "ending",
        "enough",
        "etc",
        "even",
        "ever",
        "every",
        "everyone",
        "everything",
        "everywhere",
        "except",
        "f",
        "few",
        "fifty",
        "first",
        "five",
        "for",
        "former",
        "formerly",
        "forty",
        "found", "",
        "four",
        "from",
        "further",
        "g",
        "h",
        "had",
        "has",
        "hasn't",
        "have",
        "haven't",
        "he",
        "he'd",
        "he'll",
        "he's",
        "hence",
        "her",
        "here",
        "here's",
        "hereafter",
        "hereby",
        "herein",
        "hereupon",
        "hers",
        "herself",
        "him",
        "himself",
        "his",
        "how",
        "however",
        "hundred",
        "i",
        "i'd",
        "i'll",
        "i'm",
        "i've",
        "ie",
        "if",
        "in",
        "inc.",
        "indeed",
        "instead",
        "into",
        "is",
        "isn't",
        "it",
        "it's",
        "its",
        "itself",
        "j",
        "k",
        "l",
        "last",
        "later",
        "latter",
        "latterly",
        "least",
        "less",
        "let",
        "let's",
        "like",
        "likely",
        "ltd",
        "m",
        "made",
        "make",
        "makes",
        "many",
        "maybe",
        "me",
        "meantime",
        "meanwhile",
        "might",
        "million",
        "miss",
        "more",
        "moreover",
        "most",
        "mostly",
        "mr",
        "mrs",
        "much",
        "must",
        "my",
        "myself",
        "n",
        "namely",
        "neither",
        "never",
        "nevertheless",
        "next",
        "nine",
        "ninety",
        "no",
        "nobody",
        "none",
        "nonetheless",
        "noone",
        "nor",
        "not",
        "nothing",
        "now",
        "nowhere",
        "o",
        "of",
        "off",
        "often",
        "on",
        "once",
        "one",
        "one's",
        "only",
        "onto",
        "or",
        "other",
        "others",
        "otherwise",
        "our",
        "ours",
        "ourselves",
        "out",
        "over",
        "overall",
        "own",
        "p",
        "per",
        "perhaps",
        "q",
        "r",
        "rather",
        "recent",
        "recently",
        "s",
        "same",
        "seem",
        "seemed",
        "seeming",
        "seems",
        "seven",
        "seventy",
        "several",
        "she",
        "she'd",
        "she'll",
        "she's",
        "should",
        "shouldn't",
        "since",
        "six",
        "sixty",
        "so",
        "some",
        "somehow",
        "someone",
        "something",
        "sometime",
        "sometimes",
        "somewhere",
        "still",
        "stop",
        "such",
        "t",
        "taking",
        "ten",
        "than",
        "that",
        "that'll",
        "that's",
        "that've",
        "the",
        "their",
        "them",
        "themselves",
        "then",
        "thence",
        "there",
        "there'd",
        "there'll",
        "there're",
        "there's",
        "there've",
        "thereafter",
        "thereby",
        "therefore",
        "therein",
        "thereupon",
        "these",
        "they",
        "they'd",
        "they'll",
        "they're",
        "they've",
        "thirty",
        "this",
        "those",
        "though",
        "thousand",
        "three",
        "through",
        "throughout",
        "thru",
        "thus",
        "to",
        "together",
        "too",
        "toward",
        "towards",
        "trillion",
        "twenty",
        "two",
        "u",
        "under",
        "unless",
        "unlike",
        "unlikely",
        "until",
        "up",
        "upon",
        "us",
        "used",
        "using",
        "v",
        "very",
        "via",
        "w",
        "was",
        "wasn't",
        "we",
        "we'd",
        "we'll",
        "we're",
        "we've",
        "well",
        "were",
        "weren't",
        "what",
        "what'll",
        "what's",
        "what've",
        "whatever",
        "when",
        "whence",
        "whenever",
        "where",
        "where's",
        "whereafter",
        "whereas",
        "whereby",
        "wherein",
        "whereupon",
        "wherever",
        "whether",
        "which",
        "while",
        "whither",
        "who",
        "who'd",
        "who'll",
        "who's",
        "whoever",
        "whole",
        "whom",
        "whomever",
        "whose",
        "why",
        "will",
        "with",
        "within",
        "without",
        "won't",
        "would",
        "wouldn't",
        "x",
        "y",
        "yes",
        "yet",
        "you",
        "you'd",
        "you'll",
        "you're",
        "you've",
        "your",
        "yours",
        "yourself",
        "yourselves",
        "z"
    };

    /**
     * Hashtable for quick access to the stop words
     */
    protected Hashtable stopWordsTable;

    /* !!!!! PLATFORM SPECIFIC CODE !!!!! */

    /**
     * The pathname that is location of the indexing files
     * for this search engine.
     */
     /* !!!!! NT VERSION !!!!! */
    //public String indexLocation = "e:\\starts_stuff\\indexes";

    /* !!!!! SOLARIS VERSION !!!!! */
    public String indexLocation = "/home/payette/starts_stuff/indexes";

    /* !!!!! END PLATFORM SPECIFIC CODE !!!!! */

    /**
     * Holds the in memory copy of the WAIS document table (loaded
     * at startup) which speeds access to the information within
     */
	public WAISDocuments documentTable;

    /**
     * translation table to go from STARTS operator to WAIS operator.
     * Formatted as a two-D array with column 0 being the STARTS
     * operator and column 1 the WAIS equivalent.
     */
    protected static String[][] operatorsTranslation = {
        {"and", "and"},
        {"or", "or"},
        {"and-not", "not"},
        {"prox", ""}
    };

	/**
	 * A string array indicating the modifiers supported.
	 * Each row is data for a modifier.  The elements (columns) of a row are
	 * as follows:
	 * 0 - two tokens <attribute set> <modifier>.  Note that rows for all
	 * basic-1 fields should exist.  Rows for modifiersin other
	 * attribute sets may be added
	 * 1 - a string value (true of false) indicating whether the modifieris
	 * supported at the source
	 * 2 - a set of tokens indicating special languages supported for this
	 * modifier(e.g. "en-US es-MX")
	 * Note that this information is source (not engine) specific
	 */
	public static String[][] modifiersSupported = {
		{"basic-1 <", "true", ""},  // required
		{"basic-1 <=", "true", ""}, // required
		{"basic-1 =", "true", ""}, // required
		{"basic-1 >", "true", ""}, // required
		{"basic-1 >=", "true", ""},    // required
		{"basic-1 phonetic" , "true", ""},
		{"basic-1 stem" , "false", ""},
		{"basic-1 thesaurus" , "false", ""},
		{"basic-1 right-truncation" , "true", ""},
		{"basic-1 left-truncation" , "false", ""},
		{"basic-1 case-sensitive" , "false", ""},

	};

    /**
     * A two-d string array indicating operators supported.  Column 0 is
     * the STARTS operator and column 1 true or false.
     */
    public static String[][] operatorsSupported = {
        {"and", "true"},    // required
        {"or", "true"}, // required
        {"and-not", "true"},    // required
        {"prox", "true"}    // required
    };


    /**
     * Table to translate values to STARTS modifier to WAIS modifier.
     * Formatted as a two-D array with:
     * 0 - attribute set and STARTS modifier
	 * 1 - WAIS equivalent of the STARTS modifier
	 * Note that this information is source (not engine) specific
     * Note that some translations are handled "in code" rather
     * than a direct table translation for example right-truncation.
     */

	protected static String[][] modifiersTranslation = {
		{"basic-1 <", "<"},
		{"basic-1 =", "="},
		{"basic-1 >", ">"},
		{"basic-1 phonetic" , "SOUNDEX"},
		{"basic-1 right-truncation", "right truncation"}
	};



    /**
     * A boolean indicating if proximity is supported.
     */
    public static boolean proximitySupported = true;

    /**
     * Boolean values indicating whether the source supports ranking.
     */
	public static boolean supportsRanking = true;

    /**
     * Boolean values indicating whether the source supports filters.
     */
	public static boolean supportsFilters = true;


	/**
	 * A string array giving the name of the ranking algorithm.
	 */
	public static String rankingAlgorithm = "freeWAIS";

    /**
     * A string array giving the identifiers of the tokenizers
     * used by the source.  Each element of the array is formatted
     * as follows:
     * <language-list> tokenizer-id
     * where <language-list> is an optional list of languages
     * for which the respective tokenizer is used and
     * tokenizer-id is the (single token) name of the tokenizer
     * An example array element is:
     * "en-US sp-MX Acme-1"  meaning the the tokenizer Acme-1
     * is used for both US english and Mexican spanish.
     */
	public static String[] tokenizerIDList = {
		"freeWAIS-sf"
	};

	// static initiatilization code - common to all sources.

    /**
     * Hashtable for quick modifier table translation
     */
	protected static Hashtable modifiersXlate =
		new Hashtable(modifiersTranslation.length);

    /**
     * Hashtable for quick operator table translation.
     */
    protected static Hashtable operatorsXlate =
        new Hashtable(operatorsTranslation.length);

    // load the translation hash tables.
	static {
	    int i;

		for (i = 0; i < modifiersTranslation.length; i++) {
			modifiersXlate.put(modifiersTranslation[i][0],
			    modifiersTranslation[i][1]);
		}

		for (i = 0; i < operatorsTranslation.length; i++) {
		    operatorsXlate.put(operatorsTranslation[i][0],
		        operatorsTranslation[i][1]);
		}

	}


    /**
     * Hashtable included all supported modifiers.  key is
     * <attribute-set> <modifier>
     * value is set of language tokens or empty string if no special
     * languages
     */
	public static Hashtable modifiersTable =
		new Hashtable(modifiersSupported.length);

    /**
     * Hashtable of all supported operators.  Key is operator,
     * value is null
     */
    public static Hashtable operatorsTable =
        new Hashtable(operatorsSupported.length);

	// load the above hash tables.
	static {
		int i;
		StringBuffer value;
		String field, modifier;
		StringTokenizer tokens;

		for (i = 0; i < modifiersSupported.length; i++) {
		    if (modifiersSupported[i][1].equalsIgnoreCase("true")) {
		        modifiersTable.put(modifiersSupported[i][0], modifiersSupported[i][2]);
		    }
		}


		for (i = 0; i < operatorsSupported.length; i++) {
		    if (operatorsSupported[i][1].equalsIgnoreCase("true")) {
		        operatorsTable.put(operatorsSupported[i][0], operatorsSupported[i][1]);
            }
		}

    }

	/**
	 * A boolean value indicating whether stop words can be
	 * turned off at this source
	 */
	public static boolean turnOffStopWords = false;

	/**
	 * A String indicating the default Meta attribute set
	 */
	public static String defaultMetaAttributeSet = "mbasic-1";

	/**
	 * A two element double array indicating the lower and
	 * upper bounds for relevance scores returned for each
	 * search hit.
	 */
	public static float[] scoreRange = {0.0f, 1000.0f};

  /**
   * Construct a new WAISSourceDescription
   * @param sourceID The name of the source
   */
    public WAISSourceDescription(String sourceID) {
        this.sourceID = sourceID;
        this.indexLocation = indexLocation;
        super.modifiersTable = modifiersTable;
        super.operatorsTable = operatorsTable;
        super.proximitySupported = proximitySupported;
        super.supportsRanking = supportsRanking;
        super.supportsFilters = supportsFilters;
        super.scoreRange = scoreRange;
        super.rankingAlgorithm = rankingAlgorithm;
        super.tokenizerIDList = tokenizerIDList;
        super.turnOffStopWords = turnOffStopWords;
        super.defaultMetaAttributeSet = defaultMetaAttributeSet;

        try {
            documentTable = new WAISDocuments(indexLocation, sourceID);
        }
        // Can't throw an exception here since this constructor is called
        // from static code in the ResourceDescription class.
        catch (SourceException e) {
            System.out.println(e.getMessage());
            System.out.println("Can't load the document table for source: " + sourceID);
            System.exit(1);
        }

        // Load the stop words table, this is a combination of
        // the standard WAIS stop word list and the one for this
        // source (the .stop file).
        stopWordsTable = new Hashtable(stopWords.length);

        // Standard WAIS stop words
        for (int i = 0; i < stopWords.length; i++) {
            stopWordsTable.put(stopWords[i], stopWords[i]);
        }

        LoadStopWordsFile();
    }

    /**
     * Return the pathname of the indexing files for this resource
     */
    public String GetIndexLocation() {
        return indexLocation;
    }

    /**
     * Return the in memory copy of the WAIS document table.
     */
    public WAISDocuments GetDocumentTable() {
        return documentTable;
    }

    /**
     * Return a String that is a translation from STARTS modifier
     * to WAIS modifier.
     * @param attrAndMod   Combined attribute set and modifier (space
     * separated) to be translated.
     */
    public String TranslateModifier(String attr, String mod) {
        return ((String) modifiersXlate.get(attr + " " + mod));
    }

    /**
     * Return a String that is a translation from STARTS operator to
     * WAIS operator.
     * @param op   The STARTS operator to be translated
     */
    public String TranslateOperator(String op) {
        return ((String) operatorsXlate.get(op));
    }

    /**
     * Translate from a STARTS field tree node to a WAIS one.  In
     * WAIS we handle this at the term level (from which all fields
     * descend.
     * @param field    field node to translate
     */
    public  ExpNode FieldToFilter(Field field)
        throws SourceQueryException {
        return null;
    }

    /**
     * Translate from a STARTS modifier list tree node to a WAIS one.  In
     * WAIS we handle this at the term level (from which all modifier
     * lists descend.
     * @param ml   modifier list to translate
     */
    public  ExpNode ModifierListToFilter(ModifierList ml)
        throws SourceQueryException {
        return null;
    }

    /**
     * Translate from a STARTS LString tree node to a WAIS one.  In
     * WAIS we handle this at the term level (from which all LStrings
     * descend.
     * @param string   LString to translate
     */
    public  ExpNode LStringToFilter(LString string)
        throws SourceQueryException {
        return null;
    }

    /**
     * Translate from a STARTS BoolOp to the WAIS equivalent.
     * @param bo   STARTS BoolOop to translate
     * @param leftChild    translated version of the left child of
     * this boolop
     * @param rightChild   trnslated version of the right child of this
     * boolop
     */
    public  ExpNode BoolOpToFilter(
        BoolOp bo, ExpNode leftChild, ExpNode rightChild)
        throws SourceQueryException {

        ExpNode derivedTree = null;
        String operator = bo.GetOperator();
        Query query = bo.GetQuery();

        String newOp = query.source.TranslateOperator(operator);
        if (newOp == null) {
            throw new SourceQueryException(
                "Unsupported boolean operator " + operator);
        }

        derivedTree = new WAISBoolOp(query, newOp);
        derivedTree.AddNewChild(leftChild);
        derivedTree.AddNewChild(rightChild);

        return (derivedTree);
    }

    /**
     * Translate from a STARTS ProxOp to the WAIS equivalent.
     * @param bo   STARTS ProxOp to translate
     * @param leftChild    translated version of the left child of
     * this ProxOp
     * @param rightChild   trnslated version of the right child of this
     * ProxOp
     */
    public  ExpNode ProxOpToFilter(ProxOp po,
        ExpNode leftTerm, ExpNode rightTerm)
        throws SourceQueryException {

        int distance = po.GetDistance();
        boolean wordOrderMatters = po.WordOrderMatters_p();
        Query query = po.GetQuery();

        ExpNode derivedTree = new WAISProxOp(query, wordOrderMatters, distance);
        derivedTree.AddNewChild(leftTerm);
        derivedTree.AddNewChild(rightTerm);

        return (derivedTree);
    }

    /**
     * Return the translated version of a STARTS Term node to the
     * WAIS equivalent.  NOTE - last three arguments are not used
     * for the WAIS source.  This method also makes some changes to the
     * STARTS Term for untranslatable things.
     * @param term term node to translated.
     * @param fieldXlate   tranlsated version of the field that descends
     * from this term.
     * @param modListXlate translated version of the modifier list that
     * descends from this term.
     * @param lStringXlate translated version of the LString that
     * descends from this term.
     */
    public  ExpNode TermToFilter(
        Term term, ExpNode fieldXlate, ExpNode modListXlate,
        ExpNode lStringXlset)
        throws SourceQueryException {

        ExpNode derivedTree = null, wordTree;
        String field = term.GetField().GetFieldName();
        String attrSet = term.GetField().GetAttributeSet();
        String lString = term.GetlString().GetString();
        String modArray[] = null;
        String attrArray[] = null;
        String language = term.GetlString().GetLanguage();
        ModifierList modifierList = term.GetModifierList();
        Query query = term.GetQuery();

        // Extract the string equivalent of the modifiers for the term
        if (modifierList != null) {
            Vector modifiers = modifierList.GetModifiers();
            Vector stringModifiers = new Vector(modifiers.size());
            Vector sModAttrs = new Vector(modifiers.size());
            for (int i = 0; i < modifiers.size(); i++) {
                Modifier m = (Modifier) modifiers.elementAt(i);
                stringModifiers.addElement(m.GetModifier());
                sModAttrs.addElement(m.GetAttributeSet());
            }

            modArray = new String[modifiers.size()];
            attrArray = new String[modifiers.size()];
            stringModifiers.copyInto(modArray);
            sModAttrs.copyInto(attrArray);
        }

/***************************************************************************************/
        // Tranlate the field specifier.

        String translatedField = query.source.GetAttrSet(attrSet).TranslateField(field);
        //   query.source.TranslateField(attrSet, field);

        // This should never happen becuase nothing will pass "check" that is
        // untranslatable.
        if (translatedField == null) {
            throw new SourceQueryException(
                "Unsupported field in filter: " + field);
        }

        // Pull out the LString
        String translatedLString = lString;

        // Date field requires special handling for lString.  We must go from the ISO
        // yyyy-mm-dd format to the WAIS yy-mm-dd format. NOTE!! Check for correct
        // ISO format was done earlier in check so we can be sure that lString
        // looks like yyyy-mm-dd.
        if (field.equalsIgnoreCase("date-last-modified")) {
            String year, month, day;
            Regexp reg = Regexp.compile("(\\d\\d\\d\\d)\\-(\\d\\d)\\-(\\d\\d)");
            Result result = reg.match(lString, 0);
            year = result.getMatch(1);
            month = result.getMatch(2);
            day = result.getMatch(3);
            translatedLString = year.substring(2) + month + day;
        }

        if (language != null) {
            throw new SourceQueryException(
                "Languages not supported at this source");
        }

        // Translate the modifier(s)
        String relOp = null;
        boolean soundex = false;

        // No modifier - use default '='
        if (modArray == null || modArray.length == 0) {
            if (field.equalsIgnoreCase("date-last-modified")) {
                relOp = "==";
            }
            else {
                relOp = "=";
            }
        }

        // There are modifiers to translate.
        else {
            // we will replace the list of modifiers at the end.
            Vector newMods = new Vector(modArray.length);

            // loop through the STARTS modifiers and translate.  Any operators
            // taht don't make sense we just omit.
            for (int i =0 ; i < modArray.length; i++) {

                // <, >, <=, and <= are only available for field date in wais.
                if ((modArray[i].equals(">") ||
                    modArray[i].equals("<") ||
                    modArray[i].equals("<=") ||
                    modArray[i].equals(">=")) &&
                    !field.equalsIgnoreCase("date-last-modified")) {
                    continue;
                }

                // >= and <= require special handling - will translate to a "or"
                // boolop
                else if (modArray[i].equals(">=") || modArray[i].equals("<=")) {
                    relOp = modArray[i];
                    newMods.addElement(new Modifier(query,modArray[i]));
                    continue;
                }



                String translatedMod =
                    (String) query.source.TranslateModifier(attrArray[i], modArray[i]);
                if (translatedMod == null) {
                    continue;
                }

                // Translate other modifiers - relational
                if (RelOp_p(translatedMod)) {
                    if (relOp != null) {
                        continue;
                    }
                    else {
                        relOp = translatedMod;
                    }
                }

                // soundex and right truncation
                else {
                    if (translatedMod.equalsIgnoreCase("soundex")) {
                        soundex = true;
                    }
                    else if (translatedMod.equalsIgnoreCase("right truncation")) {
                        translatedLString += "*";
                    }
                    else {
                        continue;
                    }
                }

                if (translatedField.equalsIgnoreCase("any") && !relOp.equals("=")) {
                    continue;
                }
                newMods.addElement(new Modifier(query,modArray[i]));
            }

            // If no modifiers remain, just use default =
            if (newMods.size() == 0) {
                newMods.addElement(new Modifier(query, "="));
            }

            // Now replace the modifiers in STARTS with the aftereffect of the
            // translation
            modifierList.SetModifiers(newMods);
        }

        // No relational operator defaults to '='
        if (relOp == null) {
            if (field.equalsIgnoreCase("date-last-modified")) {
                relOp = "==";
            }
            else {
                relOp = "=";
            }
        }

        // Now build up the WAIS tree
        // >= or <= is translated to or(>, =) or or(<,=)
        if ((relOp.equals(">=") || relOp.equals("<="))) {
            ExpNode child1, child2;
            wordTree = new WAISWordLiteral(query, soundex, translatedLString);
            child1 = new WAISRelOp(query, "==", translatedField);
            child1.AddNewChild((ExpNode) wordTree);
            child2 = new WAISRelOp(query, relOp.substring(0,1), translatedField);
            child2.AddNewChild((ExpNode) wordTree);
            return(BoolOpToFilter(new BoolOp(query, "or"), child1, child2));
        }
        else {
            wordTree = new WAISWordLiteral(query, soundex, translatedLString);
            derivedTree = new WAISRelOp(query, relOp, translatedField);
            derivedTree.AddNewChild((ExpNode) wordTree);
            return (derivedTree);
        }
    }


    /**
     * Return the translated version of a STARTS field node within
     * a ranking expression.  NOTE: WAIS does not support separae
     * ranking expressions that the return is the same for a filter
     * expression.
     * @param field    the STARTS field node
     */
    public  ExpNode FieldToRanking(Field field)
        throws SourceQueryException {
        return FieldToFilter(field);
    }

    /**
     * Return the translated version of a STARTS modifier list node within
     * a ranking expression.  NOTE: WAIS does not support separae
     * ranking expressions that the return is the same for a filter
     * expression..
     * @param ml    the STARTS modifier list node
     */
    public  ExpNode ModifierListToRanking(ModifierList ml)
        throws SourceQueryException {
        return ModifierListToFilter(ml);
    }

    /**
     * Return the translated version of a STARTS LString node within
     * a ranking expression.  NOTE: WAIS does not support separae
     * ranking expressions that the return is the same for a filter
     * expression..
     * @param string    the STARTS LString node
     */
    public  ExpNode LStringToRanking(LString string)
        throws SourceQueryException {
        return LStringToFilter(string);
    }

    /**
     * Return the translated version of a STARTS BoolOp node within
     * a ranking expression.  NOTE: WAIS does not support separae
     * ranking expressions that the return is the same for a filter
     * expression.
     * @param field    the STARTS BoolOp node
     * @param leftChild    translated left child of this boolop
     * @param rightChild   translated right child of this boolop
     */
    public  ExpNode BoolOpToRanking(
        BoolOp bo, ExpNode leftChild, ExpNode rightChild)
        throws SourceQueryException {

        return BoolOpToFilter(bo, leftChild, rightChild);
    }

    /**
     * Return the translated version of a STARTS ProxOp node within
     * a ranking expression.  NOTE: WAIS does not support separae
     * ranking expressions that the return is the same for a filter
     * expression..
     * @param field    the STARTS BoolOp node
     * @param leftChild    translated left child of this Prox
     * @param rightChild   translated right child of this ProxOp
     */
    public  ExpNode ProxOpToRanking(
        ProxOp po, ExpNode leftChild, ExpNode rightChild)
        throws SourceQueryException {

        return ProxOpToFilter(po, leftChild, rightChild);
    }

    /**
     * Return the translated version of a STARTS term node within a
     * ranking expresssion.  NOTE: WAIS does not support separae
     * ranking expressions that the return is the same for a filter
     * expression..
     * @param term the STARTS term node.
     * @param  fieldXlate  translated version of the field node that
     * descends from this term.
     * @param modListXlate  translated version fo the modifier list node
     * that descends from this term.
     * @param lStringXlate translated version of the lString node that
     * descends from this term.
     */
    public  ExpNode TermToRanking(Term term, ExpNode fieldXlate,
        ExpNode modListXlate,
        ExpNode lStringXlate)
        throws SourceQueryException {

        return TermToFilter(term, fieldXlate, modListXlate, lStringXlate);
    }

    /**
     * Return the translated version of a starts WeightedRanking node
     * within a ranking expression.  NOTE: WAIS does not support ranking
     * so this simply returns the original node.
     * @param wr   STARTS weighted ranking node to be translated.
     * @param child    translated version of the single childed node
     * of this weighted ranking node.
     */
    public  ExpNode WeightedRankingToRanking(WeightedRanking wr,
        ExpNode child)
        throws SourceQueryException {
        ExpNode parent = wr.GetParent();
        Vector children = parent.GetChildren();
        float weight = wr.GetRankingValue();
        float minSiblingWeight = weight;
        WAISWeightedRanking wwr;
        int repeat = 0;

        // loop through my siblings and find the minimum ranking value
        for (int i = 0; i < children.size(); i++) {
            Object o;
            float w;
            if ((o = children.elementAt(i)) instanceof WeightedRanking) {
                if ((w = ((WeightedRanking) o).GetRankingValue()) <
                    minSiblingWeight) {
                    if (w != 0.0f) {
                        minSiblingWeight = w;
                    }
                }
            }
        }

        // create a new WAISWeightedRanking with the repeat beig the
        // integer multiple of my weight of the minimum weight
        if (minSiblingWeight > 0.0f) {
            repeat = (int) (weight / minSiblingWeight);
        }
        wwr = new WAISWeightedRanking(wr.GetQuery(), repeat);
        wwr.AddNewChild(child);

        return(wwr);
    }

    /**
     * Return the translated version of a starts Ranking list node
     * within a ranking expression.  The translation is to nested "ors"
     * for each element of the ranking list.
     * @param wr   STARTS  ranking list node to be translated.
     * @param child. Vector of translated children of this ranking
     * list node.
     */

    public  ExpNode RankingListToRanking(RankingList rl, Vector children)
        throws SourceQueryException {

        return BuildRankingListTranslation(rl.GetQuery(), children, 0);
    }

    /**
     * Recursive build up a ranking list translation by nesting
     * a bunch of boolOps
     * @param query    Query context of this ranking list translation
     * @param  children    Vector of children of the original ranking
     * list
     * @param element  Element in the children vector that we are
     * currently processing.
     */
    private ExpNode BuildRankingListTranslation
        (Query query, Vector children, int element) {

        // last child, simply return it.
        if (element == children.size() - 1) {
            return((ExpNode) children.elementAt(element));
        }

        // Build up and return a new bool op that has this child
        // as its left child and a recursively built translation
        // as its right child.
        else {
            WAISBoolOp wbo = new WAISBoolOp(query, "or");
            wbo.AddNewChild((ExpNode) children.elementAt(element));
            wbo.AddNewChild(
                BuildRankingListTranslation(query, children, element+1));
            return wbo;
        }
    }


    /**
     * return true if the input string is a relation operator.
     * @param op   String representation of the operator to test.
     */
    private  boolean RelOp_p(String op) {
        return(op.equals("=") || op.equals(">") || op.equals("<"));
    }

    /**
     * Return a SOIF object that contains the Content Summary for this
     * source.
     */
    public SOIFObject ContentSummaryHeaderToSOIF() throws SourceException {
        SOIFObject outSOIF = new SOIFObject("SContentSummary");;


        outSOIF.SetAttributeValue("Version",
            ServerConfiguration.STARTSVersion);

        // no stemming in WAIS
        outSOIF.SetAttributeValue("Stemming", "F");
        // Stop words are not in the dictionary
        outSOIF.SetAttributeValue("StopWords", "F");
        // We are case sensitive
        outSOIF.SetAttributeValue("CaseSensitive", "T");
        // We can list by field
        outSOIF.SetAttributeValue("Fields", "T");

        // Number of documents
        outSOIF.SetAttributeValue("NumDocs",
            String.valueOf(documentTable.GetNumberDocuments()));

        return(outSOIF);
    }

    public void ContentSummaryFieldsDump(PrintStream out)
        throws SourceException {
        StringBuffer buf = new StringBuffer();
        // now loop through the fields t list the terms
        for (Enumeration e = fieldsTable.keys(); e.hasMoreElements();) {
            String field = (String) e.nextElement();
            String lang = (String) fieldsTable.get(field);
            if (lang.length() == 0) {
                lang = "en-US";
            }
            out.print(SOIFObject.UnparseSOIFElement("Field", field));
            out.print(SOIFObject.UnparseSOIFElement("Language", lang));

            StringTokenizer st = new StringTokenizer(field);
            String attrSet = st.nextToken();
            field = st.nextToken();

/****************************************************************************/
// sdp - changing reference to method TranslateField due to new location.
//       original code:  field = TranslateField(attrSet, field);

            field = GetAttrSet(attrSet).TranslateField(field);

/****************************************************************************/

            WAISDictionary wd = new WAISDictionary(
                indexLocation, sourceID, field);
            WAISInvertedIndex wi =
                new WAISInvertedIndex(indexLocation, sourceID, field);
            WAISDictionaryTerm wdt;
            WAISTermInfo wti;
            int count = 0;

            // loop through all the terms in the dictionary
            while ((wdt = wd.GetNextTerm()) != null) {
                long invOffset = wdt.GetInvertedFileLocation();
                String term = wdt.GetTerm();

                try {
                    wti = wi.GetWAISTermInfo(term, invOffset);
                    int docFreq = wti.GetDocumentFrequency();
                    int totalOccs = wti.GettotalOccs();
                    buf.append(term + " " + String.valueOf(totalOccs) +  " " +
                        String.valueOf(docFreq) + " ");

                    // Every 100 terms we will spit something out
                    if (count++ == 100) {
                        out.print(SOIFObject.UnparseSOIFElement("TermDocFreq", buf.toString()));
                        buf = new StringBuffer();
                        count = 0;
                    }
                }
                // Here is some silliness.  There is no way to tell howm
                // many entries there are in a WAIS dictionary file.
                // So, when we get one that is nonsence, it will raise
                // an exception in the inverted index (its offset
                // will be junk) and we just jump out of the loop.
                // I'd liketo find a better way to do this.
                catch (SourceException e1) {
                    if (buf.length() > 0) {
                        out.print(SOIFObject.UnparseSOIFElement("TermDocFreq", buf.toString()));
                    }
                    break;
                }
            }
            wd.Close();
            wi.Close();
        }

    }

    /**
     * Submit a search to the source search engine.  Return the number
     * of "hits" that the search engine returned.
     * @param query    the query on which to base the search.
     */
    public int DoSourceSearch(Query query) throws SourceException {

        int numResults = 0;
        WAISResultSet wrs = new WAISResultSet(query);
        String results[];

        // We issue both a filter and ranking query to WAIS and
        // then use both to determine the score to assign to the
        // actual document.  Refer to WAISResultSet for more
        // information.
        if (query.sourceFilter != null) {
            // Send query to the server
            String queryString = query.sourceFilter.Unparse();
            results = WAISSearchLink(queryString);
            if (results == null) {
                results = new String[0];
            }
            if ((numResults = results.length) > 0) {
                for (int i = 0; i < numResults; i++) {
                    wrs.AddFilterResult(results[i]);
                }
            }
        }

        // Do the same for the ranking
        if (query.sourceRanking != null) {
            // Send query to the server
            String queryString = query.sourceRanking.Unparse();
            results = WAISSearchLink(queryString);
            if (results == null) {
                results = new String[0];
            }
            if ((numResults = results.length) > 0) {
                for (int i = 0; i < numResults; i++) {
                    wrs.AddRankingResult(results[i]);
                }
            }
        }


        return(wrs.MergeFilterAndRanking());
    }

    /* !!!!! PLATFORM SPECIFIC CODE !!!!! */

   /* !!!!! SOLARIS VERSION !!!!! */
    /**
     * WAIS Search link to the native handler for waissearch
     * @param search   The search string
     */

   private String[] WAISSearchLink(String search)
        throws SourceException{
        return(WAISSearchSend.SearchWAIS(WAISHost, WAISPort, sourceID,
            search));
    }

    /* !!!!! NT VERSION !!!!! */
    // Version of WAISSearchLink allowing the waissearch code itself
    // (in addition to the wais engine) to run on a separate machine.
    // This allows me to debug the java code in a good environment
    // like Symantec Cafe on a NT machine without having to port
    // the WAIS junk to the NT.

    /**
    private String[] WAISSearchLink(String queryString)
        throws SourceException {
        int port = 6790;
        String host = "cs-tr.cs.cornell.edu";
        Socket s;
        String results[];

        try {
            s = new Socket(host, port);

            DataInputStream sin = new DataInputStream(s.getInputStream());
            PrintStream sout = new PrintStream(s.getOutputStream());

            DataInputStream in = new DataInputStream(System.in);

            String line;
            int numResults;

            sout.println(queryString);
            sout.println(sourceID);
            line = sin.readLine();
            if (line == null) {
                throw new SourceException
                    ("Connection closed by server.");
            }

            numResults = Integer.parseInt(line);
            results = new String[numResults];
            for (int i = 0; i < numResults; i++) {
                results[i] = sin.readLine();
                if (results[i] == null) {
                    throw new SourceException
                        ("Connection closed by server.");
                }
            }
            s.close();
            return results;
        }
        catch (IOException e) {
            throw new SourceException(e.getMessage());
        }
    }
    **/

      /* !!!!! END PLATFORM SPECIFIC CODE !!!!! */

    /**
     * Return a string array that is the set of stop words at this
     * source.
     */
    public String[] GetStopWords() {
        int i = 0;
        String retArray[] = new String[stopWordsTable.size()];
        for (Enumeration e = stopWordsTable.keys();
            e.hasMoreElements();) {
            retArray[i++] = (String) e.nextElement();
        }
        return retArray;
    }

    /**
     * Return true or false specifying whether the input string is a
     * stop word in the source.
     * @param string   word to test whether it is a stop word.
     */
    public boolean StopWord_p(String string) {
        return stopWordsTable.containsKey(string);
    }

    /**
     * Load the stop words from the source specific file for this source.
     * This file is called <sourceID>.stop.
     */
    protected void LoadStopWordsFile() {

        // Stop words for this source
        try {
            File stopWordsFile =
                new File(indexLocation, sourceID + ".stop");
            if (stopWordsFile.exists()) {
                FileInputStream inStream =
                    new FileInputStream(stopWordsFile);
                DataInputStream dStream = new DataInputStream(inStream);
                while (dStream.available() > 0) {
                    String word = dStream.readLine();
                    stopWordsTable.put(word, null);
                }
            }
        }
        catch (Exception e) {
            System.out.println(
                "Error openning stop word file for source " +
                sourceID);
            System.exit(1);
        }
    }


}
