/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import java.util.*;

import query.*;
import resource.*;
import results.*;

/**
 * An ExpNode that represents a weighted ranking in a WAIS ranking
 * expression.  The way we simulate weighted rankings (since WAIS has
 * no support for ranking expressions is to repeat the child
 * that descends from the weighted ranking multiple times.  The
 * repetition amount is the (integer) muliple that the weight of the
 * term is of the lowest weighted term).  For example, assume
 * a STARTS expression like list(("distributed" 0.6)("databases" 0.3))
 * This terms in to a WAIS expression of:
 * distribed distributed databases
 * (OR is implicit between the words).
 * So a WAISWeightedRanking has one child (the term to be repeated)
 * and an integer value which is the multiple number of times
 * this term should be repeated.
 */
public class WAISWeightedRanking extends ExpNode implements SourceNode {

    /**
     * Create a new WAISWeightedRanking
     * @param query    The query context of the word literal
     * @param repeat   "weight" (repeat count) of the child.
     */
    public WAISWeightedRanking(Query query, int repeat) {
        super(query);
        SetPropertyValue("repeat", new Integer(repeat));
    }

    public int GetRepeat() {
        Integer repeat = (Integer) GetPropertyValue("repeat");
        return(repeat.intValue());
    }

    /**
     * Return the string representation of the subtree
     * rooted at this weighted ranking
     */
    public String Unparse() {
        Vector children = GetChildren();
        StringBuffer buf = new StringBuffer();
        int repeat = GetRepeat();

        // Just repeat the unparsing of the child "repeat" times.
        String childString = ((ExpNode) children.elementAt(0)).Unparse();
        buf.append(childString);
        for (int i = 1; i < repeat; i++) {
            buf.append(" or " + childString);
        }
        return buf.toString();
    }


    /**
     * This is a slot to do some primitive ranking on this sub-tree.
     * Unimplemented now except for the recursion
     * @param doc  The document to rank using this sub-tree
     */
    public float RankBy(Document doc) {
        Vector children = GetChildren();
        float lRank = ((SourceNode) children.elementAt(0)).RankBy(doc);
        return ((float) 0.0);
    }
}
