/**********************************************************************
 *                              STARTS                                *
 *      Stanford Protocol for Internet Retrieval and Search           *
 *                                                                    *
 *                      Reference Implementation                      *
 *                                                                    *
 *                           Carl Lagoze                              *
 *                      lagoze@cs.cornell.edu                         *
 *                   Department of Computer Science                   *
 *                        Cornell University                          *
 *********************************************************************/

package wais;

import query.*;
import results.*;
import resource.*;

/**
 * An ExpNode that represents a word literal in a WAIS query
 * expression.  Each WAISWordLiteral has no children and two
 * properties
 * soundex - a boolean indicating whether this token
 * should be matched using soundex.
 * literal - the String token.
 */
public class WAISWordLiteral extends ExpNode  implements SourceNode{

    /**
     * Create a new WAISWordLiteral
     * @param query    The query context of the word literal
     * @param soundex  A boolean indicating whether this token
     * should be matched using soundex.
     * @param literal  the String token.
     */
    public WAISWordLiteral(Query query, boolean soundex, String literal) {
        super(query);
        SetPropertyValue("soundex", new Boolean(soundex));
        SetPropertyValue("literal", literal);
    }

    /**
     * Predicate determining wether this token should be matched
     * using soundex.
     */
    public boolean Soundex_p() {
        Boolean soundex = (Boolean) GetPropertyValue("soundex");
        return(soundex.booleanValue());
    }


    /**
     * Return the string that is the word literal.
     */
    public String GetLiteral() {
        return((String) GetPropertyValue("literal"));
    }

    /**
     * Return the string representation of the subtree
     * rooted at this word literal
     */
    public String Unparse() {
        if (Soundex_p()) {
            return("soundex " +GetLiteral());
        }
        else {
            return(GetLiteral());
        }

    }

    /**
     * This is a slot to do some primitive ranking on this sub-tree.
     * Unimplemented now except for the recursion
     * @param doc  The document to rank using this sub-tree
     */
    public float RankBy(Document doc) {
        return ((float) 0.0);
    }
}

